/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */


#include "flxhy_abc.h"
#include <flxrbrv_rvs.h>
#include <flxfunction_ope_calc.h>


extern FlxData* flxdata_ref;
#if FLX_DEBUG
  int FlxHy_ABC::Cinst = 0;
#endif
FlxHy_ABC* hy_abc = NULL;


  
FlxHy_ABC::FlxHy_ABC(const tuint Ni, const tuint Ncali, const tuint Npred)
: Ni(Ni), Ncali(Ncali), Npred(Npred), N(Ni+Ncali+Npred),
true_rainfall(NULL), true_discharge(NULL), measured_rain(NULL), measured_disc(NULL),
// the actual parameters
true_a(0.30), true_b(0.65), true_c(0.06), true_d(0.02), true_e(0.025), true_f(0.02), true_store(200.), true_Rstor(200.), 
true_sd_merr_rain(0.1), true_sd_merr_disc(0.02), true_sd_merr_mal(3)
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxPythonBox' created ...";
      throw FlxException("FlxHy_ABC::FlxHy_ABC_1", ssV.str() );
    }
  #endif
  // allocate dynamic memory
    true_rainfall = new tdouble[N];
    true_discharge = new tdouble[N];
    measured_rain = new tdouble[N];
    measured_disc = new tdouble[N];
  // check if parameters are valid
    if (true_a+true_b > 1.) {
      throw FlxException_Crude("FlxHy_ABC::FlxHy_2");
    }    
}

FlxHy_ABC::~FlxHy_ABC()
{
  delete [] measured_disc;
  delete [] measured_rain;
  delete [] true_discharge;
  delete [] true_rainfall;
} 
 
void FlxHy_ABC::read(FlxIstream& is)
{
  tdouble d = ZERO;
  for (tuint i=0;i<N;++i) {
    is.get_value(d);
    is.get_value(d);
    true_discharge[i] = d;
    is.get_value(d);
    true_rainfall[i] = d;
    is.get_value(d);
    is.get_value(d);
    is.get_value(d);
    measured_rain[i] = d;
    is.get_value(d);
    measured_disc[i] = d;
  }
}

void FlxHy_ABC::run_true_model(std::ostream& sout, const bool is_simple)
{
  // let it rain ...
    for (tuint i=0;i<N;++i) true_rainfall[i] = 0;
    tuint rID = 0;
    RBRV_entry_RV_Poisson* poi = new RBRV_entry_RV_Poisson("poi",rID,new FlxFunction(new FunNumber(5.)));
    tdouble gam_mu, gam_sd;
    rID = 0;
    RBRV_entry_RV_gamma* gam = new RBRV_entry_RV_gamma("gam",rID,true,new FlxFunction(new FunConst(&gam_mu)),new FlxFunction(new FunConst(&gam_sd)),NULL,false);
    tdouble da;
    for (tuint i=0;i<N;++i) {
      if (flxdata_ref->RndCreator.gen_smp_binary(0.1)==false) continue;
      // duration of rainfall
	da = flxdata_ref->RndCreator.gen_smp();
	#if FLX_DEBUG
	  poi->set_invalid(false);
	#endif
	tuint nd = poi->transform_y2x(da);
	nd += i;
	if (nd > N) nd = N;
      // simulate rainfall for duration
	gam_mu = 400.;
	gam_sd = sqrt(800.);
	for (tuint j=i;j<nd;++j) {
	  da = flxdata_ref->RndCreator.gen_smp();
	  #if FLX_DEBUG
	    gam->set_invalid(false);
	  #endif
	  tdouble rt = gam->transform_y2x(da);
	  true_rainfall[j] += rt;
	  if (rt < GlobalVar.TOL()) break;
	  gam_mu = 0.35*rt;
	  gam_sd = sqrt(0.7*rt);
	}
    }
  // measure the rainfall
    if (is_simple==false) {
      const tdouble mer_chi = sqrt(log(pow2(true_sd_merr_rain)+1));
      const tdouble mer_lambda = -0.5*pow2(mer_chi);
      flxVec mrv(true_sd_merr_mal);
      flxdata_ref->RndCreator.gen_smp(mrv);
      const tdouble mvrc = 1./sqrt(tdouble(true_sd_merr_mal));
      for (tuint i=0;i<N;++i) {
	mrv[i%true_sd_merr_mal] = flxdata_ref->RndCreator.gen_smp();
	tdouble s = 0.;
	for (tuint j=0;j<true_sd_merr_mal;++j) {
	  s += mrv[j]*mvrc;
	}
	measured_rain[i] = exp(s*mer_chi+mer_lambda) * true_rainfall[i];
      }
    } else {
      for (tuint i=0;i<N;++i) {
	measured_rain[i] = true_rainfall[i];
      }
    }
  // run the true system
    if (is_simple==false) {
      tdouble Sprev = true_store;
      tdouble Rprev = true_Rstor;
      tdouble Snext;
      gam_mu = ONE;
      gam_sd = true_sd_merr_disc;
      for (tuint i=0;i<N;++i) {
	true_discharge[i] = (ONE-true_a-true_b)*true_rainfall[i]+true_c*Sprev;
	Snext = (ONE-true_c-true_d)*Sprev+true_a*true_rainfall[i]+true_e*Rprev;
	Rprev = (ONE-true_e-true_f)*Rprev+true_d*Sprev;
	Sprev = Snext;
	// measure the discharge
	  da = flxdata_ref->RndCreator.gen_smp();
	  #if FLX_DEBUG
	    gam->set_invalid(false);
	  #endif
	  const tdouble gV = gam->transform_y2x(da);
	  Snext = true_discharge[i];
	  measured_disc[i] = gV*Snext*(0.85+0.008*Snext-0.00008*pow2(Snext));
	// output
	  sout << GlobalVar.Double2String(i,false,-1,0) << " " 
	  << GlobalVar.Double2String(true_discharge[i],false,-1,0) << " " 
	  << GlobalVar.Double2String(true_rainfall[i],false,-1,0) << " " 
	  << GlobalVar.Double2String(Sprev,false,-1,0) << " " 
	  << GlobalVar.Double2String(Rprev,false,-1,0) << " " 
	  << GlobalVar.Double2String(measured_rain[i],false,-1,0) << " "
	  << GlobalVar.Double2String(measured_disc[i],false,-1,0)
	  << std::endl;
      }
    } else {
      tdouble Sprev = true_store;
      RBRV_entry_RV_lognormal* transf = new RBRV_entry_RV_lognormal("logn",0,1,new FlxFunction(new FunNumber(1.0)),new FlxFunction(new FunNumber(0.05)), new FlxFunction(new FunNumber(ZERO)),NULL,NULL,true);
      FlxFunction* rho = flxdata_ref->ReadManager.parse_function("exp(-1*(deltax/4))");
      RBRV_set_proc* dm = new RBRV_set_proc(false,N,N,"hyabc_i_dm",false,transf,rho,1.,0,NULL,2,true,true);
      flxdata_ref->rbrv_box.register_set(dm);
      tuint itmp = 0;
      dm->print(GlobalVar.slogcout(1),"  ",itmp,false);
      flxVec yv(N);
      flxdata_ref->RndCreator.gen_smp(yv);
      dm->set_y(yv.get_tmp_vptr_const());
      dm->transform_y2x();
      dm->get_x(yv.get_tmp_vptr());
      for (tuint i=0;i<N;++i) {
	true_discharge[i] = (1.-true_a-true_b)*true_rainfall[i]+true_c*Sprev;
	Sprev = (1.-true_c-true_d)*Sprev+true_a*true_rainfall[i];
	// measure the discharge
	  measured_disc[i] = true_discharge[i] * yv[i];
	// output
	  sout << GlobalVar.Double2String(i,false,-1,0) << " " 
	  << GlobalVar.Double2String(true_discharge[i],false,-1,0) << " " 
	  << GlobalVar.Double2String(true_rainfall[i],false,-1,0) << " " 
	  << GlobalVar.Double2String(Sprev,false,-1,0) << " " 
	  << GlobalVar.Double2String(measured_rain[i],false,-1,0) << " "
	  << GlobalVar.Double2String(measured_disc[i],false,-1,0)
	  << std::endl;
      }
    }
  delete poi;
  delete gam;
}

void FlxHy_ABC::run_model(const tdouble a, const tdouble b, const tdouble c, tdouble& S0, const tuint N, const tdouble*const rt, tdouble* dt)
{
  tdouble Sprev = S0;
  for (tuint i=0;i<N;++i) {
    dt[i] = (1.-a-b)*rt[i] + c*Sprev;
    Sprev = (1.-c)*Sprev + a*rt[i];
  }
  S0 = Sprev;
}



  
  
  
  
  
  
// -------------------------- Model Init ----------------------------------------------------------------------
  
FlxObjHyABCinit::~FlxObjHyABCinit()
{
  delete fNi;
  delete fNcali;
  delete fNpred;
}

void FlxObjHyABCinit::task()
{
  if (hy_abc!=NULL) {
    std::ostringstream ssV;
    ssV << "'HyABCinit' must be executed only once.";
    throw FlxException("FlxObjHyABCinit::task", ssV.str()); 
  }
  const tuint Ni = fNi->cast2positive_or0(false);
  const tuint Ncali = fNcali->cast2positive(false);
  const tuint Npred = fNpred->cast2positive_or0(false);
  hy_abc = new FlxHy_ABC(Ni,Ncali,Npred);
}

FlxObjBase* FlxObjReadHyABCinit::read()
{
  reader->getChar('(',false);
  FlxFunction* fNi = new FlxFunction(funReader,false);
  FlxFunction* fNcali = NULL;
  FlxFunction* fNpred = NULL;
  try {
    reader->getChar(',',false);
    fNcali = new FlxFunction(funReader,false);
    reader->getChar(',',false);
    fNpred = new FlxFunction(funReader,false);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjHyABCinit(get_doLog(),fNi,fNcali,fNpred);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadHyABCinit::read",1);
    delete fNi;
    if (fNcali) delete fNcali;
    if (fNpred) delete fNpred;
    throw;
  }
}

FlxObjReadHyABCinit::~FlxObjReadHyABCinit()
{
  if (hy_abc) delete hy_abc;
}


// -------------------------- Run true model ----------------------------------------------------------------------


void FlxObjHyABCgenReality::task()
{
  hy_abc->run_true_model(sout(),is_simple);
  tuint N = hy_abc->Ncali + hy_abc->Npred;
  // get the output vector
    const std::string tvns = tv_name->eval_word(true);
    tdouble* smvp = data->ConstMtxBox.get_Vec(tvns,N);
    flxVec outpv(smvp,N);
    const flxVec discv(hy_abc->get_true_disc(),N);
    outpv = discv;
}

FlxObjReadHyABCgenReality::FlxObjReadHyABCgenReality() : FlxObjReadOutputBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"flxhy_abc::simple"));
    ParaBox.insert("simple", "flxhy_abc::simple" );
}

FlxObjBase* FlxObjReadHyABCgenReality::read()
{
  FlxString* tv_name = NULL;
  try {
    reader->getChar('(');
    tv_name = new FlxString(false,false);
    reader->getChar(')');
    read_optionalPara(false);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadHyABCgenReality::read",1);
    if (tv_name) delete tv_name;
    throw;
  }
  return new FlxObjHyABCgenReality(get_doLog(),get_stream(),get_optPara_bool("simple"),tv_name);
}


// -------------------------- Load data from file ----------------------------------------------------------------------

void FlxObjHyABCload::task()
{
  const std::string isns = is_name->eval_word(true);
  FlxIstream& is = data->IstreamBox.get(isns);
  hy_abc->read(is);
}

FlxObjBase* FlxObjReadHyABCload::read()
{
  FlxString* is_name = NULL;
  try {
    reader->getChar('(');
    is_name = new FlxString(false,false);
    reader->getChar(')');
    read_optionalPara(false);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadHyABCload::read",1);
    if (is_name) delete is_name;
    throw;
  }
  return new FlxObjHyABCload(get_doLog(),get_stream(),is_name);
}


// -------------------------- return measured discharge ----------------------------------------------------------------------


void FlxObjHyABCget_measdisc::task()
{
  tuint N = hy_abc->Ncali + hy_abc->Npred;
  // get the output vector
    const std::string tvns = tv_name->eval_word(true);
    tdouble* smvp = data->ConstMtxBox.get_Vec(tvns,N);
    flxVec outpv(smvp,N);
    const flxVec discv(get_true?(hy_abc->get_true_disc()):(hy_abc->get_measured_disc()),N);
    outpv = discv;
}

FlxObjReadHyABCget_measdisc::FlxObjReadHyABCget_measdisc() : FlxObjReadBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"flxhy_abc::get_true"));
    ParaBox.insert("get_true", "flxhy_abc::get_true" );
}

FlxObjBase* FlxObjReadHyABCget_measdisc::read()
{
  FlxString* tv_name = NULL;
  try {
    reader->getChar('(');
    tv_name = new FlxString(false,false);
    reader->getChar(')');
    read_optionalPara(false);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadHyABCget_measdisc::read",1);
    if (tv_name) delete tv_name;
    throw;
  }
  return new FlxObjHyABCget_measdisc(get_doLog(),tv_name,get_optPara_bool("get_true"));
}

// -------------------------- Run model ----------------------------------------------------------------------


FlxObjHyABCrunModel::FlxObjHyABCrunModel(const bool dolog, const std::string& ostreamV, const bool predP, FlxFunction* a, FlxFunction* b, FlxFunction* c, FlxFunction* S0, FlxFunction* mer_cor, FlxFunction* mer_sd, FlxString* tv_name, const bool do_output, const bool is_simple)
: FlxObjOutputBase(dolog,ostreamV), predP(predP), N(predP?(hy_abc->Ncali+hy_abc->Npred):hy_abc->Ncali), 
  a(a), b(b), c(c), S0(S0), mer_cor(mer_cor), mer_sd(mer_sd), tv_name(tv_name),
  rt(new tdouble[N]), dt(new tdouble[N]), do_output(do_output), is_simple(is_simple)
{
  
}

FlxObjHyABCrunModel::~FlxObjHyABCrunModel()
{
  delete a;
  delete b;
  delete c;
  delete S0;
  delete mer_cor;
  delete mer_sd;
  delete tv_name;
  delete [] rt;
  delete [] dt;
}

void FlxObjHyABCrunModel::task()
{
  // get the input-output-vector
    const std::string tvns = tv_name->eval_word(true);
    tuint N_hlp = N;
    tdouble* smvp = data->ConstMtxBox.get_Vec(tvns,N_hlp,true);
  if (is_simple==false) {
    // set up the proposed rain
    const tdouble phi = mer_cor->cast2positive_or0(false);
    const tdouble sd_eps = sqrt(1.-pow2(phi));
    const tdouble sde = mer_sd->cast2positive(false);
    const tdouble mer_chi = sqrt(log(pow2(sde)+1.));
    const tdouble mer_lambda = -0.5*pow2(mer_chi);
    rt[0] = smvp[0];
    for (tuint i=1;i<N;++i) {
      rt[i] = rt[i-1]*phi + smvp[i]*sd_eps;	// correlated standard normal
    }
    data->ConstantBox.insert("hyabc_const_re",rt[N-1]);
    for (tuint i=0;i<N;++i) {
      rt[i] = exp(rt[i]*mer_chi+mer_lambda);	// transform to LogN
      rt[i] *= hy_abc->get_measured_rain()[i]; // mutliply with measured rain
    }
  } else {
    for (tuint i=0;i<N;++i) {
      rt[i] = hy_abc->get_measured_rain()[i];
    }
  }
  // obtain the parameters
    tdouble S0d = S0->cast2positive_or0(false);
    const tdouble av = a->cast2positive_or0(false);
    const tdouble bv = b->cast2positive_or0(false);
    const tdouble cv = c->cast2positive_or0(false);
  // run the model
    hy_abc->run_model(av,bv,cv,S0d,N,rt,dt);
  data->ConstantBox.insert("hyabc_const_s0",S0d);
  // store discharge
    const flxVec discv(dt,N);
    flxVec outpv(smvp,N);
    outpv = discv;
  // calculate Likelihood
    tdouble L = 0.;
    for (tuint i=0;i<N;++i) {
      L += fabs(dt[i] - hy_abc->get_measured_disc()[i]);
    }
    L /= N;
//     L = 1./L;
    L = exp(-L);
  // output
    if (do_output) {
      for (tuint i=0;i<N;++i) {
	sout() << GlobalVar.Double2String(i,false,-1,0) 
	<< GlobalVar.Double2String(dt[i],false,-1,0) << GlobalVar.Double2String(hy_abc->get_measured_disc()[i],false,-1,0) 
	<< GlobalVar.Double2String((dt[i] - hy_abc->get_measured_disc()[i]),false,-1,0) << GlobalVar.Double2String(hy_abc->get_true_disc()[i],false,-1,0)
	<< GlobalVar.Double2String(rt[i],false,-1,0) << GlobalVar.Double2String(hy_abc->get_measured_rain()[i],false,-1,0)
	<< std::endl;
      }
    }
  data->ConstantBox.insert("hyabc_const_l",L);
}

FlxObjReadHyABCrunModel::FlxObjReadHyABCrunModel(): FlxObjReadOutputBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"flxhy_abc::output"));
    ParaBox.insert("output", "flxhy_abc::output" );
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"flxhy_abc::simple"));
    ParaBox.insert("simple", "flxhy_abc::simple" );
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"flxhy_abc::pred"));
    ParaBox.insert("pred", "flxhy_abc::pred" );
}

FlxObjBase* FlxObjReadHyABCrunModel::read()
{
  FlxFunction* a = NULL;
  FlxFunction* b = NULL;
  FlxFunction* c = NULL;
  FlxFunction* S0 = NULL;
  FlxFunction* mer_cor = NULL;
  FlxFunction* mer_sd = NULL;
  FlxString* tv_name = NULL;
  reader->getChar('(');
  try {
  a = new FlxFunction(funReader,false);
  reader->getChar(',');
  b = new FlxFunction(funReader,false);
  reader->getChar(',');
  c = new FlxFunction(funReader,false);
  reader->getChar(',');
  S0 = new FlxFunction(funReader,false);
  reader->getChar(',');
  mer_cor = new FlxFunction(funReader,false);
  reader->getChar(',');
  mer_sd = new FlxFunction(funReader,false);
  reader->getChar(',');
  tv_name = new FlxString(false,false);
  reader->getChar(')');
  read_optionalPara(false);
  return new FlxObjHyABCrunModel(get_doLog(),get_stream(),get_optPara_bool("pred"),a,b,c,S0,mer_cor,mer_sd,tv_name,get_optPara_bool("output"),get_optPara_bool("simple"));
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadHyABCrunModel::read",1);
    if (a) delete a;
    if (b) delete b;
    if (c) delete c;
    if (S0) delete S0;
    if (mer_cor) delete mer_cor;
    if (mer_sd) delete mer_sd;
    if (tv_name) delete tv_name;
    throw;
  }
}

FlxObjHyABCmodCorr::FlxObjHyABCmodCorr(const bool dolog, const std::string& ostreamV, FlxString* set_name, FlxFunction* sd, FlxFunction* rho_Rain, FlxFunction* rho_noRain, const bool predP, const bool onlyPred, const bool updateEVP)
: FlxObjOutputBase(dolog,ostreamV), set_name(set_name), sd(sd), rho_Rain(rho_Rain), rho_noRain(rho_noRain), predP(predP), onlyPred(onlyPred), updateEVP(updateEVP)
{
  
}

FlxObjHyABCmodCorr::~FlxObjHyABCmodCorr()
{
  delete set_name;
  delete sd;
  delete rho_noRain;
  delete rho_Rain;
}

void FlxObjHyABCmodCorr::task()
{
  if (predP==false&&onlyPred) throw FlxException_Crude("FlxObjHyABCmodCorr::task_1");
  const std::string name = set_name->eval_word(true);
  const tuint Ndim = (predP?(hy_abc->Ncali+hy_abc->Npred):(hy_abc->Ncali));
  const tdouble sd2 = pow2(sd->cast2positive());
  const tdouble rhoR = rho_Rain->calc();
  const tdouble rhoNR = rho_noRain->calc();
  // get the covariance matrix
    RBRV_set_base* sb = data->rbrv_box.get_set(name,false);
    RBRV_set_MVN* ts = NULL;
    RBRV_set_MVN_cond* ts2 = NULL;
    FlxMtxSym* CovM = NULL;
    if (sb) {
      if (onlyPred) {
	ts2 = dynamic_cast<RBRV_set_MVN_cond*>(sb);
      } else {
	ts = dynamic_cast<RBRV_set_MVN*>(sb);
      }
    } 
    if (ts) {
      CovM = ts->get_CovM();
      CovM->set_zeroMtx();
    } else if (ts2) {
      CovM = ts2->get_CovM();
      CovM->set_zeroMtx();
    } else {
      CovM = new FlxMtxSym(Ndim);
    }
  // set-up the covariance matrix
    bool itsRaining = false;	
    tuint phaseBegin = 0;
    tdouble prevRain = ZERO;
    const tdouble rainThres = 40.;
    const tdouble* rfall = hy_abc->get_measured_rain();
    for (tuint i=0;i<Ndim;++i) {
      if (itsRaining) {
	if (rfall[i]<=rainThres) {
	  itsRaining = false;
	  phaseBegin = i;
	} else if (rfall[i]-prevRain>rainThres) {
	  phaseBegin = i;
	}
      } else {
	if (rfall[i]>rainThres) {
	  itsRaining = true;
	  phaseBegin = i;
	}
      }
      prevRain = rfall[i];
      const tdouble r = (itsRaining?rhoR:rhoNR);
      for (tuint j=phaseBegin;j<i;++j) {
	CovM->operator()(i,j) = exp(-1*((i-j)/r))*sd2;
      }
      CovM->operator()(i,i) = sd2;
    }
  if (ts || ts2) {	// update the set
    if (updateEVP) {
      if (onlyPred) {
	ts2->update_EVP();
      } else {
	ts->update_EVP();
      }
    }
  } else {	// generate the set
    try {
      if (onlyPred) {
	flxVec* mu = new flxVec(Ndim);	// zero mean vector
	flxVec xobsv(hy_abc->Ncali);
	ts2 = new RBRV_set_MVN_cond(false,hy_abc->Npred,hy_abc->Ncali,name,false,mu,CovM,xobsv);
	data->rbrv_box.register_set(ts2);
      } else {
	flxVec* mu = new flxVec(Ndim);	// zero mean vector
	ts = new RBRV_set_MVN(false,Ndim,0,name,false,mu,CovM,2);
	data->rbrv_box.register_set(ts);
      }
      GlobalVar.slog(4) << "rbrv_mvn: created new set '" << name << "'." << std::endl;
    } catch (FlxException& e) {
      FLXMSG("FlxObjHyABCmodCorr::task_2",1);
      if (ts) delete ts;
      throw;
    }
  }
}


FlxObjReadHyABCmodCorr::FlxObjReadHyABCmodCorr(): FlxObjReadOutputBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"flxhy_abc::pred"));
    ParaBox.insert("pred", "flxhy_abc::pred" );
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"flxhy_abc::onlypred"));
    ParaBox.insert("onlypred", "flxhy_abc::onlypred" );
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"flxhy_abc::updateevp"));
    ParaBox.insert("updateevp", "flxhy_abc::updateevp" );
}

FlxObjBase* FlxObjReadHyABCmodCorr::read()
{
  FlxString* set_name = new FlxString(false,false);
  FlxFunction* sd = NULL;
  FlxFunction* rho_Rain = NULL;
  FlxFunction* rho_noRain = NULL;
  try {
    reader->getChar('(');
    sd = new FlxFunction(funReader,false);
    reader->getChar(',');
    rho_Rain = new FlxFunction(funReader,false);
    reader->getChar(',');
    rho_noRain = new FlxFunction(funReader,false);
    reader->getChar(')');
    read_optionalPara(false);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadHyABCmodCorr::read",1);
    delete set_name;
    if (sd) delete sd;
    if (rho_Rain) delete rho_Rain;
    if (rho_noRain) delete rho_noRain;
    throw;
  }
  return new FlxObjHyABCmodCorr(get_doLog(),get_stream(),set_name,sd,rho_Rain,rho_noRain,get_optPara_bool("pred"),get_optPara_bool("onlypred"),get_optPara_bool("updateevp"));
}




