/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxhy_abc_H
#define fesslix_flxhy_abc_H

#include "flxlibs_interface.h"



class FlxHy_ABC {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this class - only one class may exist
      */
      static int Cinst;
    #endif
    const tuint Ni;		// number of init-steps (only for the true model)
  public:
    const tuint Ncali;		// length of the calibration phase
    const tuint Npred;		// length of the prediction phase
  private:
    const tuint N;		// Ni+Ncali+Npred
    tdouble* true_rainfall;
    tdouble* true_discharge;
    tdouble* measured_rain;
    tdouble* measured_disc;
    
    // parameters describing the real world case
    const tdouble true_a;
    const tdouble true_b;
    const tdouble true_c;
    const tdouble true_d;
    const tdouble true_e;
    const tdouble true_f;
    const tdouble true_store;
    const tdouble true_Rstor;
    const tdouble true_sd_merr_rain;	// standard deviation of measurement error of rain gauge
    const tdouble true_sd_merr_disc;	// standard deviation of random noise in discharge measurement
    const tuint true_sd_merr_mal;	// length of moving average
  public:
    FlxHy_ABC(const tuint Ni, const tuint Ncali, const tuint Npred);
    ~FlxHy_ABC();
    /**
    * @brief reads data from file ... however, it does not read the true parameters of the model
    */
    void read(FlxIstream& is);
    
    const tdouble* const get_measured_rain() const { return measured_rain+Ni; }
    const tdouble* const get_measured_disc() const { return measured_disc+Ni; }
    const tdouble* const get_true_disc() const { return true_discharge+Ni; }
    
    void run_true_model(std::ostream& sout, const bool is_simple);
    
    /**
    * @brief run the model
    * @param S0 initial storage -> returns storage after last time-step
    * @param N number of time steps
    * @param rt rainfall vector
    * @param dt discharge vector
    */
    void run_model(const tdouble a, const tdouble b, const tdouble c, tdouble& S0, const tuint N, const tdouble*const rt, tdouble* dt);
    
};



// -------------------------- Model Init ----------------------------------------------------------------------

class FlxObjHyABCinit : public FlxObjBase {
  private:
    FlxFunction* fNi;
    FlxFunction* fNcali;
    FlxFunction* fNpred;
    void task();
  public:
    FlxObjHyABCinit( const bool dolog, FlxFunction* fNi, FlxFunction* fNcali, FlxFunction* fNpred) : FlxObjBase(dolog), fNi(fNi), fNcali(fNcali), fNpred(fNpred) {}
    ~FlxObjHyABCinit();
};

/**
* @brief object read class: 
*/
class FlxObjReadHyABCinit : public FlxObjReadBase {
  public:
    FlxObjReadHyABCinit() : FlxObjReadBase(true) {}
    ~FlxObjReadHyABCinit();
    FlxObjBase* read ();
};


// -------------------------- Run true model ----------------------------------------------------------------------

class FlxObjHyABCgenReality : public FlxObjOutputBase {
  private:
    const bool is_simple;
    FlxString* tv_name;
    
    void task();
  public:
    FlxObjHyABCgenReality( const bool dolog, const std::string& ostreamV, const bool is_simple, FlxString* tv_name) : FlxObjOutputBase(dolog,ostreamV), is_simple(is_simple), tv_name(tv_name) {}
    ~FlxObjHyABCgenReality() { delete tv_name; }
};

/**
* @brief object read class: 
*/
class FlxObjReadHyABCgenReality : public FlxObjReadOutputBase {
  public:
    FlxObjReadHyABCgenReality();
    FlxObjBase* read ();
};

// -------------------------- Load data from file ----------------------------------------------------------------------

class FlxObjHyABCload : public FlxObjOutputBase {
  private:
    FlxString* is_name;
    
    void task();
  public:
    FlxObjHyABCload( const bool dolog, const std::string& ostreamV, FlxString* is_name) : FlxObjOutputBase(dolog,ostreamV), is_name(is_name) {}
    ~FlxObjHyABCload() { delete is_name; }
};

/**
* @brief object read class: 
*/
class FlxObjReadHyABCload : public FlxObjReadOutputBase {
  public:
    FlxObjReadHyABCload() : FlxObjReadOutputBase(true) {}
    FlxObjBase* read ();
};

// -------------------------- return measured discharge ----------------------------------------------------------------------

class FlxObjHyABCget_measdisc : public FlxObjBase {
  private:
    FlxString* tv_name;
    const bool get_true;
    
    void task();
  public:
    FlxObjHyABCget_measdisc( const bool dolog, FlxString* tv_name, const bool get_true) : FlxObjBase(dolog), tv_name(tv_name), get_true(get_true) {}
    ~FlxObjHyABCget_measdisc() { delete tv_name; }
};

/**
* @brief object read class: 
*/
class FlxObjReadHyABCget_measdisc : public FlxObjReadBase {
  public:
    FlxObjReadHyABCget_measdisc();
    FlxObjBase* read ();
};


// -------------------------- Run model ----------------------------------------------------------------------

class FlxObjHyABCrunModel : public FlxObjOutputBase {
  private:
    const bool predP;
    const tuint N;
    FlxFunction* a;
    FlxFunction* b;
    FlxFunction* c;
    FlxFunction* S0;
    FlxFunction* mer_cor;
    FlxFunction* mer_sd;
    FlxString* tv_name;
    tdouble* rt;
    tdouble* dt;
    const bool do_output;
    const bool is_simple;
    
    void task();
  public:
    FlxObjHyABCrunModel( const bool dolog, const std::string& ostreamV, const bool predP, FlxFunction* a, FlxFunction* b, FlxFunction* c, FlxFunction* S0, FlxFunction* mer_cor, FlxFunction* mer_sd, FlxString* tv_name, const bool do_output, const bool is_simple);
    ~FlxObjHyABCrunModel();
};

/**
* @brief object read class: 
*/
class FlxObjReadHyABCrunModel : public FlxObjReadOutputBase {
  public:
    FlxObjReadHyABCrunModel();
    FlxObjBase* read ();
};


// -------------------------- Set up model correlation ----------------------------------------------------------------------

class FlxObjHyABCmodCorr : public FlxObjOutputBase {
  private:
    FlxString* set_name;
    FlxFunction* sd;
    FlxFunction* rho_Rain;
    FlxFunction* rho_noRain;
    const bool predP;
    const bool onlyPred;
    const bool updateEVP;
    
    void task();
  public:
    FlxObjHyABCmodCorr( const bool dolog, const std::string& ostreamV, FlxString* set_name, FlxFunction* sd, FlxFunction* rho_Rain,FlxFunction* rho_noRain, const bool predP, const bool onlyPred, const bool updateEVP );
    ~FlxObjHyABCmodCorr();
};

/**
* @brief object read class: 
*/
class FlxObjReadHyABCmodCorr : public FlxObjReadOutputBase {
  public:
    FlxObjReadHyABCmodCorr();
    FlxObjBase* read ();
};





#endif // fesslix_flxhy_abc_H

