/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */


#include "flxhy_hymod.h"


extern FlxData* flxdata_ref;
#if FLX_DEBUG
  int FlxHy_hymod::Cinst = 0;
#endif
FlxHy_hymod hy_hymod;


  
FlxHy_hymod::FlxHy_hymod(const tuint Nq_max)
: Nq_max(Nq_max),
  XCuz_0(ZERO), Xq_0(Nq_max), Xs_0(ZERO), XCuz_N(ZERO), Xq_N(Nq_max), Xs_N(ZERO), 
  cmax(ZERO), B(ZERO), Alp(ZERO), Nq(3), Kq(ZERO), Ks(ZERO)
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxHy_hymod' created ...";
      throw FlxException("FlxHy_hymod::FlxHy_hymod_1", ssV.str() );
    }
  #endif 
}

FlxHy_hymod::~FlxHy_hymod()
{

} 

void FlxHy_hymod::set_para_cmax(const tdouble V)
{
  if (V<ZERO) throw FlxException_NeglectInInteractive("FlxHy_hymod::set_para_huz","Negative values are not allowed");
  cmax = V;
}

void FlxHy_hymod::set_para_B(const tdouble V)
{
  if (V<ZERO) throw FlxException_NeglectInInteractive("FlxHy_hymod::set_para_B","Negative values are not allowed");
  B = V;
}

void FlxHy_hymod::set_para_Alp(const tdouble V)
{
  if (V<ZERO) throw FlxException_NeglectInInteractive("FlxHy_hymod::set_para_Alp","Negative values are not allowed");
  Alp = V;
}

void FlxHy_hymod::set_para_Nq(const tuint V)
{
  Nq = V;
}

void FlxHy_hymod::set_para_Kq(const tdouble V)
{
  if (V<ZERO) throw FlxException_NeglectInInteractive("FlxHy_hymod::set_para_Kq","Negative values are not allowed");
  Kq = V;
}

void FlxHy_hymod::set_para_Ks(const tdouble V)
{
  if (V<ZERO) throw FlxException_NeglectInInteractive("FlxHy_hymod::set_para_Ks","Negative values are not allowed");
  Ks = V;
}


void FlxHy_hymod::set_initial_state_S(const tdouble V)
{
  if (V<ZERO) throw FlxException_NeglectInInteractive("FlxHy_hymod::set_initial_state_XCuz","Negative values are not allowed");
  XCuz_0 = V;
}

void FlxHy_hymod::set_initial_state_Xq(const tuint NqV, const tdouble* const XqV)
{
  if (NqV>Nq_max) throw FlxException_NeglectInInteractive("FlxHy_hymod::set_initial_state_Xq_1","Vector size is too large.");
  if (NqV<Nq_max) Xq_0.set_zero();
  for (tuint i=0;i<NqV;++i) {
    if (XqV[i]<ZERO) {
      throw FlxException_NeglectInInteractive("FlxHy_hymod::set_initial_state_Xq_2","Negative values are not allowed.");
    }
    Xq_0[i] = XqV[i];
  }
}

void FlxHy_hymod::set_initial_state_Xs(const tdouble V)
{
  if (V<ZERO) throw FlxException_NeglectInInteractive("FlxHy_hymod::set_initial_state_Xs","Negative values are not allowed");
  Xs_0 = V;
}

const tdouble FlxHy_hymod::get_final_state_S() const
{
  return XCuz_N;
}

void FlxHy_hymod::get_final_state_Xq(flxVec& XqV) const
{
  if (Nq>XqV.get_N()) throw FlxException_Crude("FlxHy_hymod::get_final_state_Xq");
  if (Nq<XqV.get_N()) XqV.set_zero();
  for (tuint i=0;i<Nq;++i) {
    XqV[i] = Xq_N[i];
  }
}

const tdouble FlxHy_hymod::get_final_state_Xs() const
{
  return Xs_N;
}

const tdouble FlxHy_hymod::linres(tdouble& storage, const tdouble inflow, const tdouble K) const
{
  const tdouble tt = ONE-K;
  if (tt<=ZERO) {
    const tdouble res = storage+inflow;
    storage = ZERO;
    return res;
  } else {
    storage = tt*(storage+inflow);
    return (K/tt)*storage;
  }
}

void FlxHy_hymod::run_model(const size_t N, const flxVec PP, const flxVec PET, flxVec& D, const tdouble F)
{
  // perform some initial checks
    PP.check_size(N);
    PET.check_size(N);
    D.check_size(N);
  // initial assignments
    XCuz_N = XCuz_0;
    Xs_N = Xs_0;	// current state of the slow-flow tank
    Xq_N = Xq_0;	// current states of the quick-flow tank
  // Loop over forcing data
    const tdouble x_max = cmax/(B+ONE);
    for (size_t i=0;i<N;++i) {
      // compute excess precipitation and evapotranspiration
	const tdouble XCuz_N_prev = XCuz_N;
	const tdouble ct_prev = cmax*(ONE-pow(ONE-(XCuz_N_prev/x_max),ONE/(B+ONE)));	// C from previous time-step
	// calculate rainfall that cannot be saturated -> i.e. that directly produces runoff
	  tdouble ER1 = PP[i]-(cmax-ct_prev);
	  if (ER1<ZERO) ER1 = ZERO;
	// compute the rainfall that can be partially saturated
	  const tdouble Pval = PP[i]-ER1;
	// optain the new filling level
	  tdouble ct_now = (ct_prev+Pval)/cmax;	
	  if (ct_now>ONE) ct_now = ONE;
	// compute the respective storage
	  XCuz_N = x_max*(ONE-pow(ONE-ct_now,B+ONE));
	// calculate the additional rainfall that could not be saturated completely
	  tdouble ER2 = Pval-(XCuz_N-XCuz_N_prev);
	  if (ER2<ZERO) ER2 = ZERO;
	// storage loss due to evapotranspiration
	  const tdouble evap = (ONE-(x_max-XCuz_N)/x_max)*PET[i];
	  XCuz_N = XCuz_N-evap;
	  if (XCuz_N<ZERO) XCuz_N = ZERO;
      // total runoff at the current time-step
	const tdouble Q0 = ER1 + ER2;
      // partition Q0 between quick and slow flow tanks
	tdouble Qfast = Alp*Q0;
	tdouble Qslow = Q0-Qfast;
      // route slow flow component with single linear reservoir
	Qslow = linres(Xs_N,Qslow,Ks);
      // route quick flow component with linear reservoirs
	for (tuint k=0;k<Nq;++k) {
	  Qfast = linres(Xq_N[k],Qfast,Kq);
	}
	D[i] = Qslow + Qfast;
    }
  // transform output
    D *= F;
}



// -------------------------- Run model ----------------------------------------------------------------------


FlxObjHyModRunModel::FlxObjHyModRunModel(const bool dolog, FlxFunction* N_f, FlxFunction* F_f, FlxMtxConstFun* D_f, FlxMtxConstFun* PP_f, FlxMtxConstFun* PET_f)
: FlxObjBase(dolog), N_f(N_f), F_f(F_f), D_f(D_f), PP_f(PP_f), PET_f(PET_f)
{
  
}

FlxObjHyModRunModel::~FlxObjHyModRunModel()
{
  delete PET_f;
  delete PP_f;
  delete D_f;
  delete F_f;
  delete N_f;
}

void FlxObjHyModRunModel::task()
{
  const tuint N = N_f->cast2tuint(false);
  const tdouble F = F_f->cast2positive(false);
  tuint N_hlp = N;
  const flxVec PP(data->ConstMtxBox.get_Vec(PP_f->eval(),N_hlp,true),N);
  const flxVec PET(data->ConstMtxBox.get_Vec(PET_f->eval(),N_hlp,true),N);
  flxVec D(data->ConstMtxBox.get_Vec(D_f->eval(),N_hlp,false),N);
  hy_hymod.run_model(N,PP,PET,D,F);
}

FlxObjBase* FlxObjReadHyModrunModel::read()
{
  FlxFunction* N_f = NULL;
  FlxFunction* F_f = NULL;
  FlxMtxConstFun* D_f = NULL;
  FlxMtxConstFun* PP_f = NULL;
  FlxMtxConstFun* PET_f = NULL;
  try {
  reader->getWord("n",false);
  reader->getChar('=',false);
  N_f = new FlxFunction(funReader,false);
  reader->getChar(',',false);
  reader->getWord("d",false);
  reader->getChar('=',false);
  D_f = new FlxMtxConstFun(false);
  reader->getChar(',',false);
  reader->getWord("pp",false);
  reader->getChar('=',false);
  PP_f = new FlxMtxConstFun(true);
  reader->getChar(',',false);
  reader->getWord("pet",false);
  reader->getChar('=',false);
  PET_f = new FlxMtxConstFun(true);
  reader->getChar(',',false);
  reader->getWord("f",false);
  reader->getChar('=',false);
  F_f = new FlxFunction(funReader,false);
  read_optionalPara(false);
  return new FlxObjHyModRunModel(get_doLog(),N_f,F_f,D_f,PP_f,PET_f);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadHyModrunModel::read",1);
    if (N_f) delete N_f;
    if (F_f) delete F_f;
    if (D_f) delete D_f;
    if (PP_f) delete PP_f;
    if (PET_f) delete PET_f;
    throw;
  }
}

// -------------------------- Set parameters and states ----------------------------------------------------------------------

FlxObjHyModSet::FlxObjHyModSet(const bool dolog, const FlxObjHyModSet::HyModPSID Val2Set, FlxFunction* InpF, FlxMtxConstFun* InpV)
: FlxObjBase(dolog), Val2Set(Val2Set), InpF(InpF), InpV(InpV)
{
  
}

FlxObjHyModSet::~FlxObjHyModSet()
{
  delete InpF;
  delete InpV;
}

void FlxObjHyModSet::task()
{
  switch(Val2Set) {
    case s:
      hy_hymod.set_initial_state_S(InpF->cast2positive_or0(false));
      break;
    case xq:
      {
	tuint N=0;
	tdouble* vtmp = data->ConstMtxBox.get_Vec(InpV->eval(),N,false);
	hy_hymod.set_initial_state_Xq(N,vtmp);
      }
      break;
    case xs:
      hy_hymod.set_initial_state_Xs(InpF->cast2positive_or0(false));
      break;
    case cmax:
      hy_hymod.set_para_cmax(InpF->cast2positive_or0(false));
      break;
    case b:
      hy_hymod.set_para_B(InpF->cast2positive_or0(false));
      break;
    case alp:
      hy_hymod.set_para_Alp(InpF->cast2positive_or0(false));
      break;
    case nq:
      hy_hymod.set_para_Nq(InpF->cast2tuint(false));
      break;
    case kq:
      hy_hymod.set_para_Kq(InpF->cast2positive_or0(false));
    case ks:
      hy_hymod.set_para_Ks(InpF->cast2positive_or0(false));
      break;
  }
}

FlxObjBase* FlxObjReadHyModSet::read()
{
  // which value is to be set ???
    FlxObjHyModSet::HyModPSID Val2Set;
    const std::string V2S_str = reader->getWord(true,false);
    if (V2S_str=="s") Val2Set = FlxObjHyModSet::s;
    else if (V2S_str=="xq") Val2Set = FlxObjHyModSet::xq;
    else if (V2S_str=="xs") Val2Set = FlxObjHyModSet::xs;
    else if (V2S_str=="cmax") Val2Set = FlxObjHyModSet::cmax;
    else if (V2S_str=="b") Val2Set = FlxObjHyModSet::b;
    else if (V2S_str=="alp") Val2Set = FlxObjHyModSet::alp;
    else if (V2S_str=="nq") Val2Set = FlxObjHyModSet::nq;
    else if (V2S_str=="kq") Val2Set = FlxObjHyModSet::kq;
    else if (V2S_str=="ks") Val2Set = FlxObjHyModSet::ks;
    else {
      throw FlxException("FlxObjReadHyModSet::read","Identifier '"+V2S_str+"' is not allowed.",reader->getCurrentPos());
    }
  reader->getChar('=',false);
  FlxFunction* InpF = NULL;
  FlxMtxConstFun* InpV = NULL;
  try {
    if (Val2Set==FlxObjHyModSet::xq) {
      InpV = new FlxMtxConstFun(true);
    } else {
      InpF = new FlxFunction(funReader,false);
    }
    return new FlxObjHyModSet(get_doLog(),Val2Set,InpF,InpV);
  } catch (FlxException& e) {
    if (InpF) delete InpF;
    if (InpV) delete InpV;
    throw;
  }
}

// -------------------------- Get final states ----------------------------------------------------------------------

FlxObjHyModGet::FlxObjHyModGet(const bool dolog, const FlxObjHyModGet::HyModPGID Val2Get, FlxMtxConstFun* Vname)
: FlxObjBase(dolog), Val2Get(Val2Get), Vname(Vname)
{
  
}

FlxObjHyModGet::~FlxObjHyModGet()
{
  delete Vname;
}

void FlxObjHyModGet::task()
{
  tdouble* fp = NULL;
  if (Val2Get==s||Val2Get==xs) {
    fp = data->ConstantBox.get(Vname->eval(),true);
  } else {
    fp = data->ConstMtxBox.get_Vec(hy_hymod.get_para_Nq(),Vname->eval());
  }
  switch(Val2Get) {
    case s:
      *fp = hy_hymod.get_final_state_S();
      break;
    case xq:
      {
	flxVec XqV(fp,hy_hymod.get_para_Nq());
	hy_hymod.get_final_state_Xq(XqV);
      }
      break;
    case xs:
      *fp = hy_hymod.get_final_state_Xs();
      break;
  }
}

FlxObjBase* FlxObjReadHyModGet::read()
{
  FlxMtxConstFun* Vname = new FlxMtxConstFun(true);
  try {
    reader->getChar('=',false);
    // which value should be retrieved?
      FlxObjHyModGet::HyModPGID Val2Get;
      const std::string V2G_str = reader->getWord(true,false);
      if (V2G_str=="s") Val2Get = FlxObjHyModGet::s;
      else if (V2G_str=="xq") Val2Get = FlxObjHyModGet::xq;
      else if (V2G_str=="xs") Val2Get = FlxObjHyModGet::xs;
      else {
	throw FlxException("FlxObjReadHyModGet::read","Identifier '"+V2G_str+"' is not allowed.",reader->getCurrentPos());
      }
    return new FlxObjHyModGet(get_doLog(),Val2Get,Vname);
  } catch (FlxException& e) {
    if (Vname) delete Vname;
    throw;
  }
}








