/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxhy_hymod_H
#define fesslix_flxhy_hymod_H

#include "flxlibs_interface.h"



/**
* Hydrological model HyMod. 
* The implementation here is based on the Matlab implementation from the DREAM-package.
*/
class FlxHy_hymod {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this class - only one class may exist
      */
      static int Cinst;
    #endif
    const tuint Nq_max;
    // initial states
      tdouble XCuz_0;		// Initial basin storage [mm]
      flxVec Xq_0;		// Initial storage of quick flow tanks [mm]
      tdouble Xs_0;		// Initial storage of slow flow tanks [mm]
    // final states
      tdouble XCuz_N;
      flxVec Xq_N;
      tdouble Xs_N;
    // model parameters
      tdouble cmax;		// Maximum storage capacity encountered in the basin [mm]
      tdouble B;		// Shape parameter of the distribution of storage capacities within the basin [-]
      tdouble Alp;		// Routing ratio quick to slow pathway [-]
      tuint Nq;			// Number of quick flow linear tanks [-]
      tdouble Kq;		// Residence time quickflow routing tanks [1/day]
      tdouble Ks;		// Residence time slow routing tanks [1/day]

    /**
    * @brief linear reservoir
    * @param storage initial storage -> updated storage
    * @param inflow to reservoir (mm/day)
    * @param K leakage rate (1/day)
    * @returns output of reservoir (mm/day)
    */
    const tdouble linres(tdouble& storage, const tdouble inflow, const tdouble K) const;
  public:
    FlxHy_hymod(const tuint Nq_max=10);
    ~FlxHy_hymod();

    /**
    * @brief set model parameters
    */
    void set_para_cmax(const tdouble V);
    void set_para_B(const tdouble V);
    void set_para_Alp(const tdouble V);
    void set_para_Nq(const tuint V);
    void set_para_Kq(const tdouble V);
    void set_para_Ks(const tdouble V);
    const tuint get_para_Nq() const { return Nq; };
    /**
    * @brief set initial state of the model
    */
    void set_initial_state_S(const tdouble V);
    void set_initial_state_Xq(const tuint NqV, const tdouble* const XqV);
    void set_initial_state_Xs(const tdouble V);
    /**
    * @brief retrieve the final model sates
    */
    const tdouble get_final_state_S() const;
    void get_final_state_Xq(flxVec& XqV) const;
    const tdouble get_final_state_Xs() const;
    
    /**
    * @brief run the model
    * @param N	 how many time-steps to solve?
    * @param PP  [N] precipitation
    * @param PET [N] potential evapotranspiration
    * @param D   [N] computed discharge
    * @param F   Area factor (to translate output from mm/d to m^3/s
    */
    void run_model(const size_t N, const flxVec PP, const flxVec PET, flxVec& D, const tdouble F);
 
};



// -------------------------- Run model ----------------------------------------------------------------------

class FlxObjHyModRunModel : public FlxObjBase {
  private:
    FlxFunction* N_f;		// number of time-steps
    FlxFunction* F_f;		// area factor
    FlxMtxConstFun* D_f;	// discharge (output)
    FlxMtxConstFun* PP_f;	// precipitation
    FlxMtxConstFun* PET_f;	// evapotranspiration

    void task();
  public:
    FlxObjHyModRunModel( const bool dolog, FlxFunction* N_f, FlxFunction* F_f, FlxMtxConstFun* D_f, FlxMtxConstFun* PP_f, FlxMtxConstFun* PET_f);
    ~FlxObjHyModRunModel();
};

/**
* @brief object read class: 
*/
class FlxObjReadHyModrunModel : public FlxObjReadBase {
  public:
    FlxObjReadHyModrunModel() : FlxObjReadBase(true) {}
    FlxObjBase* read ();
};

// -------------------------- Set parameters and states ----------------------------------------------------------------------

class FlxObjHyModSet : public FlxObjBase {
  public:
    enum HyModPSID {s,xq,xs,cmax,b,alp,nq,kq,ks};
  private:
    HyModPSID Val2Set;		// ID of parameter or state to set
    FlxFunction* InpF;		// input function
    FlxMtxConstFun* InpV;	// input vector

    void task();
  public:
    FlxObjHyModSet( const bool dolog, const HyModPSID Val2Set, FlxFunction* InpF, FlxMtxConstFun* InpV);
    ~FlxObjHyModSet();
};

/**
* @brief object read class: 
*/
class FlxObjReadHyModSet : public FlxObjReadBase {
  public:
    FlxObjReadHyModSet() : FlxObjReadBase(true) {}
    FlxObjBase* read ();
};


// -------------------------- Get final states ----------------------------------------------------------------------

class FlxObjHyModGet : public FlxObjBase {
  public:
    enum HyModPGID {s,xq,xs};
  private:
    HyModPGID Val2Get;		// ID of final state to retrieve
    FlxMtxConstFun* Vname;	// name of the target variable

    void task();
  public:
    FlxObjHyModGet( const bool dolog, const HyModPGID Val2Get, FlxMtxConstFun* Vname);
    ~FlxObjHyModGet();
};

/**
* @brief object read class: 
*/
class FlxObjReadHyModGet : public FlxObjReadBase {
  public:
    FlxObjReadHyModGet() : FlxObjReadBase(true) {}
    FlxObjBase* read ();
};




#endif // fesslix_flxhy_hymod_H

