/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */


#define VERSION_REV 18
#define VERSION_DATE "2017-05-22"

#include "flxpython.h"

#if defined(_MSC_VER)
  #include "flxlibpython_export.h"
#else
  #define FLXLIBPYTHON_EXPORT
#endif

#ifdef __WINDOWS__
  #pragma comment(lib, "flxlib")
#endif

flxlibs_interface flx_iface;


#if FLX_DEBUG
  int FlxPythonBox::Cinst = 0;
#endif
FlxPythonBox PythBox;
FlxData* flxdata_ref = NULL;
  

extern "C" FLXLIBPYTHON_EXPORT void flxlib_init( FlxObjectReadBox* objReadBox, FlxData* dataBox ) {
  GlobalVar.slog(4) << "  Python-library - Version " << VERSION_REV << " (" << VERSION_DATE << ")" << std::endl;
  flx_iface.set_boxes(objReadBox,&(dataBox->FunBox));
  // objects
    flx_iface.add_obj("python_cmd", new FlxObjReadPythonCmd()); 
    flx_iface.add_obj("python_fun", new FlxObjReadPythonFun()); 
    flx_iface.add_obj("python_module", new FlxObjReadPythonModule()); 
    flx_iface.add_obj("python_path", new FlxObjReadPythonPath()); 
    flx_iface.add_obj("python_var_get", new FlxObjReadPythonVarGet()); 
    flx_iface.add_obj("python_var_set", new FlxObjReadPythonVarSet()); 
  // functions
    flxdata_ref = dataBox;
}

extern "C" FLXLIBPYTHON_EXPORT void flxlib_close() {
  flx_iface.free_memory();
}



// -------------------------- PYTHON - Extension ----------------------------------------------------------------------

static PyObject* flxpy_testfun(PyObject *self, PyObject *args)
{
  if(!PyArg_ParseTuple(args, ":numargs")) return NULL;
  return Py_BuildValue("i", 34);
}

static PyObject* flxpy_const(PyObject *self, PyObject *args)
{
  const char* cn;
  if(!PyArg_ParseTuple(args, "s", &cn)) return NULL;
  std::string sn(cn);
  const tdouble *tp = flxdata_ref->ConstantBox.get(sn,false);
  if (tp==NULL) {
    std::ostringstream ssV;
    ssV << "The Fesslix-constant '" << sn << "' does not exist.";
    throw FlxException("flxpy_const", ssV.str() );
  }
  return Py_BuildValue("f", *tp);
}

static PyObject* flxpy_var(PyObject *self, PyObject *args)
{
  const char* cn;
  if(!PyArg_ParseTuple(args, "s", &cn)) return NULL;
  std::string sn(cn);
  transform(sn.begin(), sn.end(), sn.begin(), (int(*)(int)) std::tolower);
  FlxFunction* fp = flxdata_ref->VarBox.get(sn);
  if (fp==NULL) {
    std::ostringstream ssV;
    ssV << "The Fesslix-var-variable '" << sn << "' does not exist.";
    throw FlxException("flxpy_const", ssV.str() );
  }
  const tdouble t = fp->calc();
  return Py_BuildValue("f", t);
}

static PyObject* flxpy_vec(PyObject *self, PyObject *args)
{
  const char* cn;
  if(!PyArg_ParseTuple(args, "s", &cn)) return NULL;
  std::string sn(cn);
  std::transform(sn.begin(), sn.end(), sn.begin(), (int(*)(int)) std::tolower);
  tuint N = 0;
  const tdouble *dp = flxdata_ref->ConstMtxBox.get_Vec(sn,N);
  PyObject* ptuple = PyTuple_New(N);
  if (ptuple==NULL) throw FlxException_Crude("flxpy_vec_1");
  for (tuint i=0;i<N;++i) {
    if (PyTuple_SetItem(ptuple,i,Py_BuildValue("f", dp[i]))!=0) {
      throw FlxException_Crude("flxpy_vec_2");
    }
  }
  return ptuple;
}

static PyObject* flxpy_fun(PyObject *self, PyObject *args)
{
  const char* cn;
  if(!PyArg_ParseTuple(args, "s", &cn)) return NULL;
  std::string sn(cn);
  FlxFunction* f = flxdata_ref->ReadManager.parse_function(sn);
  tdouble t;
  try {
    t = f->calc();
  } catch (FlxException &e) {
    FLXMSG("flxpy_fun",1);
    delete f;
    throw;
  }
  delete f;
  return Py_BuildValue("f", t);
}


static PyMethodDef FlxpyMethods[] = {
  {"testfun", flxpy_testfun, METH_VARARGS,
  "A simple function to test the interface."},
  {"const", flxpy_const, METH_VARARGS,
  "const('NAME'): returns the value of the Fesslix-constant NAME."},
  {"var", flxpy_var, METH_VARARGS,
  "var('NAME'): returns the value of the Fesslix-var-variable NAME."},
  {"vec", flxpy_vec, METH_VARARGS,
  "vec('NAME'): returns the value of the Fesslix-vector-variable NAME."},
  {"fun", flxpy_fun, METH_VARARGS,
  "fun('FLXFUNCTION'): evaluates the FLXFUNCTION."},
  {NULL, NULL, 0, NULL}
};

static PyModuleDef FlxModule = {
    PyModuleDef_HEAD_INIT, "flx", NULL, -1, FlxpyMethods,  NULL, NULL, NULL, NULL
};

static PyObject*
PyInit_flx(void)
{
    return PyModule_Create(&FlxModule);
}

// ------------------------------------------------------------------------------------------------
  
FlxPythonBox::FlxPythonBox()
: initialized(false)
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxPythonBox' created ...";
      throw FlxException("FlxPythonBox::FlxPythonBox", ssV.str() );
    }
  #endif
}

FlxPythonBox::~FlxPythonBox()
{
  // clean up modules
    for (std::map<std::string, PyObject*>::iterator pos = module_box.begin(); pos != module_box.end(); ++pos) Py_DECREF(pos->second);
  // Finish the Python Interpreter
    Py_Finalize();
}

void FlxPythonBox::init_python_actual()
{
  if (initialized) return;
  // Initialize the Python Interpreter
    // program name
      wchar_t *program = Py_DecodeLocale("Fesslix", NULL);
      if (program == NULL) {
	throw FlxException_Crude("FlxPythonBox::init_python_actual");
      }
      Py_SetProgramName(program);
    PyImport_AppendInittab("flx", &PyInit_flx);		// Python 3
    Py_Initialize();
    PyRun_SimpleString("import sys");
    //Py_InitModule("flx", FlxpyMethods);			// Python 2.7 
    PyRun_SimpleString("import flx");
    initialized = true;
}

PyObject* FlxPythonBox::get_module(const std::string& name)
{
  std::map<std::string, PyObject*>::iterator pos = module_box.find(name);
  if ( pos != module_box.end() ) {
    return pos->second;
  } else {
    return NULL;
  }
}

void FlxPythonBox::insert_module(const std::string& name, const bool throw_err)
{
  // search for module
    if (get_module(name)) {
      if (throw_err) {
	std::ostringstream ssV;
	ssV << "Python module '" << name << "' has already been loaded.";
	throw FlxException("FlxPythonBox::insert_module_1", ssV.str() );
      } 
      return;
    }
  init_python(); // make sure interpreter is initialized
  // load the module
    PyObject *pName = PyUnicode_DecodeFSDefault(name.c_str());
    PyObject *pModule = PyImport_Import(pName);
    Py_DECREF(pName);
    if (pModule == NULL) {
      std::ostringstream ssV;
      ssV << "There was an error while loading the Python module '" << name << "'.";
      PyErr_Print();
      throw FlxException("FlxPythonBox::insert_module_2", ssV.str() );
    }
  // insert module
    std::pair<std::string, PyObject*>Element(name, pModule);
    module_box.insert(Element);
}

PyObject* FlxPythonBox::get_function(const std::string& mod_name, const std::string& fun_name)
{
  insert_module(mod_name,false);
  PyObject *pFunc = PyDict_GetItemString(PyModule_GetDict(get_module(mod_name)),fun_name.c_str());
  if (PyCallable_Check(pFunc)==false) {
    std::ostringstream ssV;
    ssV << "There is something wrong with the python function '" << fun_name << "' in module '" << mod_name << "'.";
    throw FlxException("FlxPythonBox::get_function", ssV.str() );
  }
  return pFunc;
}

const tdouble FlxPythonBox::get_variable_double(const std::string& mod_name, const std::string& var_name)
{
  return PyFloat_AsDouble(get_variable(mod_name,var_name));
}

PyObject* FlxPythonBox::get_variable(const std::string& mod_name, const std::string& var_name)
{
  insert_module(mod_name,false);
  PyObject *pVar = PyDict_GetItemString(PyModule_GetDict(get_module(mod_name)),var_name.c_str());
  if (pVar==NULL) {
    std::ostringstream ssV;
    ssV << "There is something wrong with the python variable '" << var_name << "' in module '" << mod_name << "'.";
    throw FlxException("FlxPythonBox::get_variable", ssV.str() );
  }
  return pVar;
}

void FlxPythonBox::set_variable(const std::string& mod_name, const std::string& var_name, const tdouble value)
{
  insert_module(mod_name,false);
  PyObject * pt = PyFloat_FromDouble(value);
  PyDict_SetItemString(PyModule_GetDict(get_module(mod_name)), var_name.c_str(), pt);
  Py_DECREF(pt);
}

void FlxPythonBox::register_path(const std::string& path)
{
  init_python();
  std::string ts = "sys.path.append('" + path + "')";
  PyRun_SimpleString(ts.c_str());
}

void FlxPythonBox::execute_cmd(const std::string& cmdStr)
{
  init_python();
  PyRun_SimpleString(cmdStr.c_str());
}




void FlxObjPythonFun::task()
{
  flx_iface.add_fun( funname, new FunReadFunPython(funname,pymodname,pyfunname) );
}

void FlxObjPythonVarGet::task()
{
  flx_iface.add_fun( funname, new FunReadFunPythonVar(funname,pymodname,pyvarname) );
}

void FlxObjPythonVarSet::task()
{
  PythBox.set_variable(pymodname,pyvarname,fun->calc());
}

void FlxObjPythonModule::task()
{
  PythBox.insert_module(pymodname,false);
}

void FlxObjPythonPath::task()
{
  std::string ts = pathStr->eval(false);
  PythBox.register_path(ts);
}

void FlxObjPythonCmd::task()
{
  std::string ts = cmdStr->eval(false);
  PythBox.execute_cmd(ts);
}

FlxObjBase* FlxObjReadPythonCmd::read()
{
  FlxString* pS = new FlxString(true,false);
  try {
    read_optionalPara(false);
    return new FlxObjPythonCmd(get_doLog(),pS);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadPython::read_2",1);
    delete pS;
    throw;
  }
}

FlxObjBase* FlxObjReadPythonFun::read()
{
  const std::string funname = reader->getWord(true,false,false);
  reader->getChar('=');
  const std::string pymodname = reader->getWord(false,false,false);
  reader->getChar('.');
  const std::string pyfunname = reader->getWord(false,false,false);
  read_optionalPara(false);
  return new FlxObjPythonFun(get_doLog(),funname,pymodname,pyfunname);
}

FlxObjBase* FlxObjReadPythonModule::read()
{
  const std::string pymodname = reader->getWord(false,false,false);
  read_optionalPara(false);
  return new FlxObjPythonModule(get_doLog(),pymodname);
}

FlxObjBase* FlxObjReadPythonPath::read()
{
  FlxString* pS = new FlxString(false,false);
  try {
    read_optionalPara(false);
    return new FlxObjPythonPath(get_doLog(),pS);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadPython::read_1",1);
    delete pS;
    throw;
  }
}

FlxObjBase* FlxObjReadPythonVarGet::read()
{
  const std::string funname = reader->getWord(true,false,false);
  reader->getChar('=');
  const std::string pymodname = reader->getWord(false,false,false);
  reader->getChar('.');
  const std::string pyvarname = reader->getWord(false,false,false);
  read_optionalPara(false);
  return new FlxObjPythonVarGet(get_doLog(),funname,pymodname,pyvarname);
}

FlxObjBase* FlxObjReadPythonVarSet::read()
{
  const std::string pymodname = reader->getWord(false,false,false);
  reader->getChar('.');
  const std::string pyvarname = reader->getWord(false,false,false);
  reader->getChar('=');
  FlxFunction* fun = new FlxFunction(funReader,false);
  try {      
    read_optionalPara(false);
    return new FlxObjPythonVarSet(get_doLog(),fun,pymodname,pyvarname);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadPython::read_3",1);
    delete fun;
    throw;
  }
}

const std::string FunPython::write_v()
{
  return funname;
}

const tdouble FunPython::calc()
{
  const tuint S = ParaList->size();
  PyObject* pArgs = PyTuple_New(S);
  for (tuint i=0;i<S;++i) {
    PyObject* pValue = PyFloat_FromDouble(ParaListP[i]->calc());
    if (!pValue) {
      throw FlxException_Crude("FunPython::calc_1");
    }
    PyTuple_SetItem(pArgs, i, pValue); 
  }
  
  PyObject* pValue = PyObject_CallObject(pyfun, pArgs);
  if (pArgs != NULL)
  {
    Py_DECREF(pArgs);
  }
  if (pValue) {
    const tdouble t = PyFloat_AsDouble(pValue);
    Py_DECREF(pValue);
    return t;
  } else {
    std::ostringstream ssV;
    ssV << "Error in Python.";
    PyErr_Print();
    throw FlxException_NeglectInInteractive("FunPython::calc_2", ssV.str() );
  }
}

FunReadFunPython::FunReadFunPython(const std::string& funnameV, const std::string& pymodnameV, const std::string& pyfunnameV)
: FunReadFunBase(true), funname(funnameV), pyfun(NULL)
{
  pyfun = PythBox.get_function(pymodnameV,pyfunnameV);
}

FunBase* FunReadFunPython::read(bool errSerious)
{
  return new FunPython( funname, pyfun, read_parameters(-1,errSerious) );
}

const tdouble FunPythonVar::calc()
{
  return PythBox.get_variable_double(pymodname,pyvarname);
}

const std::string FunPythonVar::write()
{
  return funname + "()";
}

FunReadFunPythonVar::FunReadFunPythonVar(const std::string& varname, const std::string& pymodname, const std::string& pyvarname)
: FunReadFunBase(true), varname(varname), pymodname(pymodname), pyvarname(pyvarname)
{
  
}

FunBase* FunReadFunPythonVar::read(bool errSerious)
{
  return new FunPythonVar(varname,pymodname,pyvarname);
}





