/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */


#define VERSION_REV 8
#define VERSION_DATE "2017-05-18"

#if defined(_MSC_VER)
  #include "flxlibrsm_export.h"
#else
  #define FLXLIBRSM_EXPORT
#endif

#include "flxrsm.h"
#include "flxlibs_interface.h"

#ifdef __WINDOWS__
  #pragma comment(lib, "flxlibssfem")
#endif

flxlibs_interface flx_iface;

flxRSMmap RSMmap;


extern "C" FLXLIBRSM_EXPORT void flxlib_init( FlxObjectReadBox* objReadBox, FlxData* dataBox ) {
  GlobalVar.slog(4) << "  RSM-library - Version " << VERSION_REV << " (" << VERSION_DATE << ")" << std::endl;
  flx_iface.set_boxes(objReadBox,&(dataBox->FunBox));
  // objects
    flx_iface.add_obj("rsm", new FlxObjReadRSM());
    flx_iface.add_obj("rsm_set_fun", new FlxObjReadRSM_set_fun());
    flx_iface.add_obj("rsm_set_para", new FlxObjReadRSM_set_para());
    flx_iface.add_obj("rsm_print_coeff", new FlxObjReadRSM_print_coeff());
    flx_iface.add_obj("rsm_coeff_vec", new FlxObjReadRSM_coeff_vec());
    flx_iface.add_obj("rsm_pc_project", new FlxObjReadRSM_pc_project());
  // functions
    flx_iface.add_fun("rsm", new FunReadFunRSM() );
    flx_iface.add_fun("rsm_terms", new FunReadFunRSM_terms() );
}

extern "C" FLXLIBRSM_EXPORT void flxlib_close() {
  flx_iface.free_memory();
}


//-------------------------------------------------------------------------------------------------

flxRSMmap::~flxRSMmap()
{
  for (std::map<std::string,FlxRSM_Fun_base*>::iterator pos = box.begin(); pos != box.end(); ++pos) {
    delete pos->second;
  }
}

void flxRSMmap::set(const std::string& ID, FlxRSM_Fun_base* rsmp)
{
  std::pair<std::string,FlxRSM_Fun_base*> Element(ID, rsmp);
  std::map<std::string,FlxRSM_Fun_base*>::iterator pos = box.find(ID);
  if ( pos != box.end() ) {
    delete pos->second; 
    pos->second = rsmp;
    return;
  }
  box.insert(Element);
}

FlxRSM_Fun_base& flxRSMmap::get(const std::string& ID)
{
  std::map<std::string,FlxRSM_Fun_base*>::iterator pos = box.find(ID);
  if ( pos != box.end() ) {
    FlxRSM_Fun_base* res = pos->second;
    return (*res);
  } else {
    std::ostringstream ssV;
    ssV << "The response surface '" << ID << "' does not exist.";
    throw FlxException("flxRSMmap::get", ssV.str()); 
  }
}


//-------------------------------------------------------------------------------------------------



FlxRSM_Fun_base::FlxRSM_Fun_base(const std::string setStr, flxVec* paraVec)
: setStr(setStr), RndBox(setStr,data->rbrv_box), DIM( RndBox.get_NRV() ), paraVec(paraVec), tempVec(NULL), xVec(NULL)
{
  if (paraVec) {
    tempVec = new flxVec(paraVec->get_N());
    xVec = new flxVec(DIM);
  }
}

FlxRSM_Fun_base::~FlxRSM_Fun_base()
{
  if (paraVec) delete paraVec;
  if (tempVec) delete tempVec;
  if (xVec) delete xVec;
}

void FlxRSM_Fun_base::set_paraVec(flxVec* paraVecV)
{
  if (paraVecV->get_N() != get_NumbOfTerms()) {
    std::ostringstream ssV_2;
    ssV_2 << "Parameter vector has wrong dimension.";
    throw FlxException("FlxRSM_Fun_base::set_paraVec", ssV_2.str() );  
  }
  if (paraVec) delete paraVec;
  paraVec = paraVecV;
  if (tempVec) {
    if (tempVec->get_N()==paraVec->get_N()) return;
    delete tempVec;
  }
  tempVec = new flxVec(paraVec->get_N());
  if (xVec) {
    if (xVec->get_N()==DIM) return;
    delete xVec;
  }
  xVec = new flxVec(DIM);
}

void FlxRSM_Fun_base::set_para(const tuint index, const tdouble paraV)
{
  if (paraVec->get_N() <= index) {
    std::ostringstream ssV_2;
    ssV_2 << "Index (" << index << ") out of range [0;" << paraVec->get_N()-1 << "].";
    throw FlxException("FlxRSM_Fun_base::set_para", ssV_2.str() );  
  }
  (*paraVec)[index] = paraV;
}

const tdouble FlxRSM_Fun_base::get_para(const tuint index)
{
  if (paraVec->get_N() <= index) {
    std::ostringstream ssV_2;
    ssV_2 << "Index (" << index << ") out of range [0;" << paraVec->get_N()-1 << "].";
    throw FlxException("FlxRSM_Fun_base::get_para", ssV_2.str() );  
  }
  return (*paraVec)[index];
}

const tdouble FlxRSM_Fun_base::calc_solution()
{
  if (paraVec == NULL) {
    std::ostringstream ssV_2;
    ssV_2 << "A response surface has not yet been defined.";
    throw FlxException("FlxRSM_Fun_base::calc_solutionVec", ssV_2.str() );  
  }
  RndBox.get_x_Vec(xVec->get_tmp_vptr());
  calc_termVec(*xVec,*tempVec);
  return (*paraVec)*(*tempVec);
}

FlxRSM_Fun_base* FlxRSM_Fun_base::createRSMfun(const std::string setStr, bool errSerious, const std::string ftype)
{
  std::string funType;
  if (ftype=="") {
    funType = reader->getWord(true,errSerious);
  } else {
    funType = ftype;
  }
  if (funType == "p2womt") {
    return new FlxRSM_Fun_p2womt(setStr);
  } else if ( funType == "p2wmt" ) {
    return new FlxRSM_Fun_p2wmt(setStr);
  } else if ( funType == "p1" ) {
    return new FlxRSM_Fun_p1(setStr);
  } else if ( funType == "pc" ) {
    reader->getChar('(',errSerious);
    const int order = reader->get_UInt<int>(false);
    reader->getChar(')',errSerious);
    return new FlxRSM_Fun_pc(setStr,order);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "Type of RSM-function not known.";
    FlxError(errSerious,"FlxRSM_Fun_base::createRSMfun_2", ssV_2.str() ); return NULL; // dummy-return
  }
}

void FlxRSM_Fun_p2womt::calc_termVec(const flxVec& xVec, flxVec& termVec)
{
  #if FLX_DEBUG
    if (termVec.get_N() != get_NumbOfTerms()) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR. (" << termVec.get_N() << ", " << get_NumbOfTerms() << ")";
      throw FlxException("FlxRSM_Fun_p2womt::calc_termVec_0", ssV_2.str() );  
    }
  #endif
  if (xVec.get_N() != DIM) {
    std::ostringstream ssV_2;
    ssV_2 << "Number of random variables has changed.";
    throw FlxException("FlxRSM_Fun_p2womt::calc_termVec_1", ssV_2.str() );  
  }
  termVec[0]=ONE;
  for (tuint i = 1; i <= DIM; ++i) {
    termVec[i] = xVec[i-1];
    termVec[DIM+i] = pow2( xVec[i-1]);
  }
}

FlxRSM_Fun_base* FlxRSM_Fun_p2womt::copy()
{
  return new FlxRSM_Fun_p2womt(setStr, new flxVec(*paraVec));
}

void FlxRSM_Fun_p2wmt::calc_termVec(const flxVec& xVec, flxVec& termVec)
{
  tuint k = FlxRSM_Fun_p2womt::get_NumbOfTerms();
  FlxRSM_Fun_p2womt::calc_termVec(xVec, termVec);
  for (tuint i = 0; i < DIM; ++i) {
    for (tuint j = i+1; j < DIM; ++j, ++k) {
      termVec[k] = xVec[i]*xVec[j];
    }
  }
}

FlxRSM_Fun_base* FlxRSM_Fun_p2wmt::copy()
{
  return new FlxRSM_Fun_p2wmt(setStr, new flxVec(*paraVec));
}

FlxRSM_Fun_base* FlxRSM_Fun_p1::copy()
{
  return new FlxRSM_Fun_p1(setStr, new flxVec(*paraVec));
}

void FlxRSM_Fun_p1::calc_termVec(const flxVec& xVec, flxVec& termVec)
{
  #if FLX_DEBUG
    if (termVec.get_N() != get_NumbOfTerms()) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR. (" << termVec.get_N() << ", " << get_NumbOfTerms() << ")";
      throw FlxException("FlxRSM_Fun_p2womt::calc_termVec_0", ssV_2.str() );  
    }
  #endif
  if (xVec.get_N() != DIM) {
    std::ostringstream ssV_2;
    ssV_2 << "Number of random variables has changed.";
    throw FlxException("FlxRSM_Fun_p2womt::calc_termVec_1", ssV_2.str() );  
  }
  termVec[0]=ONE;
  for (tuint i = 1; i <= DIM; ++i) {
    termVec[i] = xVec[i-1];
  }
}

FlxRSM_Fun_base* FlxRSM_Fun_pc::copy()
{
  return new FlxRSM_Fun_pc(setStr, pc.get_order(), new flxVec(*paraVec));
}

void FlxRSM_Fun_pc::calc_termVec(const flxVec& xVec, flxVec& termVec)
{
  const tulong k = pc.get_SeqNumb();
  #if FLX_DEBUG
    if (termVec.get_N() != k) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR. (" << termVec.get_N() << ", " << get_NumbOfTerms() << ")";
      throw FlxException("FlxRSM_Fun_pc::calc_termVec_0", ssV_2.str() );  
    }
  #endif
  if (xVec.get_N() != DIM) {
    std::ostringstream ssV_2;
    ssV_2 << "Number of random variables has changed.";
    throw FlxException("FlxRSM_Fun_pc::calc_termVec_1", ssV_2.str() );  
  }
  for (tulong i=0;i<k;++i) {
    termVec[i] = pc.calc_Poly(i,xVec);
  }  
}

FlxRSM_SmpPoints_base* FlxRSM_SmpPoints_base::createRSMpts(bool errSerious)
{
  std::string ptsType = reader->getWord(true);
  if (ptsType == "joa") {
    return new FlxRSM_SmpPoints_joa(errSerious);
  } else if (ptsType == "cc") { 
    return new FlxRSM_SmpPoints_cc(errSerious);
  } else if (ptsType == "ff") { 
    return new FlxRSM_SmpPoints_ff(errSerious);
  } else if (ptsType == "p1") { 
    return new FlxRSM_SmpPoints_p1(errSerious);
  } else if (ptsType == "rnd") { 
    return new FlxRSM_SmpPoints_rnd(errSerious);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "Type of RSM-sampling rule not known.";
    FlxError(errSerious,"FlxRSM_SmpPoints_base::createRSMpts", ssV_2.str() ); return NULL; // dummy-return
  }
}

FlxRSM_SmpPoints_joa::FlxRSM_SmpPoints_joa(const bool errSerious) : centerY(NULL)
{
  reader->getWord("centery",errSerious);
  centerYF = new FlxMtxConstFun(true);
  try {
    reader->getWord("k",errSerious);
    kF = new FlxFunction(funReader,errSerious);
  } catch ( FlxException &e) {
    FLXMSG("FlxRSM_SmpPoints_joa::FlxRSM_SmpPoints_joa",1);
    delete centerYF;
    throw;
  }
}

FlxRSM_SmpPoints_joa::~FlxRSM_SmpPoints_joa()
{
  delete centerYF;
  delete kF;
  if (centerY) delete centerY;
}

void FlxRSM_SmpPoints_joa::updateStart(const tuint M)
{
  FlxRSM_SmpPoints_base::updateStart(M);
  if (centerY) delete centerY;
  tuint centerYN = 0;
  centerY = new flxVec(data->ConstMtxBox.get_Vec(centerYF->eval(),centerYN),centerYN);
  if (centerYN != DIM) {
    std::ostringstream ssV_2;
    ssV_2 << "'centerY' has wrong dimension (" << centerYN << " instead of " << DIM << ")";
    throw FlxException("FlxRSM_SmpPoints_joa::updateStart", ssV_2.str() );  
  }
  k = kF->calc();
  index = 0;
}

void FlxRSM_SmpPoints_joa::get_nextPoint(flxVec& y)
{
  #if FLX_DEBUG
    if (DIM != y.get_N()) {
      throw FlxException_Crude("FlxRSM_SmpPoints_joa::get_nextPoint_0" );  
    }
  #endif
  ++index;
  y = *centerY;
  if (index > 1) {
    if (index > get_NumbOfPoints()) {
      throw FlxException_Crude("FlxRSM_SmpPoints_joa::get_nextPoint_1" );  
    }
    tuint j = index/2-1;
    if (index % 2 == 0) {
      y[j] += k;
    } else {
      y[j] -= k;
    }
  }
}

void FlxRSM_SmpPoints_cc::get_nextPoint(flxVec& y)
{
   if ( index < FlxRSM_SmpPoints_joa::get_NumbOfPoints() ) {
     FlxRSM_SmpPoints_joa::get_nextPoint(y);
     return;
   }
   y = *centerY;
   y.add(Comb->Next(),k2);
}

FlxRSM_SmpPoints_cc::FlxRSM_SmpPoints_cc(const bool errSerious) : FlxRSM_SmpPoints_joa(), Comb(NULL)
{
  reader->getWord("k2",errSerious);
  kF2 = new FlxFunction(funReader,errSerious);
}

FlxRSM_SmpPoints_cc::~FlxRSM_SmpPoints_cc()
{
  delete kF2;
  if (Comb != NULL) delete Comb;
}

void FlxRSM_SmpPoints_cc::updateStart(const tuint M)
{
  FlxRSM_SmpPoints_joa::updateStart(M);
  k2 = kF2->calc();
  if (Comb) delete Comb;
  Comb = new FlxCombination(-ONE, ONE, 2*ONE, DIM);
}

void FlxObjRSM::task()
{ 
  GlobalVar.slog(4) << "rsm: generating a response surface" << std::endl;
  RBRV_constructor& RndBox = RSMfunction->get_RndBox();
  SmpPoints->updateStart(RndBox.get_NRV());	
  const tuint Pneed = RSMfunction->get_NumbOfTerms();
  const tuint Pprov = SmpPoints->get_NumbOfPoints();
  
  FlxMtx A(Pprov, Pneed);
  GlobalVar.slog(4) << "rsm: terms in approximation: " << GlobalVar.Double2String(Pneed) << "; points provided: " << GlobalVar.Double2String(Pprov) << std::endl;
  if (Pprov<Pneed) {
    std::ostringstream ssV;
    ssV << "You have to provide at least as many points as terms are in the approximation." << std::endl << "\tPoints provided: " << GlobalVar.Double2String(Pprov) << "; Terms in the approximation: " << GlobalVar.Double2String(Pneed);
    throw FlxException("FlxObjRSM::task", ssV.str() );
  }
  
  flxVec b(Pprov);
  flxVec y(RndBox.get_NRV());
  flxVec x(RndBox.get_NOX());
  flxVec t(Pneed);
  flxVec *p = new flxVec(Pneed);
  
  // Assemble A and b
  for (tuint i = 0; i < Pprov; ++i) {
    SmpPoints->get_nextPoint(y);
    RndBox.set_smp(y);
    RndBox.get_x_Vec(x.get_tmp_vptr());
    RSMfunction->calc_termVec(x,t);
    for (tuint j = 0; j < Pneed; ++j) {
	A(i,j) = t[j];
    }
    b[i] = fung->calc();
  };
  // Solve LSE
    solMeth->solve(A,b,*p);
    RSMfunction->set_paraVec(p);
  RSMmap.set(rsmID,RSMfunction->copy());
  GlobalVar.slog(4) << "rsm: response surface '" << rsmID << "' initialized." << std::endl;
}

FlxObjRSM::~FlxObjRSM()
{
  delete RSMfunction;
  delete SmpPoints;
  delete solMeth;
  delete fung;
}

FlxObjBase* FlxObjReadRSM::read()
{
  FlxFunction* fung=NULL; FlxRSM_Fun_base* RSMfun=NULL;FlxRSM_SmpPoints_base* SmpPts=NULL;
  FlxRSM_Solve_base* solMeth=NULL;
  FlxString* rbrvsets = NULL;
  try {
    reader->getChar('(',false); 
    const std::string rsmID = reader->getWord(true,false);
    reader->getChar(';',false); 
    fung = new FlxFunction(funReader,false);
    reader->getChar(';',false); 
    rbrvsets = new FlxString(false,false);
    const std::string setstr = rbrvsets->eval(true);
    delete rbrvsets; rbrvsets = NULL;
    reader->getChar(';',false); 
    RSMfun = FlxRSM_Fun_base::createRSMfun(setstr,false);
    reader->getChar(';',false); 
    SmpPts = FlxRSM_SmpPoints_base::createRSMpts(false);
    reader->getChar(';',false); 
    solMeth = FlxRSM_Solve_base::createRSMsolve(false);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjRSM(get_doLog(),rsmID,RSMfun,SmpPts,solMeth,fung);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadRSM::read",1);
    if (fung) delete fung;
    if (RSMfun) delete RSMfun;
    if (SmpPts) delete SmpPts;
    if (solMeth) delete solMeth;
    if (rbrvsets) delete rbrvsets;
    throw;
  }
}

FlxObjRSM_set_fun::~FlxObjRSM_set_fun()
{
  delete RSMfunction;
}

void FlxObjRSM_set_fun::task()
{
  GlobalVar.slog(4) << "rsm: generating an empty response surface" << std::endl;
  const tuint Pneed = RSMfunction->get_NumbOfTerms();
  GlobalVar.slog(4) << "rsm: terms in approximation: " << GlobalVar.Double2String(Pneed) << std::endl;
  flxVec *p = new flxVec(Pneed);
  RSMfunction->set_paraVec(p);
  RSMmap.set(rsmID,RSMfunction->copy());
}

FlxObjBase* FlxObjReadRSM_set_fun::read()
{
  reader->getChar('(',false);
  const std::string rsmID = reader->getWord(true,false);
  reader->getChar(';',false);
  FlxString* rbrvsets = new FlxString(false,false);
  std::string setstr;
  try {
    setstr = rbrvsets->eval(true);
  } catch (FlxException &e) {
    delete rbrvsets;
    throw;
  }
  delete rbrvsets; rbrvsets = NULL;
  reader->getChar(';',false);
  FlxRSM_Fun_base* RSMfun = FlxRSM_Fun_base::createRSMfun(setstr,false);
  try {
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjRSM_set_fun(get_doLog(),rsmID,RSMfun);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadRSM_set_fun::read",1);
    delete RSMfun;
    throw;
  }
}


FlxObjRSM_set_para::~FlxObjRSM_set_para()
{
  delete index;
  delete value;
}

void FlxObjRSM_set_para::task()
{
  FlxRSM_Fun_base& rs = RSMmap.get(rsmID);
  rs.set_para(index->cast2tuintW0(),value->calc());
}

FlxObjBase* FlxObjReadRSM_set_para::read()
{
  const std::string rsmID = reader->getWord(true,false);
  reader->getChar('(',false);
  FlxFunction* index = new FlxFunction(funReader,false);
  FlxFunction* valueP=NULL;
  try {
    reader->getChar(')',false);
    reader->getChar('=',false);
    valueP = new FlxFunction(funReader,false);
    return new FlxObjRSM_set_para(get_doLog(),index,rsmID,valueP);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadRSM_set::read_2",1);
    delete index;
    if (valueP) delete valueP;
    throw;
  }
}

FlxObjRSM_pc_project::~FlxObjRSM_pc_project()
{
  delete rv_fun;
  delete RSMfunction;
  delete intervalsF;
  delete smpPintervF;
  delete smpRedistF;
  delete diffusionF;
  delete Bound;
}

void FlxObjRSM_pc_project::task()
{
  FlxRSM_Fun_pc* pcRSM = dynamic_cast<FlxRSM_Fun_pc*>(RSMfunction);
  #if FLX_DEBUG
    if (pcRSM==NULL) throw FlxException_Crude("FlxObjRSM_pc_project::task");
  #endif
  const tuint Pneed = RSMfunction->get_NumbOfTerms();
  GlobalVar.slog(4) << "rsm: generating a polynomial chaos response surface "
    << "(order=" << pcRSM->get_order()  
    << "; terms=" << Pneed
    << ")" << std::endl;
  flxVec *p = new flxVec(Pneed);

  calc_PC_projection_numerical_MCI pcpnmci(pcRSM->pc_reference(),RSMfunction->get_RndBox());
  pcpnmci.calc_expectation(*p,rv_fun,intervalsF->cast2tuint(),smpPintervF->cast2tuint(),smpRedistF->cast2tuint(),diffusionF->calc(),Bound->calc());

  RSMfunction->set_paraVec(p);
  RSMmap.set(rsmID,RSMfunction->copy());
}

FlxObjReadRSM_pc_project::FlxObjReadRSM_pc_project()
{
  AllDefParaBox->insert(new FlxOptionalParaFun(tdouble(608),"rsm::project::intervals"));
  ParaBox.insert("intervals", "rsm::project::intervals" );
  AllDefParaBox->insert(new FlxOptionalParaFun(tdouble(100),"rsm::project::samplesperinterval"));
  ParaBox.insert("samplesperinterval", "rsm::project::samplesperinterval" );
  AllDefParaBox->insert(new FlxOptionalParaFun(tdouble(1e6),"rsm::project::redistsmp"));
  ParaBox.insert("redistsmp", "rsm::project::redistsmp" );
  AllDefParaBox->insert(new FlxOptionalParaFun(tdouble(0.9),"rsm::project::diffusion"));
  ParaBox.insert("diffusion", "rsm::project::diffusion" );
  AllDefParaBox->insert(new FlxOptionalParaFun(tdouble(38),"rsm::project::bound"));
  ParaBox.insert("bound", "rsm::project::bound" );
}

FlxObjBase* FlxObjReadRSM_pc_project::read()
{
  FlxRSM_Fun_base* RSMfun=NULL;
  FlxFunction* rvFun=NULL;
  const std::string rsmID = reader->getWord(true,false);
  try {
    reader->getChar('(');
    FlxString* rbrvsets = new FlxString(false,false);
    std::string setstr;
    try {
      setstr = rbrvsets->eval(true);
    } catch (FlxException &e) {
      delete rbrvsets;
      throw;
    }
    delete rbrvsets; rbrvsets = NULL;
    reader->getChar(')');
    RSMfun = FlxRSM_Fun_base::createRSMfun(setstr,false,"pc");
    reader->getChar('=');
    rvFun = new FlxFunction(funReader,false);
    read_optionalPara(false);
    return new FlxObjRSM_pc_project(get_doLog(),rsmID,RSMfun,rvFun,get_optPara_FlxFunction("intervals"),get_optPara_FlxFunction("samplesperinterval"),get_optPara_FlxFunction("redistsmp"),get_optPara_FlxFunction("diffusion"),get_optPara_FlxFunction("bound"));
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadRSM_set::read_3",1);
    if (RSMfun) delete RSMfun;
    if (rvFun) delete rvFun;
    throw;
  }
}


void FlxObjRSM_print_coeff::task()
{
  FlxRSM_Fun_base& rs = RSMmap.get(rsmID);
  const tuint N = rs.get_NumbOfTerms();
  for (tuint i = 0; i < N; ++i) {
    if (is_verbose()) {
      sout() << i << "\t";
    }
    sout() << GlobalVar.Double2String(rs.get_para(i)) << std::endl;
  }
}

FlxObjBase* FlxObjReadRSM_print_coeff::read()
{
  const std::string rsmID = reader->getWord(true,false);
  read_optionalPara(false);
  return new FlxObjRSM_print_coeff(get_doLog(), get_stream(), get_verbose(), rsmID);
}


void FlxObjRSM_coeff_vec::task()
{
  FlxRSM_Fun_base& rs = RSMmap.get(rsmID);
  data->ConstMtxBox.insert(vname->eval_word(true),new FlxSMtx(rs.get_paraVec()));
}

FlxObjBase* FlxObjReadRSM_coeff_vec::read()
{
  FlxString* vname = new FlxString(false,false);
  try {
    reader->getChar('=',false);
    const std::string rsmID = reader->getWord(true,false);
    read_optionalPara(false);
    return new FlxObjRSM_coeff_vec(get_doLog(),vname,rsmID);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadRSM_coeff_vec::read",1);
    delete vname;
    throw;
  }
}

FlxRSM_Solve_base* FlxRSM_Solve_base::createRSMsolve(bool errSerious)
{
  const std::string ptsType = reader->getWord(true,errSerious);
  if (ptsType == "transpose") {
    return new FlxRSM_Solve_CG(errSerious);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "Type of RSM-solution method not known.";
    FlxError(errSerious,"FlxRSM_Solve_base::createRSMsolve", ssV_2.str() ); return NULL; // dummy-return 
  }
}

void FlxRSM_Solve_CG::solve(const FlxMtx& A, const flxVec& b, flxVec& sol)
{
  FlxMtxSym ATA(sol.get_N());
  flxVec ATb(sol.get_N());
  A.TransposeMmultM(ATA);
  A.TransposeMmultVec(b,ATb);
  tnlong iter = iterF->cast2tnlong();
  tdouble eps = epsF->calc();
  if (eps <= ZERO) {
    std::ostringstream ssV_2;
    ssV_2 << "'eps' (" << eps << ") must be larger than zero.";
    throw FlxException("FlxRSM_Solve_CG::solve_1", ssV_2.str() );  
  }
  int pcn = pcnF->cast2int();
  if ( ! ATA.solve_CG(sol,ATb,eps,iter,pcn,false) ) {
    std::ostringstream ssV_2;
    ssV_2 << "System could not be solved.";
    std::ostringstream ssV_3;
    ssV_2 << "iter=" << iter << "; eps=" << eps;
    throw FlxException("FlxRSM_Solve_CG::solve_2", ssV_2.str(), ssV_3.str() );  
  }
  if (print) {
    GlobalVar.slog(4) << "  Solution: iter=" << GlobalVar.Double2String(iter) << "; eps=" << GlobalVar.Double2String(eps) << "(" << GlobalVar.Double2String(eps/(ATb.get_Norm2())) << ")";
    if (printS) { GlobalVar.slog(4) << "; sol=" << sol; }
    GlobalVar.slog(4) << std::endl;
  }
}

FlxRSM_Solve_CG::~FlxRSM_Solve_CG()
{
  delete iterF;
  delete epsF;
  delete pcnF;
}

FlxRSM_Solve_CG::FlxRSM_Solve_CG(const bool errSerious) :
iterF(new FlxFunction(new FunNumber(1e4))),
epsF(new FlxFunction(new FunNumber(1e-8))),
pcnF(new FlxFunction(new FunNumber(ONE))),
print(true),printS(false)
{
  if (reader->whatIsNextChar() == ';') return;
  try {
    while ( true ) {
      const std::string strV = reader->getWord(true,errSerious);
      reader->getChar('=',errSerious);
      if (strV == "eps") {
	delete epsF; epsF=NULL;
	epsF = new FlxFunction(funReader,errSerious);
      } else if ( strV == "iter" ) {
	delete iterF; iterF=NULL;
	iterF = new FlxFunction(funReader,errSerious);
      } else if ( strV == "pcn" ) {
	delete pcnF; pcnF=NULL;
	pcnF = new FlxFunction(funReader,errSerious);
      } else if ( strV == "print" ) {
	print = reader->getBool(errSerious);
      } else if ( strV == "printsol" ) {
	printS = reader->getBool(errSerious);
      } else {
	std::ostringstream ssV_2;
	ssV_2 << "Unknown keyword (" << strV << ")";
	FlxError(errSerious,"FlxRSM_Solve_CG::FlxRSM_Solve_CG_1", ssV_2.str() );  
      }
      if (reader->whatIsNextChar() != ',' ) break;
      reader->getChar(',',errSerious);
    }
  } catch (FlxException &e) {
    FLXMSG("FlxRSM_Solve_CG::FlxRSM_Solve_CG_2",1);
    if (epsF) delete epsF;
    if (iterF) delete iterF;
    if (pcnF) delete pcnF;
    throw;
  }
}

const bool FunRSM::dependOn_Const(tdouble* thenumber)
{
  return false;
}

const tdouble FunRSM::calc()
{	
  FlxRSM_Fun_base& rs = RSMmap.get(rsmID);
  return rs.calc_solution();
}

const tdouble FunRSM_terms::calc()
{
  FlxRSM_Fun_base& rs = RSMmap.get(rsmID);
  return tdouble(rs.get_NumbOfTerms());
}

FlxCombination::FlxCombination(const tdouble start, const tdouble end, const tdouble step, const tuint dim)
: start(start), end(end), step(step), dim(dim), cv(dim), runs(false)
{
  cv = start;
}

const flxVec& FlxCombination::Next()
{
  if (!runs) {
    runs = true;
    return cv;
  }
  for (tuint i = 0; i < dim; ++i) {
    if (cv[i] >= end) {
      cv[i] = start;
    } else {
      cv[i]+=step;
      return cv;
    }
  }
  std::ostringstream ssV_2;
  ssV_2 << "No combinations can be created anymore.";
  throw FlxException("FlxCombination::Next", ssV_2.str() ); 
}

void FlxRSM_SmpPoints_p1::get_nextPoint(flxVec& y)
{
  #if FLX_DEBUG
    if (DIM != y.get_N()) {
      throw FlxException_Crude("FlxRSM_SmpPoints_p1::FlxRSM_SmpPoints_p1_1" );  
    }
  #endif
  y = *centerY;
  if (index > 0) {
    if (index >= get_NumbOfPoints()) {
      throw FlxException_Crude("FlxRSM_SmpPoints_p1::FlxRSM_SmpPoints_p1_2" );  
    }
    if ( y[index-1] > ZERO ) {
      y[index-1] -= k;
    } else {
      y[index-1] += k;
    }
  }
  ++index;
}


FlxRSM_SmpPoints_ff::FlxRSM_SmpPoints_ff(const bool errSerious)
: FlxRSM_SmpPoints_joa(), Comb(NULL)
{
  reader->getWord("sphere",errSerious);
  sphere = reader->getBool(errSerious);
}

FlxRSM_SmpPoints_ff::~FlxRSM_SmpPoints_ff()
{
  if (Comb) delete Comb;
}	

void FlxRSM_SmpPoints_ff::updateStart(const tuint M)
{
  FlxRSM_SmpPoints_joa::updateStart(M);
  if (Comb) delete Comb;
  Comb = new FlxCombination(-ONE, ONE, ONE, DIM);
}

void FlxRSM_SmpPoints_ff::get_nextPoint(flxVec& y)
{
  y = *centerY;
  const flxVec& t = Comb->Next();
  if (sphere) {
    const tdouble l = t.get_Norm2();
    if (l > GlobalVar.TOL()) {
      y.add(t,k/l);
    }
  } else {
    y.add(t,k);
  }
}

FlxRSM_SmpPoints_rnd::FlxRSM_SmpPoints_rnd(const bool errSerious)
: FlxRSM_SmpPoints_joa()
{
  reader->getWord("numb",errSerious);
  numb = new FlxFunction(funReader,errSerious);
}

void FlxRSM_SmpPoints_rnd::get_nextPoint(flxVec& y)
{
  if (y.get_N() != centerY->get_N()) {
    std::ostringstream ssV_2;
    ssV_2 << "ERROR.";
    throw FlxException("FlxRSM_SmpPoints_rnd::get_nextPoint", ssV_2.str() ); 
  }
  y = *centerY;
  for (tuint i = 0; i < y.get_N(); ++i) {
    y[i] += rv_normal() * k;
  }
}


const tnlong calc_PC_projection_numerical_MCI::calc_expectation(flxVec& IntegT, FlxFunction* fun, const tnlong& N_Interv, const tnlong& N_mciI_1, const tnlong& N_mciI_2, const tdouble& pRedistribute, const tdouble Bound)
{
  GlobalVar.slog(4) << "rsm_set pc_project: estimate the moments using Monte Carlo integration ..." << std::endl;
  #if FLX_DEBUG
    if (IntegT.get_N() != PC.get_SeqNumb() ) {
      throw FlxException_Crude("calc_PC_projection_numerical_MCI::calc_expectation");
    }
  #endif
  P = IntegT.get_N();
  funR = fun;
  // remember x and y
    flxVec orig_X(RndBoxN.get_NOX());
    RndBoxN.get_x_Vec(orig_X.get_tmp_vptr());
    flxVec orig_Y(RndBoxN.get_NRV());
    RndBoxN.get_y_Vec(orig_Y.get_tmp_vptr());
     
  flxpVec Ivec(P*N_Interv);
  tnlong i,j;
  lb = new FunNumber(-ONE);
  ub = new FunNumber(ONE);
  uD = new RBRV_entry_RV_uniform("p",0,new FlxFunction(lb),new FlxFunction(ub),false);
    
  const tdouble step=Bound/N_Interv;
  const tdouble &h_inv = Bound;
  
  std::valarray<tnlong> N_Vec(N_mciI_1,N_Interv);
  if (N_Interv==1) {
    N_Vec[0] += N_mciI_2;
    do_IntegrationLoop(Ivec,step,N_Interv,N_Vec);
  } else {
    // estimate the integral
    GlobalVar.slog(4) << "  estimation step ..." << std::endl;
    do_IntegrationLoop(Ivec,step,N_Interv,N_Vec);
    
    // perform a better integration based on the estimate (for each term in the PC)   
    std::valarray<tnlong> N_Vec_sub(N_Interv);
    tdouble t;
    for (i=0; i<P; ++i) {
      GlobalVar.slog(4) << "  term " << i << " (out of " << P << ") ..." << std::endl;
      t = ZERO;
      for (j=0; j<N_Interv; ++j) {
	t+=sqrt(fabs( (Ivec[j*P+i]).cast2double() )/N_Vec[j]);	// evaluate the estimated ith integral value (square)
      }
      // redistribution
      tnlong max = 0;
      tnlong maxI = 0;
      tnlong count = 0;
      for (j=0;j<N_Interv;++j) {
	N_Vec_sub[j] = tnlong(std::floor((sqrt(fabs((Ivec[j*P+i]).cast2double())/N_Vec[j])/t)*N_mciI_2));
	if (N_Vec_sub[j]>max) {
	  maxI = j;
	  max = N_Vec_sub[j];
	}
	count+=N_Vec_sub[j];
      }
      N_Vec_sub[maxI] += N_mciI_2 - count;
      // redistribute (diffusion)
	count = 0;
	tnlong rd = tnlong(N_Vec_sub[0]*pRedistribute/2);
	N_Vec_sub[0]-=rd;
	count +=N_Vec_sub[0];
	for (j=1;j<N_Interv-1;++j) {
	  N_Vec_sub[j]+=rd;
	  rd = tnlong((N_Vec_sub[j]-rd)*pRedistribute/2);
	  N_Vec_sub[j]-=2*rd;
	  N_Vec_sub[j-1]+=rd;
	  count += N_Vec_sub[j]+rd;
	}
	N_Vec_sub[N_Interv-1]+=rd;
	rd = tnlong((N_Vec_sub[N_Interv-1]-rd)*pRedistribute/2);
	N_Vec_sub[N_Interv-1]-=rd;
	N_Vec_sub[N_Interv-2]+=rd;
	count += N_Vec_sub[N_Interv-1]+rd;
	N_Vec_sub[maxI] += N_mciI_2 - count;
      // take old results into account
	for (j=0;j<N_Interv;++j) {
	  if (N_Vec[j]>=N_Vec_sub[j]) N_Vec_sub[j] = 0;
	  else N_Vec_sub[j]-=N_Vec[j];
	}
      // evaluate the integral (estimate based on previous estimate)
      do_IntegrationLoop(Ivec,step,N_Interv,N_Vec_sub);
      N_Vec+=N_Vec_sub;
    }
  }

  // sum of everything
    j=0;
    for (i=0;i<N_Interv;++i) {
      j+=N_Vec[i];
    }
    const tdouble jm = tdouble(j)/tdouble(N_Interv);
    flxpVec Integ(P);
    for (i=0;i<N_Interv;++i) {
      flxpVec IvecP(&(Ivec[i*P]),P);
      Integ.add(IvecP,jm/N_Vec[i]);
    }
    IntegT = Integ;
    IntegT *= h_inv / j;
  
  // restore x and y
    RndBoxN.set_smp_y(orig_Y);
    RndBoxN.set_smp_x(orig_X);
  // free memory
    funR = NULL;
    delete uD;
    GlobalVar.slog(4) << "  samples required for the integration: " << GlobalVar.Double2String(j) << std::endl;
    return j;
}

void calc_PC_projection_numerical_MCI::do_IntegrationLoop( flxpVec& Ivec, const tdouble& step, const tnlong& N_Interv, const std::valarray< tnlong >& N_Vec)
{
  tdouble x_u=ZERO, x_o=step;
  tnlong N_mciI_T, N_mciI_T_I;
  flxpVec IntegI(P), IntegI2(P);
  register tnlong i,j,m;
  for (i=0;i<N_Interv;++i) {
    // change properties of uniform distribution
      lb->set_thenumber(x_u);
      ub->set_thenumber(x_o);
    IntegI.set_zero();
    N_mciI_T = N_Vec[i];
    if (N_mciI_T>=1e4) {
      N_mciI_T_I = tnlong(floor(sqrt(tdouble(N_mciI_T))));
      for (int l=1;l<=2;++l) {
	for (j=(l==2)?N_mciI_T_I-1:1;j<N_mciI_T_I;++j) {
	  IntegI2.set_zero();
	  for (m=0;m<N_mciI_T_I;++m) {
	    Integrationstep(IntegI2);
	  }
	  IntegI+=IntegI2;
	}
	N_mciI_T_I = N_mciI_T - pow2(N_mciI_T_I) + N_mciI_T_I;
      }
    } else {
      for (j=0;j<N_mciI_T;++j) {
	Integrationstep(IntegI);
      }
    }
    flxpVec Ivec_tmp(&(Ivec[i*P]),P);
    Ivec_tmp+=IntegI;
    x_u = x_o;
    x_o += step;
  }
}

void calc_PC_projection_numerical_MCI::Integrationstep(flxpVec& IntegI)
{
  // generate y-vector
    flxVec &y = tempVec;
    RndBoxN.propose_y(y);

  // transform to z-space
    const tdouble ly2 = y.get_Norm2_NOroot();
    const tdouble lz2 = uD->Inv_cdf_x(rv_cdf_ChiSquare(DIM,ly2));
    flxVec &z = tempVec;
    z*= sqrt(lz2/ly2);
  
  // calculate f
    const tdouble f = rv_pdf_ChiSquare(DIM,lz2);
    
  // update standard normal space
    RndBoxN.set_smp(z);
    
  // return function * foverh;
    const tdouble cv = funR->calc() * f;
    for (register tnlong i=0;i<P;++i) {
      IntegI[i]+=cv*PC.calc_Poly(i,z);      
    }
}
