/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#pragma once

#include "flxobjects.h"
#include "ssfem_kernel.h"


// ------------------------------------------------------------------------------------------------------------------

class FlxRSM_Fun_base : public FlxReaderBase2, public FlxDataBase {
  protected:
    const std::string setStr;
    RBRV_constructor RndBox;
    const tuint DIM;
    flxVec* paraVec;
    flxVec* tempVec; 
    flxVec* xVec;    
  public:
    FlxRSM_Fun_base(const std::string setStr, flxVec* paraVec=NULL);
    virtual ~FlxRSM_Fun_base();
    virtual FlxRSM_Fun_base* copy() = 0;
    virtual const tuint get_NumbOfTerms() const = 0;
    virtual void calc_termVec(const flxVec& xVec, flxVec& termVec) = 0;
    void set_paraVec(flxVec* paraVecV);
    void set_para(const tuint index, const tdouble paraV);
    const tdouble calc_solution();
    const tdouble get_para(const tuint index);
    const flxVec &get_paraVec() { return *paraVec; }
    RBRV_constructor& get_RndBox() { return RndBox; }
    
    static FlxRSM_Fun_base* createRSMfun(const std::string setStr, bool errSerious=true, const std::string ftype="" );
};

class FlxRSM_Fun_p1 : public FlxRSM_Fun_base {
  public:
    FlxRSM_Fun_p1(const std::string setStr, flxVec* paraVec=NULL) : FlxRSM_Fun_base(setStr,paraVec) {}
    FlxRSM_Fun_base* copy();
    virtual const tuint get_NumbOfTerms() const { return DIM+1; }
    virtual void calc_termVec(const flxVec& xVec, flxVec& termVec);
};

class FlxRSM_Fun_p2womt : public FlxRSM_Fun_base {
  public:
    FlxRSM_Fun_p2womt(const std::string setStr, flxVec* paraVec=NULL) : FlxRSM_Fun_base(setStr,paraVec) {}
    FlxRSM_Fun_base* copy();
    virtual const tuint get_NumbOfTerms() const { return 2*DIM+1; }
    virtual void calc_termVec(const flxVec& xVec, flxVec& termVec);
};

class FlxRSM_Fun_p2wmt : public FlxRSM_Fun_p2womt {
  public:
    FlxRSM_Fun_p2wmt(const std::string setStr, flxVec* paraVec=NULL) : FlxRSM_Fun_p2womt(setStr,paraVec) {}
    FlxRSM_Fun_base* copy();
    virtual const tuint get_NumbOfTerms() const { return 2*DIM+1+(DIM*DIM-DIM)/2; }
    virtual void calc_termVec(const flxVec& xVec, flxVec& termVec);
};


// -------------------------- PC-Response Surface ----------------------------------------------------------------------

class FlxRSM_Fun_pc : public FlxRSM_Fun_base {
  private:
    /**
    * @brief calss for generating the polynomial chaos
    */
    PolyChaos pc;
  public:
    FlxRSM_Fun_pc(const std::string setStr, const int order, flxVec* paraVec=NULL) : FlxRSM_Fun_base(setStr,paraVec), pc(DIM,order) {}
    FlxRSM_Fun_base* copy();
    virtual const tuint get_NumbOfTerms() const { return pc.get_SeqNumb(); }
    void calc_termVec(const flxVec& xVec, flxVec& termVec);
    
    const int get_order() { return pc.get_order(); }
    const PolyChaos &pc_reference() { return pc; }
};


/**
* @brief calculates the expectation value of a function numerically
* ... for functions depending on more than one variable
*/
class calc_PC_projection_numerical_MCI : public FlxBoxBaseR {
  private:
    /**
    * @brief The function to calculate the expectation from
    */
    FlxFunction* funR;
    /**
    * @brief The polynomial chaos to work with
    */
    const PolyChaos& PC;
    /**
    * @brief Dimension of the problem
    */
    tuint DIM;
    flxVec tempVec;
    /**
    * @brief number of terms in the polynomial chaos
    */
    tnlong P;
    /**
    * @brief uniform distribution
    */
    FunNumber* lb;
    FunNumber* ub;
    RBRV_entry_RV_uniform* uD;
    RBRV_constructor& RndBoxN;
    /**
    * @brief one MCI-Integrationstep
    */
    inline void Integrationstep(flxpVec &IntegI);
    /**
    * @brief integrates once
    * @param step the stepsize
    * @param N_Interv number of intervals to split the sampling
    */
    void do_IntegrationLoop(flxpVec& Ivec, const tdouble& step, const tnlong& N_Interv, const std::valarray<tnlong> &N_Vec);
  public:
    calc_PC_projection_numerical_MCI(const PolyChaos& PC, RBRV_constructor& RndBoxN) : PC(PC), DIM(PC.get_dim()), tempVec(DIM), RndBoxN(RndBoxN) {}
    /**
    * @brief calculates the expectation of fun * PC_terms
    * @param IntegT returns the expectations
    * @param fun The function to calculate the expectation from
    * @param pc the polynomial chaos to project the solution onto
    * @param N_Interv number of intervals to split the sampling
    * @param N_mciI_1 number of MCI-samples per interval (for the first round)
    * @param N_mciI_2 number of MCI-samples in general (for the second round) - adaptive sampling
    * @return number of samples
    */
    const tnlong calc_expectation(flxVec& IntegT, FlxFunction* fun, const tnlong& N_Interv, const tnlong& N_mciI_1, const tnlong& N_mciI_2, const tdouble& pRedistribute, const tdouble Bound);
};


// ---------------------------------------------------------------------------------------------------------------------

class FlxCombination {
  private:
    tdouble start;
    tdouble end;
    tdouble step;
    const tuint dim;
    flxVec cv;
    bool runs;
  public:
    FlxCombination(const tdouble start, const tdouble end, const tdouble step, const tuint dim);
    const flxVec& Next();
};

class FlxRSM_SmpPoints_base : public FlxReaderBase2, public FlxDataBase {
  protected:
    tuint DIM;
  public:
    FlxRSM_SmpPoints_base() : DIM( 0 ) {}
    virtual ~FlxRSM_SmpPoints_base() {};
    virtual const tuint get_NumbOfPoints() = 0;
    virtual void get_nextPoint(flxVec& y) = 0;
    virtual void updateStart(const tuint M) { DIM = M; };
    
    static FlxRSM_SmpPoints_base* createRSMpts(bool errSerious=true);
};

class FlxRSM_SmpPoints_joa : public FlxRSM_SmpPoints_base {
  protected:
    tuint index;
    FlxMtxConstFun* centerYF;
    FlxFunction* kF;
    flxVec* centerY;
    tdouble k;
  public:
    FlxRSM_SmpPoints_joa(const bool errSerious=true);
    virtual ~FlxRSM_SmpPoints_joa();
    virtual void updateStart(const tuint M);
    virtual const tuint get_NumbOfPoints() { return 2*DIM+1; }
    virtual void get_nextPoint(flxVec& y);
};

class FlxRSM_SmpPoints_cc : public FlxRSM_SmpPoints_joa {
  private:
    FlxFunction* kF2;
    tdouble k2;
    FlxCombination* Comb;
  public:
    FlxRSM_SmpPoints_cc(const bool errSerious=true);
    virtual ~FlxRSM_SmpPoints_cc();
    virtual void updateStart(const tuint M);
    virtual const tuint get_NumbOfPoints() { return 2*DIM+1+pow_b2(DIM); }
    virtual void get_nextPoint(flxVec& y);
};

class FlxRSM_SmpPoints_ff : public FlxRSM_SmpPoints_joa {
  private:
    FlxCombination* Comb;
    bool sphere;
  public:
    FlxRSM_SmpPoints_ff(const bool errSerious=true);
    virtual ~FlxRSM_SmpPoints_ff();
    virtual void updateStart(const tuint M);
    virtual const tuint get_NumbOfPoints() { return pow_int(3,DIM); }
    virtual void get_nextPoint(flxVec& y);
};

class FlxRSM_SmpPoints_p1 : public FlxRSM_SmpPoints_joa {
  public:
    FlxRSM_SmpPoints_p1(const bool errSerious=true) : FlxRSM_SmpPoints_joa(errSerious) { };
    virtual const tuint get_NumbOfPoints() { return DIM+1; }
    virtual void get_nextPoint(flxVec& y);
};

class FlxRSM_SmpPoints_rnd : public FlxRSM_SmpPoints_joa {
  private:
    FlxFunction* numb;
  public:
    FlxRSM_SmpPoints_rnd(const bool errSerious=true);
    ~FlxRSM_SmpPoints_rnd() {delete numb;}
    virtual const tuint get_NumbOfPoints() { return numb->cast2tuint(); }
    void get_nextPoint(flxVec& y);
};

class FlxRSM_Solve_base : public FlxReaderBase2, public FlxDataBase {
  public:
    virtual ~FlxRSM_Solve_base() {};
    virtual void solve(const FlxMtx& A, const flxVec& b, flxVec& sol) = 0;
    static FlxRSM_Solve_base* createRSMsolve(bool errSerious=true);
};

// ---------------------------------------------------------------------------------------------------------------------


class FlxRSM_Solve_CG : public FlxRSM_Solve_base {
  private:
    FlxFunction* iterF;
    FlxFunction* epsF;
    FlxFunction* pcnF;
    bool print;
    bool printS;
  public:
    FlxRSM_Solve_CG(const bool errSerious=true);
    ~FlxRSM_Solve_CG();
    void solve(const FlxMtx& A, const flxVec& b, flxVec& sol);
};

// ---------------------------------------------------------------------------------------------------------------------

/**
* @brief object class: Response Surface
*
* RSM (g; RSMfunction; SmpPoints; SolMeth) { RSM_CODE};
*/
class FlxObjRSM : public FlxObjBase {
  protected:
    const std::string rsmID;
    /**
    * @brief the approximation function
    */
    FlxRSM_Fun_base* RSMfunction;
    /**
    * @brief the sample points (or better the rule for sampling points)
    */
    FlxRSM_SmpPoints_base* SmpPoints;
    /**
    * @brief the solution method
    */
    FlxRSM_Solve_base* solMeth;
    /**
    * @brief the function to approximate
    */
    FlxFunction* fung;
    
    void task();
  public:
    FlxObjRSM ( bool dolog, const std::string rsmID, FlxRSM_Fun_base* RSMfunction, FlxRSM_SmpPoints_base* SmpPoints, FlxRSM_Solve_base* solMeth, FlxFunction* fung ) 
      : FlxObjBase(dolog), rsmID(rsmID), RSMfunction(RSMfunction), SmpPoints(SmpPoints),solMeth(solMeth),fung(fung) {}
    virtual ~FlxObjRSM();
};

/**
* @brief object read class: for FlxObjRSM
*/
class FlxObjReadRSM : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

// ---------------------------------------------------------------------------------------------------------------------

/**
* @brief object class: set a response surface function
*
* RSM_set fun RSM_function_TYPE;
*/
class FlxObjRSM_set_fun : public FlxObjBase {
  protected:
    const std::string rsmID;
    /**
    * @brief the approximation function
    */
    FlxRSM_Fun_base* RSMfunction;
    
    void task();
  public:
    FlxObjRSM_set_fun ( const bool dolog, const std::string& rsmID, FlxRSM_Fun_base* RSMfunction ) : FlxObjBase(dolog), rsmID(rsmID), RSMfunction(RSMfunction) {}
    virtual ~FlxObjRSM_set_fun();
};

/**
* @brief object read class: for FlxObjRSM
*/
class FlxObjReadRSM_set_fun : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

// ---------------------------------------------------------------------------------------------------------------------


/**
* @brief object class: set the coefficients of a response surface function
*/
class FlxObjRSM_set_para : public FlxObjBase {
  protected:
    FlxFunction* index;
    const std::string rsmID;
    FlxFunction* value;
    void task();
  public:
    FlxObjRSM_set_para( const bool dolog, FlxFunction* index, const std::string &rsmID, FlxFunction* value ) : FlxObjBase(dolog), index(index), rsmID(rsmID), value(value) {}
    virtual ~FlxObjRSM_set_para();
};

/**
* @brief object read class: for FlxObjRSM
*/
class FlxObjReadRSM_set_para : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

// ---------------------------------------------------------------------------------------------------------------------

/**
* @brief object class
*/
class FlxObjRSM_print_coeff : public FlxObjOutputBase {
  protected:
    const std::string rsmID;
    void task();
  public:
    FlxObjRSM_print_coeff ( const bool dolog, const std::string& ostreamV, const bool verboseV, const std::string& rsmID) : FlxObjOutputBase(dolog,ostreamV,verboseV), rsmID(rsmID) {}
};

/**
* @brief object read class: for FlxObjRSM
*/
class FlxObjReadRSM_print_coeff : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

// ---------------------------------------------------------------------------------------------------------------------

/**
* @brief object class
*/
class FlxObjRSM_coeff_vec : public FlxObjBase {
  protected:
    FlxString* vname;
    const std::string rsmID;
    void task();
  public:
    FlxObjRSM_coeff_vec ( bool dolog, FlxString* vname, const std::string& rsmID) : FlxObjBase(dolog), vname(vname), rsmID(rsmID) {}
    ~FlxObjRSM_coeff_vec() { delete vname; }
};

/**
* @brief object read class: for FlxObjRSM
*/
class FlxObjReadRSM_coeff_vec : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

// ---------------------------------------------------------------------------------------------------------------------

/**
* @brief object class: set a polynomial chaos response surface function (using the projection technique)
*
* RSM_set para index=value;
*/
class FlxObjRSM_pc_project : public FlxObjBase {
  protected:
    const std::string rsmID;
    FlxRSM_Fun_base* RSMfunction;
    FlxFunction* rv_fun;
    FlxFunction* intervalsF;
    FlxFunction* smpPintervF;
    FlxFunction* smpRedistF;
    FlxFunction* diffusionF;
    FlxFunction* Bound;
    void task();
  public:
    FlxObjRSM_pc_project ( const bool dolog, const std::string& rsmID, FlxRSM_Fun_base* RSMfunction, FlxFunction* rv_fun, FlxFunction* intervalsF, FlxFunction* smpPintervF, FlxFunction* smpRedistF, FlxFunction* diffusionF, FlxFunction* Bound ) : FlxObjBase(dolog), rsmID(rsmID), RSMfunction(RSMfunction), rv_fun(rv_fun), intervalsF(intervalsF), smpPintervF(smpPintervF), smpRedistF(smpRedistF), diffusionF(diffusionF), Bound(Bound) {}
    virtual ~FlxObjRSM_pc_project();
};


/**
* @brief object read class: for FlxObjRSM
*/
class FlxObjReadRSM_pc_project : public FlxObjReadOutputBase {
  public:
    FlxObjReadRSM_pc_project();
    FlxObjBase* read ();
};

// ---------------------------------------------------------------------------------------------------------------------

class FunRSM : public FunBaseFun_multPara, public FlxDataBase {
  private:
    const std::string rsmID;
  public:
    FunRSM(const std::string& rsmID, std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV), rsmID(rsmID) {};
    const tdouble calc();
    const std::string write() { return "rsm(" + rsmID + ")"; }
    const std::string write_v() { return "rsm"; }
    const bool dependOn_Const(tdouble *thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

class FunReadFunRSM : public FunReadFunBase {
  public:
    FunBase* read( bool errSerious ) { return new FunRSM( reader->getWord(true,false), read_parameters(0,errSerious) );}
};

// ---------------------------------------------------------------------------------------------------------------------

class FunRSM_terms : public FunBaseFun_multPara, public FlxDataBase {
  private:
    const std::string rsmID;
  public:
    FunRSM_terms(const std::string& rsmID, std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV), rsmID(rsmID) {};
    const tdouble calc();
    const std::string write() { return "rsm_terms(" + rsmID + ")"; }
    const std::string write_v() { return "rsm_terms"; }
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

class FunReadFunRSM_terms : public FunReadFunBase {
    FunBase* read( bool errSerious ) { return new FunRSM_terms( reader->getWord(true,false), read_parameters(0,errSerious) );}
};


// ---------------------------------------------------------------------------------------------------------------------

class flxRSMmap {
  private:
    std::map<std::string,FlxRSM_Fun_base*> box;
  public:
    ~flxRSMmap();
    void set(const std::string& ID,FlxRSM_Fun_base* rsmp);
    FlxRSM_Fun_base& get(const std::string& ID);
};



