/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */


#define VERSION_REV 22
#define VERSION_DATE "2017-11-08"

#include "flxstat.h"
#include "flxfunction_ope_calc.h"
#include "flxrbrv_rvs_read.h"
#include <fstream>

#if defined(_MSC_VER)
  #include "flxlibstat_export.h"
#else
  #define FLXLIBSTAT_EXPORT
#endif

#ifdef __WINDOWS__
  #pragma comment(lib, "flxlib")
#endif

flxlibs_interface flx_iface;
FlxData* flxdata_ref = NULL;
flxStatBox_container StatBox_cnt;
flx_cdf_container CDFbox;

#if FLX_DEBUG
int flxStatBox_container::Cinst = 0;
#endif


extern "C" FLXLIBSTAT_EXPORT void flxlib_init( FlxObjectReadBox* objReadBox, FlxData* dataBox ) {
  GlobalVar.slog(4) << "  Stat-library - Version " << VERSION_REV << " (" << VERSION_DATE << ")" << std::endl;
  flx_iface.set_boxes(objReadBox,&(dataBox->FunBox));
  // objects
    flx_iface.add_obj("stat_box_new", new FlxObjReadStat_Box_new()); 
    flx_iface.add_obj("stat_box_add", new FlxObjReadStat_Box_add()); 
    flx_iface.add_obj("stat_box_vec_mean", new FlxObjReadStat_Box_vec_mean()); 
    flx_iface.add_obj("stat_box_confi", new FlxObjReadStat_Box_confi());  
    flx_iface.add_obj("stat_cdf_new", new FlxObjReadStat_CDF_new()); 
    flx_iface.add_obj("stat_cdf_clear", new FlxObjReadStat_CDF_clear()); 
    flx_iface.add_obj("stat_cdf_append", new FlxObjReadStat_CDF_append()); 
    flx_iface.add_obj("stat_cdf_average", new FlxObjReadStat_CDF_average()); 
    flx_iface.add_obj("stat_cdf_plot", new FlxObjReadStat_CDF_plot()); 
    flx_iface.add_obj("stat_cdf_qq_plot", new FlxObjReadStat_CDF_qq_plot()); 
    flx_iface.add_obj("stat_cdf_save", new FlxObjReadStat_CDF_save()); 
    flx_iface.add_obj("stat_cdf_load", new FlxObjReadStat_CDF_load()); 
  // functions
    flxdata_ref = dataBox;
    flx_iface.add_fun( "stat_cdf",new FunReadStat_cdf(false) );
    flx_iface.add_fun( "stat_cdf_credible",new FunReadStat_cdf_credible() );
    flx_iface.add_fun( "stat_cdf_inv",new FunReadStat_cdf(true) );
}

extern "C" FLXLIBSTAT_EXPORT void flxlib_close() {
  flx_iface.free_memory();
}

// ------------------------------------------------------------------------------------------------

flxStatBox_container::flxStatBox_container()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'flxStatBox_container' created ...";
      throw FlxException("flxStatBox_container::flxStatBox_container", ssV.str() );
    }
  #endif
}

flxStatBox_container::~flxStatBox_container()
{
  for (std::map<std::string, FlxStatBox*>::iterator pos = box.begin(); pos != box.end(); ++pos) {
    delete pos->second;
  }
}

void flxStatBox_container::insert(const std::string name, const tuint N, const tuint M)
{
  std::map<std::string, FlxStatBox*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    FlxStatBox* sb = pos->second;
    if (sb->N==N && sb->M==M) {
      sb->clear();
    } else {
      delete sb;
      pos->second = new FlxStatBox(N,M);
    }
  } else {
    FlxStatBox* valuep = new FlxStatBox(N,M);
    std::pair<std::string, FlxStatBox*> Element(name, valuep);
    box.insert(Element);
  }
}

FlxStatBox& flxStatBox_container::get(const std::string& name)
{
  std::map<std::string, FlxStatBox*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    return (*pos->second);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "The stat-box '" << name << "' does not exist.";
    throw FlxException("flxStatBox_container::get_1", ssV_2.str(), "You have to define the container before you can use it."); 
  }
}


// ------------------------------------------------------------------------------------------------

FlxObjStat_Box_new::~FlxObjStat_Box_new()
{
  delete fname;
  delete M;
  delete N;
}

void FlxObjStat_Box_new::task()
{
  const tuint Ni = N->cast2tuint(false);
  const tuint Mi = M->cast2tuint(false);
  const std::string name = fname->eval_word(true);
  StatBox_cnt.insert(name,Ni,Mi);
  GlobalVar.slog(4) << "stat: defined new container '" << name << "' with N=" << Ni << " and M=" << Mi << std::endl;
}

FlxObjBase* FlxObjReadStat_Box_new::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxFunction* N = NULL;
  FlxFunction* M = NULL;
  try {
    reader->getChar('(');
    N = new FlxFunction(funReader,false);
    reader->getChar(',');
    M = new FlxFunction(funReader,false);
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjStat_Box_new(get_doLog(),fname,N,M);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_Box_new::read",1);
    delete fname;
    if (N) delete N;
    if (M) delete M;
    throw;
  }
}

// ------------------------------------------------------------------------------------------------

FlxObjStat_Box_add::~FlxObjStat_Box_add()
{
  delete fname;
  delete vecName;
}

void FlxObjStat_Box_add::task()
{
  const std::string name = fname->eval_word(true);
  FlxStatBox& sb = StatBox_cnt.get(name);
  tuint N = 0;
  tdouble* tvp = data->ConstMtxBox.get_Vec(vecName->eval_word(true),N);
  flxVec tv(tvp,N);
  sb.add(tv);
}

FlxObjBase* FlxObjReadStat_Box_add::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxString* vecName = NULL;
  try {
    reader->getChar('(');
    vecName = new FlxString(false,false);
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjStat_Box_add(get_doLog(),fname,vecName);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_Box_add::read",1);
    delete fname;
    if (vecName) delete vecName;
    throw;
  }
}

// ------------------------------------------------------------------------------------------------

void FlxObjStat_Box_vec_mean::task()
{
  const std::string name = fname->eval_word(true);
  FlxStatBox& sb = StatBox_cnt.get(name);
  tuint Nc = sb.M;
  tdouble* tvp = data->ConstMtxBox.get_Vec(vecName->eval_word(true),Nc);
  flxVec tv(tvp,Nc);
  sb.compute_mean(tv);
}

FlxObjStat_Box_vec_mean::~FlxObjStat_Box_vec_mean()
{
  delete fname;
  delete vecName;
}

FlxObjBase* FlxObjReadStat_Box_vec_mean::read()
{
  FlxString* vecName = new FlxString(false,false);
  FlxString* fname = NULL;
  try {
    reader->getChar('=');
    fname = new FlxString(false,false);
    read_optionalPara(false);
    return new FlxObjStat_Box_vec_mean(get_doLog(),fname,vecName);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_Box_vec_mean::read",1);
    delete vecName;
    if (fname) delete fname;
    throw;
  }
}

// ------------------------------------------------------------------------------------------------

void FlxObjStat_Box_confi::task()
{
  const std::string name = fname->eval_word(true);
  FlxStatBox& sb = StatBox_cnt.get(name);
  tuint N = 0;
  flxVec clevels(data->ConstMtxBox.get_Vec(clf->eval(),N),N);
  if (userColName.empty()) {
    sb.plot_confidence(sout(),clevels,fixW,NULL);
  } else {
    N = 0;
    tdouble* userCol = data->ConstMtxBox.get_Vec(userColName,N);
    flxVec uc(userCol,N);
    sb.plot_confidence(sout(),clevels,fixW,&uc);
  }
}

FlxObjReadStat_Box_confi::FlxObjReadStat_Box_confi()
 : FlxObjReadOutputBase(true) 
{
  AllDefParaBox->insert(new FlxOptionalParaFlxString("","stat_box_confi::usercol",false));
  ParaBox.insert("usercol", "stat_box_confi::usercol" );
}

FlxObjBase* FlxObjReadStat_Box_confi::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxMtxConstFun* clf = NULL;
  try {
    reader->getChar('(');
    clf = new FlxMtxConstFun(true);
    reader->getChar(')');
    read_optionalPara(false);
    FlxString* uc = get_optPara_FlxString("usercol");
    std::string uce("");
    try {
      uce = uc->eval_word(true,true);
      delete uc; uc = NULL;
    } catch (FlxException& e) {
      FLXMSG("FlxObjReadStat_Box_confi::read_1",1);
      delete uc;
      throw;
    }
    return new FlxObjStat_Box_confi(get_doLog(),get_stream(),fname,clf,get_fixW(),uce);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_Box_confi::read_2",1);
    delete fname;
    if (clf) delete clf;
    throw;
  }  
}

// -------------------------------------------------------------------------------------------------------------------------------------------------------------------

flx_cdf_container::~flx_cdf_container()
{
  for (std::map<std::string,flx_cdf*>::iterator pos = box.begin(); pos != box.end(); ++pos) {
    delete pos->second;
  }
}

void flx_cdf_container::insert(const std::string& name, const size_t Nreserve)
{
  std::map<std::string, flx_cdf*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    throw FlxException("flx_cdf_container::insert", "A data-CDF with name '" + name + "' is already declared.");
  } else {
    flx_cdf* valuep = new flx_cdf(Nreserve);
    std::pair<std::string, flx_cdf*> Element(name, valuep);
    box.insert(Element);
  }
}

flx_cdf& flx_cdf_container::get(const std::string& name)
{
  std::map<std::string, flx_cdf*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    return (*pos->second);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "The data-CDF '" << name << "' does not exist.";
    throw FlxException("flx_cdf_container::get_1", ssV_2.str(), "You have to define the data-CDF before you can use it."); 
  }
}

flx_cdf::flx_cdf(const size_t Nreserve)
: N(ZERO), count_up2date(false), true_count(true)
{
  list.reserve(Nreserve);
  countList.reserve(Nreserve);
}

void flx_cdf::clear()
{
  N = ZERO;
  count_up2date = false;
  true_count = true;
  list.clear();
  countList.clear();
}

void flx_cdf::append(flxVec& dV)
{
  size_t i = 0;		// counter variable
  for (size_t j = 0; j<dV.get_N(); ++j) {	// loop over each number to append to the data-CDF
    const float dVe = dV[j];
    while ( (i<list.size()) && (list[i].value<dVe) ) {		// jump to position in list that has relevant value
      ++i;
    };
    ++N;
    if ( (i<list.size()) && (list[i].value == dVe) ) {	// increase counter of list[i]
      list[i].occ += 1;
    } else {				// insert new entry before list[i]
      flx_cdf_entry e = { float(dVe), 1 };
      list.insert(list.begin()+i,e);
      countList.push_back(0);		// dummy insert in order to maintain list.size()==countList.size()
    }
  };
  count_up2date = false;
}

void flx_cdf::plot_help(const tdouble q, const tdouble count, tdouble& p, tdouble& p_l, tdouble& p_u)
{
  if (!true_count) {
    p = count/N;
    p_l = ZERO;
    p_u = ONE;
    return;
  }
  const tdouble H = count+ONE/2;
  const tdouble K = N+ONE;
  const tdouble& alpha = H;
  const tdouble beta = K-H;
  p = H/K;
  p_l = iBeta_reg_inv(alpha,beta,q);
  p_u = iBeta_reg_inv(alpha,beta,ONE-q);
}

void flx_cdf::plot(const tdouble q, std::ostream& sout, const bool checkTOL, const int prec, int fixW)
{
  tdouble count = ZERO;
  tdouble px,px_l,px_u,qpx=ONE,qpx_l,qpx_u;
  plot_help(q,N-count,qpx,qpx_l,qpx_u);
  plot_help(q,count,px,px_l,px_u);
  size_t Nlist = list.size();
  for (size_t i=0;i<Nlist;++i) {
    sout << GlobalVar.Double2String(list[i].value,checkTOL,prec,fixW) << ' ' 
      << GlobalVar.Double2String(px,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(px_l,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(px_u,checkTOL,prec,fixW) << ' ' 
      << GlobalVar.Double2String(qpx,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(qpx_l,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(qpx_u,checkTOL,prec,fixW) << ' ' 
      << std::endl;
    plot_help(q,N-count,qpx,qpx_l,qpx_u);
    count += list[i].occ;
    plot_help(q,count,px,px_l,px_u);
    sout << GlobalVar.Double2String(list[i].value,checkTOL,prec,fixW) << ' ' 
      << GlobalVar.Double2String(px,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(px_l,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(px_u,checkTOL,prec,fixW) << ' ' 
      << GlobalVar.Double2String(qpx,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(qpx_l,checkTOL,prec,fixW) << ' ' << GlobalVar.Double2String(qpx_u,checkTOL,prec,fixW) << ' ' 
      << std::endl;
  };
}

void flx_cdf::qq_plot(RBRV_entry_RV_base* rep, std::ostream& sout)
{
  tdouble count = ZERO;
  size_t Nlist = list.size();
  for (size_t i=0;i<Nlist;++i) {
    const tdouble d = list[i].value;
    const tdouble p = (count+list[i].occ);    
    const tdouble y = (p<N/2)?(rv_InvPhi(p/N)):(-rv_InvPhi((N-p)/N));
      const tdouble x = rep->transform_y2x(y);
      sout << GlobalVar.Double2String(d) << ' ' << GlobalVar.Double2String(x) << std::endl; 
    count += list[i].occ;
  }
}

void flx_cdf::save(const std::string& fileS)
{
  std::ofstream file(fileS.c_str(), std::ios::out | std::ios::binary);
  if (file.is_open()==false) {
    throw FlxException("flx_cdf::save", "File '" + fileS + "' could not be opened.");
  }
  file.write((char*)&true_count,sizeof(bool));
  file.write((char*)&(list[0]),sizeof(flx_cdf_entry)*list.size());
  file.close();
}

void flx_cdf::load(const std::string& fileS)
{
  if (N>0) throw FlxException("flx_cdf::load_01", "A stored data-CDF can only be loaded to an empty data-CDF.");
  std::ifstream file(fileS.c_str(), std::ios::in | std::ios::binary | std::ios::ate );
  if (file.is_open()==false || file.good()==false) {
    throw FlxException("flx_cdf::load_02", "File '" + fileS + "' could not be opened.");
  }
  size_t size = (size_t)file.tellg() - sizeof(bool);
  size_t sizeN = size/sizeof(flx_cdf_entry);
  if ( size != sizeN*sizeof(flx_cdf_entry) ) {
    throw FlxException("flx_cdf::load_03", "The file '" + fileS + "' seems to be either corrupt or does not contain a data-CDF.");
  }
  list.resize(sizeN);
  countList.resize(sizeN);
  file.seekg (0, std::ios::beg);
  file.read((char*)&true_count,sizeof(bool));
  file.read((char*)&(list[0]),size);
  file.close();
  // count N
    const size_t Nlist = list.size();
    for (size_t i=0;i<Nlist;++i) {
      N += list[i].occ;
      countList[i] = N;
    };
  count_up2date = true;
}

void flx_cdf::average(tuint A, FlxFunction* transform)
{
  if (A<=1) return;
  // normalize occ's
    {
      size_t Nlist = list.size();
      for (size_t i=0;i<Nlist;++i) {
	list[i].occ /= N;
      }
      N = ONE;
      true_count = false;
    }
  // remember initial list
    std::vector<flx_cdf_entry> list_1 = list;
  // recursive call
    average_help(A,transform,list_1);
  // divide by A
    size_t Nlist = list.size();
    tdouble c = 0;
    for (size_t i=0;i<Nlist;++i) {
      list[i].value /= A;
      c += list[i].occ;
    }
}

void flx_cdf::average_help(tuint A, FlxFunction* transform, const std::vector<flx_cdf_entry>& list_1)
{
  if (A>2) {
    if (A%2) {
      --A;
      average_help(A,transform,list_1);
      std::vector<flx_cdf_entry> list_a = list;
      average_help_add(list_a,list_1,transform);
    } else {
      A /= 2;
      average_help(A,transform,list_1);      
      std::vector<flx_cdf_entry> list_a = list;
      average_help_add(list_a,list_a,transform);
    }
  } else {
    if (A<2) throw FlxException_Crude("flx_cdf::average_help");
    average_help_add(list_1,list_1,transform);
  }  
}

void flx_cdf::average_help_add(const std::vector< flx_cdf_entry >& list_a, const std::vector< flx_cdf_entry >& list_b, FlxFunction* transform)
{
  N = 0; count_up2date=false;
  list.clear();
  countList.clear();
  size_t Nlista = list_a.size();
  for (size_t i=0;i<Nlista;++i) {
    average_help_add2(list_a[i], list_b,transform);
  }
}

void flx_cdf::average_help_add2(const flx_cdf_entry& entry_a, const std::vector< flx_cdf_entry >& list_b, FlxFunction* transform)
{
  size_t i = 0;		// counter variable
  size_t Nlistb = list_b.size();
  tdouble &x_ref = flxdata_ref->ConstantBox.getRef("x");
  const tdouble x_old = x_ref;
  for (size_t j = 0; j<Nlistb; ++j) {	// loop over each number to append to the data-CDF
    x_ref = list_b[j].value+entry_a.value;
    const float dVe = transform->calc();
    const flxcdfn dvN = list_b[j].occ*entry_a.occ;
    while ( (i<list.size()) && (list[i].value<dVe) ) {		// jump to position in list that has relevant value
      ++i;
    };
    N += dvN;
    if ( (i<list.size()) && (list[i].value == dVe) ) {	// increase counter of list[i]
      list[i].occ += dvN;
    } else {				// insert new entry before list[i]
      flx_cdf_entry e = { float(dVe), dvN };
      list.insert(list.begin()+i,e);
      countList.push_back(0);		// dummy insert in order to maintain list.size()==countList.size()
    }
  };
  x_ref = x_old;
  count_up2date = false;
}


void flx_cdf::update_count()
{
  if (count_up2date) return;
  const size_t Nlist = list.size();
  tdouble Nh = 0;
  for (size_t i=0;i<Nlist;++i) {
    Nh += list[i].occ;
    countList[i] = Nh;
  };
  if (fabs(Nh-N)/N>1e-3) { throw FlxException_Crude("flx_cdf::update_count"); };
  N = Nh;
  count_up2date = true;
}

const tdouble flx_cdf::hits_le_val(const tdouble val)
{
  // ensure that data-CDF is properly initialized
    if ( N==0 ) {
      throw FlxException("flx_cdf::hits_le_val","The data-CDF does not contain data values.");
    }
    update_count();
  // finds the largest index 'id' in 'list' whose value is smaller or equal than val
    const size_t Nlist = list.size();	// number of unique entries in data-CDF
    if (val<list[0].value) return 0;
    if (val>=list[Nlist-1].value) return N;
    size_t id = 0;
    size_t length = Nlist;
    while (length>1) {
      const size_t check = id + (length/2);
      if (val<list[check].value) {
	length = check-id;
      } else if (val==list[check].value) {
	id = check;
	length = 1;
      } else {
	id = check;
	length = length-(length/2);
      }
    };
  return countList[id];
}


const tdouble flx_cdf::cdf(const tdouble val)
{
  if (!true_count) return hits_le_val(val);
  const tdouble H = hits_le_val(val) + 0.5;
  const tdouble K = N + ONE;
  return H/K;
}

const tdouble flx_cdf::cdf_credible(const tdouble val, const tdouble q)
{
  if (!true_count) return ZERO;
  // parameters of beta distribution
    const tdouble H = hits_le_val(val) + 0.5;
    const tdouble K = N + ONE;
    const tdouble alpha = H;
    const tdouble beta = K-H;
  return iBeta_reg_inv(alpha,beta,q);
}

const tdouble flx_cdf::cdf_inv(const tdouble p)
{
  update_count();
  const size_t Nlist = list.size();
  // check bounds
    if (N==0) {
      throw FlxException_NeglectInInteractive("flx_cdf::cdf_inv_01", "The inverse CDF is not defined, as the data-CDF is empty." );
    }
    if (p<ZERO||p>ONE) {
      std::ostringstream ssV;
      ssV << "Value '" << GlobalVar.Double2String(p) << "' is not within the valid bounds [0;1]";
      throw FlxException_NeglectInInteractive("flx_cdf::cdf_inv_02", ssV.str() );
    }
    if (true_count) {
      if (p >= (N+ONE/2)/(N+ONE)) return list[Nlist-1].value;
    } else {
      if (p >= ONE) return list[Nlist-1].value;
    }
  // find value that belongs to 'p'
      size_t id = 0;
      size_t length = Nlist;
      while (length>1) {
	const size_t check = id + (length/2);
	const tdouble pc = true_count?((countList[check]+ONE/2)/(N+ONE)):(countList[check]/N);
	if (p<pc) {
	  length = check-id;
	} else if (fabs(p-pc)<=GlobalVar.TOL()) {
	  id = check;
	  length = 1;
	} else {
	  id = check;
	  length = length-(length/2);
	}
      };
  return list[id].value;
}

void FlxObjStat_CDF_new::task()
{
  const size_t Nreserve = Nreservef->cast2tulong(false);
  const std::string name = fname->eval_word(true);
  CDFbox.insert(name,Nreserve);
  GlobalVar.slog(4) << "stat: defined new data-CDF '" << name << "' with capacity for " << Nreserve << " entries." << std::endl;
}

FlxObjReadStat_CDF_new::FlxObjReadStat_CDF_new()
: FlxObjReadBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaFun(1e7,"stat_cdf_new::nreserve"));
  ParaBox.insert("nreserve", "stat_cdf_new::nreserve" );
}

FlxObjBase* FlxObjReadStat_CDF_new::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxFunction* Nreserve = NULL;
  try {
    read_optionalPara(false);
    Nreserve = get_optPara_FlxFunction("nreserve");
    return new FlxObjStat_CDF_new(get_doLog(),fname,Nreserve);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_CDF_new::read",1);
    delete fname;
    if (Nreserve) delete Nreserve;
    throw;
  }
}

void FlxObjStat_CDF_clear::task()
{
  // get data-CDF
    const std::string cdf_name = fname->eval_word(true);
    flx_cdf& dcdf = CDFbox.get(cdf_name);
  dcdf.clear();
}

FlxObjBase* FlxObjReadStat_CDF_clear::read()
{
  FlxString* fname = new FlxString(false,false);
  try {
    read_optionalPara(false);
    return new FlxObjStat_CDF_clear(get_doLog(),fname);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_CDF_clear::read",1);
    delete fname;
    throw;
  }
}

void FlxObjStat_CDF_append::task()
{
  // get data-CDF
    const std::string cdf_name = fname->eval_word(true);
    flx_cdf& dcdf = CDFbox.get(cdf_name);
  if (istrmf) {
    // get input stream
      const std::string isn = istrmf->eval_word(true);
      FlxIstream &is = data->IstreamBox.get(isn);
    // bulk-size
      size_t bN = bulkf->cast2tuint(false);
    flxVec dV(bN);
    tdoublePtr dP = dV.get_tmp_vptr();
    tuint lastIndex = 0;
    bool b;
    tdouble &x_ref = data->ConstantBox.getRef("x");
    const tdouble x_old = x_ref;
    do {
      b = is.get_vec(dV,lastIndex,true);
      // transform the vector
	for (size_t i=0;i<lastIndex;++i) {
	  x_ref = dP[i];
	  dP[i] = transform->calc();
	};
      std::sort(dP,dP+lastIndex);
      flxVec tV(dP,lastIndex);
      dcdf.append(tV);
    } while (b);
    x_ref = x_old;
  } else {
    if (mtxf==NULL) throw FlxException_Crude("FlxObjStat_CDF_append::task");
    tuint Nr = 0;
    tuint Nc = 0;
    tdoublePtr dPh = data->ConstMtxBox.get_Mtx(mtxf->eval(),Nr,Nc);
    flxVec dV(dPh,Nr*Nc,true);
    tdoublePtr dP = dV.get_tmp_vptr();
    // transform the vector
      tdouble &x_ref = data->ConstantBox.getRef("x");
      const tdouble x_old = x_ref;
      for (size_t i=0;i<Nr*Nc;++i) {
	x_ref = dP[i];
	dP[i] = transform->calc();
      }
      x_ref = x_old;
    std::sort(dP,dP+Nr*Nc);
    dcdf.append(dV);
  }
}

FlxObjReadStat_CDF_append::FlxObjReadStat_CDF_append()
: FlxObjReadBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaFun(1e7,"stat_cdf_append::bulk"));
  ParaBox.insert("bulk", "stat_cdf_append::bulk" );
  AllDefParaBox->insert(new FlxOptionalParaFun(new FlxFunction(new FunConst(data->ConstantBox.get("x",true))),"stat_cdf_append::transform"));
  ParaBox.insert("transform", "stat_cdf_append::transform" );
}

FlxObjBase* FlxObjReadStat_CDF_append::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxString* istrmf = NULL;
  FlxMtxConstFun* mtxf = NULL;
  try {
    reader->getChar('(',false);
    if (reader->whatIsNextChar()=='{') {
      reader->getChar('{',false);
      mtxf = new FlxMtxConstFun(true);
      reader->getChar('}',false);
    } else {
      istrmf = new FlxString(false,false);
    }
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjStat_CDF_append(get_doLog(), fname, istrmf, mtxf, get_optPara_FlxFunction("bulk"), get_optPara_FlxFunction("transform"));
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadStat_CDF_append::read",1);
    delete fname;
    if (istrmf) delete istrmf;
    if (mtxf) delete mtxf;
    throw;
  }
}

void FlxObjStat_CDF_average::task()
{
  const tuint A = Afun->cast2tuint();
  // get data-CDF
    const std::string cdf_name = fname->eval_word(true);
    flx_cdf& dcdf = CDFbox.get(cdf_name);
  dcdf.average(A,transform);
}

FlxObjReadStat_CDF_average::FlxObjReadStat_CDF_average()
: FlxObjReadBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaFun(new FlxFunction(new FunConst(data->ConstantBox.get("x",true))),"stat_cdf_average::transform"));
  ParaBox.insert("transform", "stat_cdf_average::transform" );
}

FlxObjBase* FlxObjReadStat_CDF_average::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxFunction* Afun = NULL;
  try {
    reader->getChar('(',false);
    Afun = new FlxFunction(funReader,false);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjStat_CDF_average(get_doLog(), fname, Afun, get_optPara_FlxFunction("transform"));
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadStat_CDF_average::read",1);
    delete fname;
    if (Afun) delete Afun;
    throw;
  }
}

void FlxObjStat_CDF_plot::task()
{
  // get data-CDF
    const std::string cdf_name = fname->eval_word(true);
    flx_cdf& dcdf = CDFbox.get(cdf_name);
  // evaluate the quantile for the credible interval
    const tdouble q = qF->calc();
    if (q<ZERO || q>ONE/2) {
      std::ostringstream ssV;
      ssV << "The specified quantile value (" << GlobalVar.Double2String(q) << ") must be between [0,0.5].";
      throw FlxException("FlxObjStat_CDF_plot::task", ssV.str() );
    }
  dcdf.plot(q,sout(),checkTOL,prec,fixW);
}

FlxObjReadStat_CDF_plot::FlxObjReadStat_CDF_plot() 
: FlxObjReadOutputBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaFun(0.05,"stat_cdf_plot::credible"));
  ParaBox.insert("credible", "stat_cdf_plot::credible" );
}

FlxObjBase* FlxObjReadStat_CDF_plot::read()
{
  FlxString* fname = new FlxString(false,false);
  try {
    read_optionalPara(false);
    return new FlxObjStat_CDF_plot(get_doLog(),get_stream(),fname,get_optPara_FlxFunction("credible"),get_prec(),get_fixW());
  } catch (FlxException &e) {
    delete fname;
    throw;
  }
}

void FlxObjStat_CDF_qq_plot::task()
{
  // get data-CDF
    const std::string cdf_name = fname->eval_word(true);
    flx_cdf& dcdf = CDFbox.get(cdf_name);
  dcdf.qq_plot(rep,sout());
}

FlxObjReadStat_CDF_qq_plot::FlxObjReadStat_CDF_qq_plot()
: FlxObjReadOutputBase(true)
{
  
}

FlxObjBase* FlxObjReadStat_CDF_qq_plot::read()
{
  FlxString* fname = new FlxString(false,false);
  RBRV_entry_RV_base* rep = NULL;
  try {
    reader->getChar('(');
    rep = RBRV_entry_read_base::read_gen_entry(false);
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjStat_CDF_qq_plot(get_doLog(),get_stream(),fname,rep);
  } catch (FlxException &e) {
    delete fname;
    if (rep) delete rep;
    throw;
  }
}

void FlxObjStat_CDF_save::task()
{
  const std::string name = fname->eval_word(true);
  const std::string fileS = fileF->eval(false);
  flx_cdf& dcdf = CDFbox.get(name);
  dcdf.save(fileS);
  GlobalVar.slog(4) << "stat: stored data-CDF '" << name << "' in file '" << fileS << "'." << std::endl;
}

FlxObjBase* FlxObjReadStat_CDF_save::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxString* fileF = NULL;
  try {
    reader->getChar('(');
    fileF = new FlxString(false,false);
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjStat_CDF_save(get_doLog(),fname,fileF);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_CDF_save::read",1);
    delete fname;
    if (fileF) delete fileF;
    throw;
  }
}

void FlxObjStat_CDF_load::task()
{
  const std::string name = fname->eval_word(true);
  const std::string fileS = fileF->eval(false);
  flx_cdf& dcdf = CDFbox.get(name);
  dcdf.load(fileS);
  GlobalVar.slog(4) << "stat: loaded data-CDF '" << name << "' from file '" << fileS << "'." << std::endl;
}

FlxObjBase* FlxObjReadStat_CDF_load::read()
{
  FlxString* fname = new FlxString(false,false);
  FlxString* fileF = NULL;
  try {
    reader->getChar('(');
    fileF = new FlxString(false,false);
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjStat_CDF_load(get_doLog(),fname,fileF);
  } catch (FlxException& e) {
    FLXMSG("FlxObjReadStat_CDF_load::read",1);
    delete fname;
    if (fileF) delete fileF;
    throw;
  }
}

const tdouble FunStat_cdf::calc()
{
  if (dcdf==NULL) {
    const std::string name = fname->eval_word(true);
    dcdf = &(CDFbox.get(name));
  }
  if (isInv) {
    return dcdf->cdf_inv(valF->calc());
  } else {
    return dcdf->cdf(valF->calc());
  }
}

const std::string FunStat_cdf::write()
{
  return "stat_cdf(" + fname->write() + "," + valF->write() + ")";
}

const bool FunStat_cdf::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  child_optimize(valF,foi);
  return false;
}

FunBase* FunReadStat_cdf::read(bool errSerious)
{
  FlxString* fname = new FlxString(false,false);
  FunBase* valF = NULL;
  try {
    reader->getChar(',');
    valF = FunctionList->read(errSerious);
    return new FunStat_cdf(fname,valF,isInv);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadStat_CDF_load::read",1);
    delete fname;
    if (valF) delete valF;
    throw;
  }
}

const tdouble FunStat_cdf_credible::calc()
{
  if (dcdf==NULL) {
    const std::string name = fname->eval_word(true);
    dcdf = &(CDFbox.get(name));
  }
  const tdouble q = qF->calc();
  if (q<ZERO || q>ONE) {
    std::ostringstream ssV;
    ssV << "The specified quantile value (" << GlobalVar.Double2String(q) << ") must be between [0,1].";
    throw FlxException("FunStat_cdf_credible::calc", ssV.str() );
  }
  return dcdf->cdf_credible(valF->calc(),q);
}

const bool FunStat_cdf_credible::search_circref(FlxFunction* fcr)
{
  return valF->search_circref(fcr) || qF->search_circref(fcr);
}

const std::string FunStat_cdf_credible::write()
{
  return "stat_cdf_credible(" + fname->write() + "," + valF->write() + "," + qF->write() + ")";
}

const bool FunStat_cdf_credible::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  child_optimize(valF,foi);
  child_optimize(qF,foi);
  return false;
}

FunBase* FunReadStat_cdf_credible::read(bool errSerious)
{
  FlxString* fname = new FlxString(false,false);
  FunBase* valF = NULL;
  FunBase* qF = NULL;
  try {
    reader->getChar(',');
    valF = FunctionList->read(errSerious);
    reader->getChar(',');
    qF = FunctionList->read(errSerious);
    return new FunStat_cdf_credible(fname,valF,qF);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadStat_CDF_load::read",1);
    delete fname;
    if (valF) delete valF;
    if (qF) delete qF;
    throw;
  }
}



