/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2016 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxfunction_ext.h"
#include "flxfunction_fun.h"
#include "flxfunction_ope.h"
#include "flxfunction_ope_calc.h"



FlxFunPoint::FoR FlxFunPoint::get_FoR(char strlFoR, bool errSerious)
{
  switch (strlFoR) {
    case 'c':
      return cartesian;
    case 'p':
      return cylindrical;
    case 's':
      return spherical;
    default:
      std::ostringstream ssV;
      ssV << "Unkown frame of reference '" << strlFoR << "'.";
      FlxError(errSerious,"FunReadPara::set_NumbOfPara", ssV.str() );
      return cartesian;	// DUMMY-RETURN
  }
}

char FlxFunPoint::get_FoR(FlxFunPoint::FoR elFoR)
{
  switch(elFoR) {
    case cartesian:
      return 'c';
      break;
    case cylindrical:
      return 'p';
      break;
    case spherical:
      return 's';
      break;
    default:
      return ' ';
  }
}

FlxFunPoint::FlxFunPoint(const FlxFunPoint& ref)
: FrORef(ref.FrORef), d1(new FlxFunction(*(ref.d1))), d2(new FlxFunction(*(ref.d2))), d3(new FlxFunction(*(ref.d3)))
{
  
}

FlxFunPoint& FlxFunPoint::operator=(const FlxFunPoint& ref)
{
  if (this != & ref) {
    FrORef = ref.FrORef;
    delete d1; delete d2; delete d3;
    d1 = new FlxFunction(*(ref.d1));
    d2 = new FlxFunction(*(ref.d2));
    d3 = new FlxFunction(*(ref.d3));
  }
  return *this;
}

FlxFunPoint::FlxFunPoint(ReadStream* reader, FlxFunctionReader* funReader, bool errSerious)
:FrORef(cartesian),d1(NULL),d2(NULL),d3(NULL)
{
  char c = reader->getChar();
  // Cartesian frame of reference
    if (c=='[' || c=='c' || c=='C') {
      FrORef = cartesian;
    } 
  // Cylindrical frame of reference
    else if (c=='p' || c=='P') {
      FrORef = cylindrical;
    }
  // Spherical frame of reference
    else if (c=='s' || c=='S') {
      FrORef = spherical;
    }
  // Syntax-Error
    else {
      std::ostringstream ssV;
      ssV << "Expected input of a point and not '" << c << "'.";
      FlxError(errSerious,"FlxFunPoint::FlxFunPoint_1", ssV.str() );
    }
  if (c!='[') reader->getChar('[',errSerious);
  // input coordinate-information
    d1 = new FlxFunction(funReader,errSerious);
    if (reader->whatIsNextChar()==']') {
      d2 = new FlxFunction(new FunNumber(0.0) );
    } else {
      reader->getChar(',');
      d2 = new FlxFunction(funReader,errSerious);
    }
    if (reader->whatIsNextChar()==']') {
      d3 = new FlxFunction(new FunNumber(0.0) );
    } else {
      reader->getChar(',');
      d3 = new FlxFunction(funReader,errSerious);
    }
  // Finish the input of the point
    reader->getChar(']',errSerious);
}

FlxFunPoint::FlxFunPoint(const FlxFunPoint::FoR FrORef, FlxFunction* d1, FlxFunction* d2, FlxFunction* d3)
: FrORef(FrORef), d1(d1), d2(d2), d3(d3) 
{
   
}

FlxFunPoint::~FlxFunPoint()
{
  delete d1; 
  delete d2; 
  delete d3;
}

const flxPoint FlxFunPoint::calc() const
{
  tdouble rr;
  tdouble pp;
  tdouble tt;
  switch (FrORef) {
    case cartesian:
      return flxPoint(d1->calc(), d2->calc(), d3->calc());
    case cylindrical:
      rr = d1->calc();
      pp = d2->calc();
      return flxPoint(rr*cos(pp), rr*sin(pp), d3->calc());
    case spherical:
      rr = d1->calc();
      pp = d2->calc();
      tt = d3->calc();
      return flxPoint(rr*cos(pp)*sin(tt), rr*sin(pp)*sin(tt), rr*cos(tt));      
    default:
      #if FLX_DEBUG
	throw FlxException_Crude("FlxFunPoint::calc");
      #else
	return flxPoint(ZERO,ZERO,ZERO);
      #endif
  }
}

const bool FlxFunPoint::is_direction_zero(const char dir)
{
  switch (dir) {
    case 'x':
      return d1->is_zero();
    case 'y':
      return d2->is_zero();
    case 'z':
      return d3->is_zero();
    default:
      throw FlxException_Crude("FlxFunPoint::is_direction_zero");
  }
}

std::ostream& operator<<(std::ostream& os, const FlxFunPoint& val)
{
  return os << val.get_FoR(val.FrORef) << "[" << val.d1->write() << "," << val.d2->write() << "," << val.d3->write() << "]";
}


const tdouble FlxFun_GaussIntegration(FunBase* funI, tdouble* theconst, const tdouble& start, const tdouble& end, const tuint& GPN, GaussIntegration& GI)
{
  const tdouble theconst_orig = *theconst;	// remember the value of theconst 
  // get number of integration points
    GI.check_GA(GPN);
    const tdouble* GP = GI.get_GP(GPN);
    const tdouble* GW = GI.get_GW(GPN);
  // determine start and end
    const tdouble c1 = (end-start)/2;
    const tdouble c2 = (end+start)/2;
  // perform the integration
    tdouble sum = ZERO;
    for (tuint p=0;p<GPN;++p) {
      *theconst = c1*GP[p] + c2;
      sum += funI->calc() * GW[p];
    }
    sum*=c1;
    #if FLX_DEBUG
      if ( std::isnan(sum) ) throw FlxException_Crude("FlxFun_GaussIntegration");
    #endif
  *theconst = theconst_orig;
  return sum;
}

const tuint FlxFun_EstimateGaussPoint(FunBase* funI, tdouble* theconst, const tdouble& start, const tdouble& end, const tuint& GPmax, const tuint& GPtest, const tdouble& err, GaussIntegration& GI)
{
  #if FLX_DEBUG
    if (GPmax+2 > GPtest) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxFun_EstimateGaussPoint", ssV.str() );
    }
  #endif
  // check if the function is constant
    const tdouble theconst_orig = *theconst;
    *theconst = start; 
      tdouble t = funI->calc();
    *theconst = end; 
      tdouble t2 = funI->calc();
    if (t==t2) {
      *theconst = (start+end)/2; 
	t2 = funI->calc();
      if (t==t2) {
	const tdouble c1 = (end-start)/2;
	const tdouble c2 = (end+start)/2;
	*theconst = c2+c1/sqrt(3.); 
	  t2 = funI->calc();
	if (t==t2) {
	  *theconst = c2-c1/sqrt(3.); 
	    t2 = funI->calc();
	  if (t==t2) {
	    *theconst = theconst_orig;
	    return 0;
	  }
	}
      }
    }
    *theconst = theconst_orig;
  
  // compute the reference value
    const tdouble refRes = FlxFun_GaussIntegration(funI,theconst,start,end,GPtest,GI);
  
  // do the actual iteration
    for (tuint i=1;i<=GPmax;++i) {
      t = FlxFun_GaussIntegration(funI,theconst,start,end,i,GI);
      if (fabs((t-refRes)/refRes)<=err) {
	t = FlxFun_GaussIntegration(funI,theconst,start,end,i+1,GI);
	if (fabs((t-refRes)/refRes)<=err) {
	  return i;
	}
      }
    }
  return GPmax+1;  
}


const tdouble FlxFun_RootSearch_Bisec(FunBase* funR, tdouble* theconst, tdouble start, tdouble end, const tdouble dx, const tdouble dy, ostreamp streamp)
{
  const int fixW = 15;		// output-width for logging
  if (streamp) {
    *streamp << std::endl;
    *streamp << "*************************************" << std::endl;
    *streamp << "* root-search: bisection method     *" << std::endl;
    *streamp << "*************************************" << std::endl;
    *streamp << " dx = " << GlobalVar.Double2String(dx) << std::endl;
    *streamp << " dy = " << GlobalVar.Double2String(dy) << std::endl;    
  }
  // check the intervals
    if (end<=start) {
      std::ostringstream ssV;
      ssV << "Error with root-search boundaries.";
      throw FlxException("FlxFun_RootSearch_Bisec_1", ssV.str() );
    }
    if (dx<ZERO) {
      std::ostringstream ssV;
      ssV << "'dx' threshold must not be negativ";
      throw FlxException("FlxFun_RootSearch_Bisec_2", ssV.str() );
    }
    if (dy<ZERO) {
      std::ostringstream ssV;
      ssV << "'dy' threshold must not be negativ";
      throw FlxException("FlxFun_RootSearch_Bisec_3", ssV.str() );
    }
    if (dx < GlobalVar.TOL() && dy < GlobalVar.TOL()) {
      GlobalVar.alert.alert("FlxFun_RootSearch_Bisec_4","The 'dx' and 'dy' thresholds might be too restrictive.");
    }
  const tdouble storedF = *theconst;
  bool found = false;
  // start - configuration
    tdouble res = ZERO;  // dummy initialization to prevent compiler warning
    *theconst = start;
    tdouble f1 = funR->calc();
    *theconst = end;
    tdouble f2 = funR->calc();
    if (fabs(f1)<=GlobalVar.TOL()) {
      res = start;
      found = true;
    } else if (fabs(f2)<=GlobalVar.TOL()) {
      res = end;
      found = true;
    } else if (f2*f1>=ZERO) {
      std::ostringstream ssV;
      ssV << "Function has the same sign at both boundaries.";
      throw FlxException("FlxFun_RootSearch_Bisec_5", ssV.str() );
    }
  if (streamp) {
    *streamp << " f(" << GlobalVar.Double2String(start,false,-1,fixW) << ") = " << GlobalVar.Double2String(f1,false,-1,fixW) << std::endl;
    *streamp << " f(" << GlobalVar.Double2String(end,false,-1,fixW) << ") = "   << GlobalVar.Double2String(f2,false,-1,fixW) << std::endl;
  }
  if (!found) {
    while (end-start>dx) {
      res = (start+end)/2;
      *theconst = res;
      const tdouble f3 = funR->calc();
      if (streamp) {
	*streamp << " f(" << GlobalVar.Double2String(res,false,-1,fixW) << ") = " << GlobalVar.Double2String(f3,false,-1,fixW) << std::endl;
      }
      if (fabs(f3)<=dy) {
	found = true;
	break;
      }
      if ( f1*f3>ZERO ) {
	start = res;
	f1 = f3;
      } else {
	end = res;
	f2 = f3;
      }
    }
  }
  if (!found) {
    res = (start+end)/2;
  }
  if (streamp) {
    if (!found) {
      *streamp << " ALERT: root could not be localized for the specified dy-error-bound." << std::endl;
    }
    *streamp << " result: " << GlobalVar.Double2String(res,false,-1,fixW) << std::endl;
    *streamp << "*************************************" << std::endl;
  }
  *theconst = storedF;
  return res;
}

const tdouble FlxFun_RootSearch_RegulaFalsi(FunBase* funR, tdouble* theconst, tdouble start, tdouble end, const tdouble dx, const tdouble dy, ostreamp streamp)
{
  const int fixW = 15;		// output-width for logging
  if (streamp) {
    *streamp << std::endl;
    *streamp << "*************************************" << std::endl;
    *streamp << "* root-search: regula falsi         *" << std::endl;
    *streamp << "*************************************" << std::endl;
    *streamp << " dx = " << GlobalVar.Double2String(dx) << std::endl;
    *streamp << " dy = " << GlobalVar.Double2String(dy) << std::endl;   
  }
  const tuint NMAX = tuint(1e4);
  // check the intervals
    if (end<=start) {
      std::ostringstream ssV;
      ssV << "Error with root-search boundaries.";
      throw FlxException("FlxFun_RootSearch_RegulaFalsi_1", ssV.str() );
    }
    if (dx<ZERO) {
      std::ostringstream ssV;
      ssV << "'dx' threshold must not be negativ";
      throw FlxException("FlxFun_RootSearch_RegulaFalsi_2", ssV.str() );
    }
    if (dy<ZERO) {
      std::ostringstream ssV;
      ssV << "'dy' threshold must not be negativ";
      throw FlxException("FlxFun_RootSearch_RegulaFalsi_3", ssV.str() );
    }
    if (dx < GlobalVar.TOL() && dy < GlobalVar.TOL()) {
      GlobalVar.alert.alert("FlxFun_RootSearch_RegulaFalsi_4","The 'dx' and 'dy' thresholds might be too restrictive.");
    }
  const tdouble storedF = *theconst;
  bool found = false;
  // start - configuration
    tdouble res = ZERO;
    *theconst = start;
    tdouble f1 = funR->calc();
    *theconst = end;
    tdouble f2 = funR->calc();
    if (streamp) {
      *streamp << " f(" << GlobalVar.Double2String(start,false,-1,fixW) << ") = " << GlobalVar.Double2String(f1,false,-1,fixW) << std::endl;
      *streamp << " f(" << GlobalVar.Double2String(end,false,-1,fixW) << ") = "   << GlobalVar.Double2String(f2,false,-1,fixW) << std::endl;
    }
    if (fabs(f1)<=GlobalVar.TOL()) {
      res = start;
      found = true;
    } else if (fabs(f2)<=GlobalVar.TOL()) {
      res = end;
      found = true;
    }
  if (!found) {
    tuint c = 0;
    while (fabs(end-start)>dx && c<NMAX) {
      ++c;
      tdouble f3;
      if (f1*f2<=ZERO) {	// Pegasus-algorithm
	res = (start*f2-end*f1)/(f2-f1);
	*theconst = res;
	f3 = funR->calc();
	if (streamp) {
	  *streamp << " f(" << GlobalVar.Double2String(res,false,-1,fixW) << ") = " << GlobalVar.Double2String(f3,false,-1,fixW) << " (Pegasus-algorithm)" << std::endl;
	}
	if ( f2*f3<ZERO ) {
	  start = end;
	  f1 = f2;
	  end = res;
	  f2 = f3;
	} else {
	  const tdouble m = f2/(f2+f3);
	  f1 = m*f1;
	  end = res;
	  f2 = f3;
	}
      } else {			// Sekantenverfahren
	res = end - ((end-start)/(f2-f1))*f2;
	*theconst = res;
	f3 = funR->calc();
	if (streamp) {
	  *streamp << " f(" << GlobalVar.Double2String(res,false,-1,fixW) << ") = " << GlobalVar.Double2String(f3,false,-1,fixW) << " (Secant-algorithm)" << std::endl;
	}
	f1 = f2;
	start = end;
	f2 = f3;
	end = res;
      }
      if (fabs(f3)<=dy) {
	found = true;
	break;
      }
      #if FLX_DEBUG
	if (std::isnan(start) || std::isnan(end) ) throw FlxException_Crude("FlxFun_RootSearch_RegulaFalsi_5");
      #endif
    }
    if (c>=NMAX) {
      res = FlxFun_RootSearch_Bisec(funR,theconst,start,end,dx,dy,streamp);
      found = true;
    }
  }
  if (!found) {
    res = (start+end)/2;
  }
  if (streamp) {
    if (!found) {
      *streamp << " ALERT: root could not be localized for the specified error bounds." << std::endl;
    }
    *streamp << " result: " << GlobalVar.Double2String(res,false,-1,fixW) << std::endl;
    *streamp << "*************************************" << std::endl;
  }
  #if FLX_DEBUG
    if (std::isnan(res)) {
      throw FlxException_Crude("FlxFun_RootSearch_RegulaFalsi_6");
    }
  #endif
  *theconst = storedF;
  return res;
}

const tdouble FlxFun_Optimize1D(FunBase* funR, tdouble* theconst, tdouble start, tdouble end, const tdouble tau, const tuint nex)
{
  // check the intervals
    if (end<=start) {
      std::ostringstream ssV;
      ssV << "Error with optimization boundaries.";
      throw FlxException("FlxFun_Optimize1D_1", ssV.str() );
    }
    if (tau<ZERO) {
      std::ostringstream ssV;
      ssV << "'tau' threshold must not be negativ";
      throw FlxException("FlxFun_Optimize1D_2", ssV.str() );
    }
    if (tau < GlobalVar.TOL()) {
      GlobalVar.alert.alert("FlxFun_Optimize1D_3","The 'tau' threshold might be too restrictive.");
    }
  // store value of theconst
    const tdouble storedF = *theconst;
  // initialize variables
    const tdouble phi = (ONE + sqrt(tdouble(5.0))) / (2*ONE);
    const tdouble resphi = 2*ONE - phi;
    const tdouble phih = (2*ONE)/(sqrt(tdouble(5.0))-ONE);
    tdouble a = start;
    tdouble b = start + resphi*(end-start);
    tdouble c = end;
    *theconst = a;
      tdouble fa = funR->calc();
    *theconst = b;
      tdouble fb = funR->calc();
    *theconst = c;
      tdouble fc = funR->calc();
  // preliminary search (Brent's exponential search http://linneus20.ethz.ch:8080/1_5_2.html)
    if (fa > fb && fb > fc) {
      for (tuint i=0;(i<nex)&&(fa > fb && fb > fc);++i) {
	const tdouble x = c + phih*(c-b);
	a = b;
	b = c;
	c = x;
	*theconst = c;
	fa = fb;
	fb = fc;
	fc = funR->calc();
      }
      if (fa > fb && fb > fc) {
	std::ostringstream ssV;
	ssV << "Maximum number of interations reached.";
	throw FlxException("FlxFun_Optimize1D_4", ssV.str() );
      }
    } else if (fa < fb && fb < fc) {
      for (tuint i=0;(i<nex)&&(fa < fb && fb < fc);++i) {
	const tdouble x = a - phih*(b-a);
	c = b;
	b = a;
	a = x;
	*theconst = a;
	fc = fb;
	fb = fa;
	fa = funR->calc();
      }
      if (fa < fb && fb < fc) {
	std::ostringstream ssV;
	ssV << "Maximum number of interations reached.";
	throw FlxException("FlxFun_Optimize1D_5", ssV.str() );
      }
    } else if (fa < fb && fb > fc) {
      std::ostringstream ssV;
      ssV << "There appears to be a maximum a not a minimum within the chosen interval.";
      throw FlxException("FlxFun_Optimize1D_6", ssV.str() );
    } else if (fa == fb && fb == fc) {
      std::ostringstream ssV;
      ssV << "Is it possible that the function is constant on the chosen interval?";
      throw FlxException("FlxFun_Optimize1D_7", ssV.str() );
    }
  // actual algorithm
    double x,fx;
    do {
      if (c-b>b-a) {
	x = b + resphi*(c-b);
      } else {
	x = b - resphi*(b-a);
      }
      if (fabs(c-a)<tau*(fabs(b)+fabs(x))) {
	*theconst = storedF;
	return (c+a)/(2*ONE);
      }
      *theconst = x;
      fx = funR->calc();
      if (fx < fb) {
	if ( c-b > b-a) {
	  a = b; fa = fb;
	  b = x; fb = fx;
	} else {
	  c = b; fc = fb;
	  b = x; fb = fx;
	}
      } else {
	if ( c-b > b-a) {
	  c = x; fc = fx;
	} else {
	  a = x; fa = fx;
	}
      }
    } while (true);
}


