/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2016 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxmath.h"

#include <boost/math/special_functions.hpp>
#include <boost/math/special_functions/gamma.hpp>


/**
* @brief Array Size of FactorialLn-Function (stores calculated values)
*/
const int FactorialLnArraySize = 100;
/**
* @brief Array Size of Factorial-Function (stores calculated values up to !(FactorialArraySize+1) )
*/
const int FactorialArraySize = 33;
/**
* @brief Array Size of DoubleFactorial-Function (stores calculated values up to !(DoubleFactorialArraySize+1) )
*/
const int DoubleFactorialArraySize = 33;


const tdouble GammaLn(const tdouble &xx)
{
  static const tdouble cof[6]={76.18009172947146,-86.50532032941677,24.01409824083091,-1.231739572450155,0.1208650973866179e-2,-0.5395239384953e-5};
  tdouble y=xx;
  const tdouble &x=xx;
  tdouble tmp=x+5.5;
  tmp-=(x+0.5)*std::log(tmp);
  tdouble ser=1.000000000190015;
  for (int j=0;j<6;j++) ser+=cof[j]/++y;
  return -tmp+std::log(2.5066282746310005*ser/x);
}
// {
//   return boost::math::lgamma(xx);
// }

const tdouble flxgamma(const tdouble x)
{
  return boost::math::tgamma(x);
}

const tdouble flxgamma(const tdouble a, const tdouble z)
{
  return boost::math::tgamma(a,z);
}

const tdouble flxgamma_l(const tdouble a, const tdouble z)
{
  return boost::math::tgamma_lower(a,z);
}

const tdouble flxgamma_ru(const tdouble a, const tdouble z)
{
  return boost::math::gamma_q(a,z);
}

const tdouble flxgamma_rl(const tdouble a, const tdouble z)
{
  return boost::math::gamma_p(a,z);
}

const tdouble flxgamma_ru_inv(const tdouble a, const tdouble q)
{
  return boost::math::gamma_q_inv(a,q);
}

const tdouble flxgamma_rl_inv(const tdouble a, const tdouble p)
{
  return boost::math::gamma_p_inv(a,p);
}

const tdouble flxerf_inv(const tdouble p)
{
  return boost::math::erf_inv(p);
}

const tdouble iBeta_reg(const tdouble alpha, const tdouble beta, const tdouble x)
{
  return boost::math::ibeta(alpha,beta,x);
}

const tdouble iBeta_regc(const tdouble alpha, const tdouble beta, const tdouble x)
{
  return boost::math::ibetac(alpha,beta,x);
}

const tdouble iBeta_reg_inv(const tdouble alpha, const tdouble beta, const tdouble p)
{
  return boost::math::ibeta_inv(alpha,beta,p);
}

const tdouble iBetac_reg_inv(const tdouble alpha, const tdouble beta, const tdouble p)
{
  return boost::math::ibetac_inv(alpha,beta,p);
}

const tdouble flxdigamma(const tdouble x)
{
  return boost::math::digamma(x);
}

const tdouble Factorial(const int n) {
  static int ntop=4;
  static tdouble a[FactorialArraySize]={2.0,6.0,24.0};
  if (n<=1) return ONE;
  else if (n>FactorialArraySize+1) return exp(GammaLn(tdouble(n)+ONE));
  while (ntop<n) {
    const int j=ntop++;
    a[ntop-2]=a[j-2]*ntop;
  }
  return a[n-2];
}

const tdouble DoubleFactorial(const int n)
{
  static int ntop=5;
  static tdouble a[FactorialArraySize]={2.0,3.0,8.0,15.0};
  if (n<=1) return ONE;
  else if (n>FactorialArraySize+1) {
    if (n%2 == 0) {
      const int k = n/2;
      return pow(2.,k)*Factorial(k);			// TODO efficiency of pow?
    } else {
      const int k = (n+1)/2;
      return Factorial(2*k)/(pow(2.,k)*Factorial(k));	// TODO efficiency of pow?
    }
  }
  else {
    while (ntop<n) {
      int j=ntop++;
      a[ntop-2]=a[j-3]*ntop;
    }
    return a[n-2];
  }
}

const tdouble FactorialLn(const int n) {
  static tdouble a[FactorialLnArraySize];
  if (n<=1) return ZERO; 
  if (n<=FactorialLnArraySize+1) return (a[n-2]!=ZERO?a[n-2]:(a[n-2]=GammaLn(tdouble(n)+ONE)));
  else return GammaLn(tdouble(n)+ONE);
}

