/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2016 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxmath_H
#define fesslix_flxmath_H

#include "flxglobal.h"

#include <cmath>


/**
* @brief throw away all numbers before the dot
*/
inline const tdouble frac(const tdouble &x) {  // digits after the dot
  if ( x >= ZERO ) return x - std::floor(x);
  return std::ceil(x) - x;
}
inline const tdouble arccot(const tdouble &x) {
  return PI * ONE/2 - std::atan(x);
}
inline const tdouble arctanh(const tdouble &x) {
  return (ONE/2) * std::log((ONE + x) / (ONE - x));
}

inline const tdouble sign(const tdouble &x) {
  return (x>0)?ONE:((x<ZERO)?-ONE:ZERO);
}

template <typename T>
inline T pow_int(T x, const tuint n)
{
  switch(n) {
    case 0:
      return 1;
    case 1:
      return x;
    case 2:
      return x*x;
    case 3:
      return x*x*x;
    case 4:
    {
      T v = x*x;
      return v*v;
    }
    case 5:
    {
      T v = x*x;
      return v*v*x;
    }
    default:
    {
      T v = T(1);
      for (tuint i=0;i<n;++i) v *= x;
      return v;
    }
  };
}

template <typename T>
inline T pow2(T x)
{
  return x*x;
}

template <typename T>
inline T pow_b2(T n)
{
  return 1<<n;
}

/**
* @brief calculate the Logarithm of the Gamma FunctionList
* @brief returns the value ln(GAMMA(xx)] for xx > 0
*/
FLXLIB_EXPORT const tdouble GammaLn(const tdouble &xx);

inline const tdouble BetaFun(const tdouble &z, const tdouble &w) {
  return std::exp(GammaLn(z)+GammaLn(w)-GammaLn(z+w));
}

/**
* @brief error function
*/
inline const tdouble flxerf(const tdouble x) { return erf(x); }
/**
* @brief inverse error function
*/
FLXLIB_EXPORT const tdouble flxerf_inv(const tdouble p);

/**
* @brief gamma function
*/
FLXLIB_EXPORT const tdouble flxgamma(const tdouble x);
/**
* @brief incomplete upper gamma function
*/
FLXLIB_EXPORT const tdouble flxgamma(const tdouble a, const tdouble z);
/**
* @brief incomplete lower gamma function
*/
FLXLIB_EXPORT const tdouble flxgamma_l(const tdouble a, const tdouble z);
/**
* @brief regularized incomplete upper gamma function
*/
FLXLIB_EXPORT const tdouble flxgamma_ru(const tdouble a, const tdouble z);
/**
* @brief regularized incomplete lower gamma function
*/
FLXLIB_EXPORT const tdouble flxgamma_rl(const tdouble a, const tdouble z);
/**
* @brief inverse regularized incomplete upper gamma function
*/
FLXLIB_EXPORT const tdouble flxgamma_ru_inv(const tdouble a, const tdouble q);
/**
* @brief inverse regularized incomplete lower gamma function
*/
FLXLIB_EXPORT const tdouble flxgamma_rl_inv(const tdouble a, const tdouble p);
/**
* @brief regularized incomplete Beta function
*/
FLXLIB_EXPORT const tdouble iBeta_reg(const tdouble alpha, const tdouble beta, const tdouble x);
/**
* @brief complement of regularized incomplete Beta function
*/
FLXLIB_EXPORT const tdouble iBeta_regc(const tdouble alpha, const tdouble beta, const tdouble x);
/**
* @brief inverse of the regularized incomplete Beta function
*/
FLXLIB_EXPORT const tdouble iBeta_reg_inv(const tdouble alpha, const tdouble beta, const tdouble p);
/**
* @brief inverse of the complement of the regularized incomplete Beta function
*/
FLXLIB_EXPORT const tdouble iBetac_reg_inv(const tdouble alpha, const tdouble beta, const tdouble p);
/**
* @brief Digamma function
*/
FLXLIB_EXPORT const tdouble flxdigamma(const tdouble x);


/**
* @brief Factorial Function
*/
FLXLIB_EXPORT const tdouble Factorial(const int n);
FLXLIB_EXPORT const tdouble DoubleFactorial(const int n);
FLXLIB_EXPORT const tdouble FactorialLn(const int n);

/**
* @brief calculate the log of the Binomial Coefficient (n, m)
*/
inline const tdouble BinomCoeff_ln(const int n, const int m) {
  return GammaLn(n+1)-GammaLn(m+1)-GammaLn(n-m+1);
}

/**
* @brief calculate the Binomial Coefficient (n, m)
*/
inline const tdouble BinomCoeff(const int n, const int m)
{
  //return std::floor(ONE/2+std::exp(FactorialLn(n)-FactorialLn(m)-FactorialLn(n-m)));
  return exp(BinomCoeff_ln(n,m));
}


inline const double round_flx(const double r) {
  return (r > ZERO) ? std::floor(r + ONE/2) : std::ceil(r - ONE/2);
}

inline const double round_flx(double r, const tuint n) {
  const double p = pow(tdouble(10),n);
  r *= p;
  return ((r > ZERO) ? std::floor(r + ONE/2) : std::ceil(r - ONE/2))/p;
}


#endif 	// fesslix_flxmath_H
