/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2015 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxpeval.h"



FlxPeval_term::FlxPeval_term(int M, int p, int degree_of_p) : M(M)
{
  #if FLX_DEBUG
    if (M <= 0) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR: M <= 0";
      throw FlxException("FlxPeval_term::FlxPeval_term_1", ssV_2.str() );
    }
    if (p < 1 || p > M) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR: p";
      throw FlxException("FlxPeval_term::FlxPeval_term_2", ssV_2.str() );
    }
  #endif
  arr = new int[M];
  for (int i=0; i < M; ++i) arr[i] = 0;
  arr[p-1] = degree_of_p;
}

FlxPeval_term::FlxPeval_term(const FlxPeval_term& lhs, const FlxPeval_term& rhs) 
: M(lhs.M), arr(new int[M])
{
  #if FLX_DEBUG
    if (lhs.M != rhs.M) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR";
      throw FlxException("FlxPeval_term::FlxPeval_term_3", ssV_2.str() );
    }
  #endif
  for (register int i=0; i < M; ++i) arr[i] = lhs.arr[i] + rhs.arr[i];
}

FlxPeval_term::FlxPeval_term(const FlxPeval_term& fpt) 
: M(fpt.M), arr(new int[M])
{
  for (register int i=0; i < M; ++i) arr[i] = fpt.arr[i];
}

FlxPeval_term& FlxPeval_term::operator=(const FlxPeval_term& fpt)
{
  if (this != &fpt) {
    if (M!=fpt.M) {
      delete [] arr;
      arr = new int[fpt.M];
    }
    M = fpt.M;
    for (register int i=0; i < M; ++i) arr[i] = fpt.arr[i];
  }
  return *this;
}

#if FLX_DEBUG
void FlxPeval_term::M_check(int rhsM) const
{
  if (M != rhsM) {
    std::ostringstream ssV_2;
    ssV_2 << "ERROR";
    throw FlxException("FlxPeval_term::operator<", ssV_2.str() );
  }
}
#endif

// const bool FlxPeval_term::operator<(const FlxPeval_term& rhs) const
// {
//   #if FLX_DEBUG
//     M_check(rhs.M);
//   #endif
//   for (register int i=0; i < M; ++i) {
//     if (arr[i] == rhs.arr[i]) continue;
//     else if (arr[i] < rhs.arr[i]) return true;
//     else return false;
//   }
//   return false;
// }

const bool FlxPeval_term::operator>(const FlxPeval_term& rhs) const
{
  #if FLX_DEBUG
    M_check(rhs.M);
  #endif
  for (register int i=0; i < M; ++i) {
    if (arr[i] == rhs.arr[i]) continue;
    else if (arr[i] > rhs.arr[i]) return true;
    else return false;
  }
  return false;
}

const bool FlxPeval_term::operator==(const FlxPeval_term& rhs) const
{
  #if FLX_DEBUG
    M_check(rhs.M);
  #endif
  for (int i=0; i < M; ++i) {
    if (arr[i] != rhs.arr[i]) {
      return false;
    }
  }
  return true;
}

tdouble FlxPeval_term::calc_Expectation(const std::vector< FlxRnd_base* >& rndvec) const
{
  #if FLX_DEBUG
    if (rndvec.size() != static_cast<unsigned int>(M)) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR";
      throw FlxException("FlxPeval_term::calc_Expectation", ssV_2.str() );
    }
  #endif
  tdouble E = 1;
  for (int i = 0; i < M; ++i) {
    if (arr[i] != 0) {
      E*=(rndvec[i])->moment(arr[i]);
    }
  }
  return E;  
}

FlxPeval& FlxPeval::add_term(const FlxPeval_term& tp, const tdouble& tv)
{
  std::pair< FlxPeval_term, tdouble > pv(tp, tv);
  return operator+=(pv);
}

FlxPeval& FlxPeval::operator+=(const std::pair< FlxPeval_term, tdouble >& pv)
{
  #if FLX_DEBUG
    if (pv.first.get_dim() != M) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR";
      throw FlxException("FlxPeval::operator+=", ssV_2.str() );
    }
  #endif
  if (pv.second == 0.0) return *this;
//   std::pair<std::map<FlxPeval_term, tdouble>::iterator,bool> par = tl.insert(pv);
//   if (!par.second) {
//     par.first->second += pv.second;
//     if (par.first->second == 0.0) {
//       tl.erase(par.first);
//     }
//   }
  std::map<FlxPeval_term, tdouble>::iterator pos = tl.find(pv.first);
  if ( pos != tl.end() ) {	// already in there
    pos->second += pv.second;
    if (pos->second == 0.0) {
      tl.erase(pos);
    }
  } else {
    tl.insert(pv);
  }
  return *this;
}

FlxPeval& FlxPeval::operator+=(const FlxPeval& rhs)
{
  for (flxtlmap::const_iterator iter = rhs.tl.begin();iter!=rhs.tl.end();++iter) {
    add_term(iter->first,iter->second);
  }
  return *this;
}

FlxPeval& FlxPeval::operator*=(const tdouble& rhs)
{
  if (rhs == 0.0) {
    tl.clear();
  } else {
    for (flxtlmap::iterator iter = tl.begin();iter!=tl.end();++iter) {
      iter->second*=rhs;
    }
  }
  return *this;
}

tdouble FlxPeval::get_coeff(FlxPeval_term& fpt)
{
  std::map<FlxPeval_term, tdouble>::iterator pos = tl.find(fpt);
  if ( pos != tl.end() ) {	
    return pos->second;
  } else {
    return 0.0;
  }
}

FlxPeval::FlxPeval(int M, bool isone) : M(M) 
{
  if (isone) {
    add_term(FlxPeval_term(M),1.0);
  }
}

FlxPeval::FlxPeval(const FlxPeval& lhs, const FlxPeval& rhs) : M(lhs.M)
{
  #if FLX_DEBUG
    if (lhs.M != rhs.M) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR";
      throw FlxException("FlxPeval::FlxPeval_1", ssV_2.str() );
    }
  #endif
  
  for (flxtlmap::const_iterator riter = rhs.tl.begin();riter!=rhs.tl.end();++riter) {
    for (flxtlmap::const_iterator liter = lhs.tl.begin();liter!=lhs.tl.end();++liter) {
      add_term(FlxPeval_term(liter->first,riter->first),liter->second*riter->second);
    }
  }
  
}

tdouble FlxPeval::calc_Expectation(const std::vector< FlxRnd_base* >& rndvec) const
{
  #if FLX_DEBUG
    if (rndvec.size() != static_cast<unsigned int>(M)) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR";
      throw FlxException("FlxPeval_term::calc_Expectation", ssV_2.str() );
    }
  #endif
  tdouble E = 0;
  for (flxtlmap::const_iterator iter = tl.begin();iter!=tl.end();++iter) {
    E+=iter->second * ((iter->first).calc_Expectation(rndvec));
  }
  return E; 
}

std::ostream& operator<<(std::ostream& os, const FlxPeval_term& val)
{
  bool first = true;
  for (int i = 0; i < val.get_dim(); ++i) {
    if ( val.arr[i] != 0) {
      if (first) {
	first = false;
      } else {
	os << "*";
      }
      os << "(" << i+1 << ")^" << val.arr[i];
    }
  }
  if (first) os << "1";
  return os;
}

std::ostream& operator<<(std::ostream& os, const FlxPeval& val)
{
  bool first = true;
  for (FlxPeval::flxtlmap::const_iterator iter = val.tl.begin();iter!=val.tl.end();++iter) {
    if (first) {
      first = false;
    } else {
      if (iter->second >= 0.0) {
	os << "+";
      }
    }
    os << iter->second << "*" << iter->first;
  }
  return os;
}
