/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2015 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef sfem_sfem_H
#define sfem_sfem_H

#include "flxobjfem.h"
#include "sfem_kernel.h"
#include "flxobjrandom.h"

namespace sfem {

class FlxCreateObjReaders_SFEM : public FlxCreateObjReaders {
  public:
    void createObjReaders (FlxObjectReadBox* objReadBox );
    void createFunReaders (FlxData* dataBox );
};






// -------------------------- FlxObjSFEM ----------------------------------------------------------------------

/**
* @brief object class: define a function
*
* sfem new
*/
class FlxObjSFEM_new : public FlxObjOutputBase {
  private:
    int p;
    void task();
  public:
    FlxObjSFEM_new(bool dolog,int p, std::string ostreamV, bool verbose ) :FlxObjOutputBase(dolog,ostreamV,verbose),p(p){};
};


/**
* @brief object read class: for FlxObjSFEM
*/
class FlxObjReadSFEM : public FlxObjReadControlFEM {
  private:
    virtual FlxObjBase* read_new(bool errSerious=false);
    virtual int get_order();
  public:
    FlxObjReadSFEM();
};


// -------------------------- FlxObjSFEMtest ----------------------------------------------------------------------

/**
* @brief object class: SFEM test class
*
* calc FlxFunction
*/
class FlxObjSFEMtest : public FlxObjOutputBase {
  private:
    int M,p;
    void task();
  public:
    FlxObjSFEMtest ( bool dolog, int MV, int pV, std::string ostreamV ) : FlxObjOutputBase(dolog,ostreamV), M(MV), p(pV) {}
    //~FlxObjCalc() { delete fun; }
};

/**
* @brief object read class: for FlxObjSFEMtest
*/
class FlxObjReadSFEMtest : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};


// -------------------------- PC-Response Surface ----------------------------------------------------------------------

class FlxRSM_Fun_pc : public FlxRSM_Fun_base {
  private:
    /**
    * @brief calss for generating the polynomial chaos
    */
    PolyChaos pc;
    FlxFunction* orderF;
  public:
    FlxRSM_Fun_pc(FlxFunction* orderF) : FlxRSM_Fun_base(), pc(1,1), orderF(orderF) {}
    FlxRSM_Fun_pc(const tuint DIM, const int order, flxVec* paraVec) : FlxRSM_Fun_base(DIM,paraVec),pc(DIM,order),orderF(new FlxFunction(new FunNumber(tdouble(order)))) {}
    ~FlxRSM_Fun_pc() { delete orderF; }
    FlxRSM_Fun_base* copy();
    virtual const tuint get_NumbOfTerms() const { return pc.get_SeqNumb(); }
    void calc_termVec(const flxVec& yVec, const flxVec& xVec, flxVec& termVec);
    void updateStart();
    
    const int get_order() { return pc.get_order(); }
    const PolyChaos &pc_reference() { return pc; }
};


// -------------------------- Functions ---------------------------------------------------------------------

class FunSFEM_moment : public FunFEM_base {
  private:
    FunBase* moment;
  public:
    FunSFEM_moment(FunBase* node,FlxDof::dof dofV,FunBase* moment):FunFEM_base(node,dofV),moment(moment){}
    ~FunSFEM_moment() { delete moment;}
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr) {return ( FunFEM_base::search_circref(fcr) || moment->search_circref(fcr));};
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { }
    const bool dependOn_Const(const tdouble* const thenumber);
    const tdouble expectation() { return calc(); }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic read class of FunSFEM_moment
*/
class FunReadSFEM_moment : public FunReadFEM_base {
  public:
    FunBase* read ( bool errSerious );
};


class FunSFEM_covariance : public FunFEM_base {
  private:
    FunBase* node2;
    FunBase* joint2;
    FlxDof::dof dof2;
  public:
    FunSFEM_covariance(FunBase* node,FlxDof::dof dofV,FunBase* node2,FlxDof::dof dof2):FunFEM_base(node,dofV),node2(node2),dof2(dof2){}
    ~FunSFEM_covariance() { delete node2; delete joint2;}
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr) {return ( FunFEM_base::search_circref(fcr) || node2->search_circref(fcr) || joint2->search_circref(fcr) );};
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { }
    const bool dependOn_Const(const tdouble* const thenumber);
    const tdouble expectation() { return calc(); }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic read class of FunReadSFEM_covariance
*/
class FunReadSFEM_covariance : public FunReadFEM_base {
  public:
    FunBase* read ( bool errSerious );
};


class FunSFEM_pc_term : public FunBase {
  private:
    const PolyChaos& pc;
    tulong j;
  public:
    FunSFEM_pc_term(const PolyChaos& pc, tulong j) : pc(pc), j(j) {}
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr) { return false; }
    const std::string write() { return ""; }
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
    const tdouble expectation();
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
    
    void set_j(const tulong jV) { j = jV; }
};



} // end of namespace 'sfem'

#endif // sfem_sfem_H
