/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2015 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "sfem_kernel.h"

using namespace std;
using namespace sfem;


// -------------------------- PolyChaos ----------------------------------------------------------------------

PolyChaos::PolyChaos(const int MV, const int pV):qS(NULL),H(NULL),ExpSq(NULL)
{
  GenerateSequence(MV,pV);
}

PolyChaos::~PolyChaos()
{
  delMem();
}

int* PolyChaos::ExtrSeq(const int n, int *balls) const
{
  int *a = new int[M];
  a[0]=balls[0];
  for (int i = 1; i<M-1;i++) {
    a[i] = balls[i]-balls[i-1]-1;
  }
  a[M-1]=n-balls[M-2]-1;
  return a;
}

const bool PolyChaos::NxtSmp(const int n, const int m, int* balls) const
{
  for (int i=0;i<m;i++) {
    if (balls[i]!=i) {
      balls[i]-=1;
      for (int j=i-1;j>=0;j--) {
	balls[j]=balls[j+1]-1;
      }
      return true;
    }
  }
  return false;
}

void PolyChaos::GenerateSequence(const int MV, const int pV)
{

  // Delete Sequences
  // ================
  delMem();
  M = MV;
  p=pV;

  // Construction of 1D Hermite Polynomials
  // ======================================
  H = new tdouble* [p+1];
  H[0] = new tdouble [1];
  H[1] = new tdouble [1];
  H[0][0] = 1.0;
  H[1][0] = 1.0;
  for (int i = 2; i <= p; i++) {
    int t = (i+2)/2;	// number of coefficients unequal to 0
    int g = (i+1) % 2;	// even number=1; uneven number=0
    H[i] = new tdouble [t];
    // Construction-part of previous polynomial
    int tu = (i+1)/2;	// number of coefficients of the previous polynomial
    H[i][0]=0;
    for (int j = 0; j < tu; j++) {
      H[i][j+g]=H[i-1][j];
    }
    // Construction-part of last but one
    for (int j = 0; j < t-1; j++) {
      H[i][j] += (-1.0)*(tdouble(i)-1.0)*H[i-2][j];
    }
  }
  
//   // DEBUG-OUTPUT
//   for (int i = 0; i<=p; i++) {
//     cout << i << ".: ";
//     for (int j = 0; j<(i+2)/2;j++){
// 	cout << H[i][j] << ", ";	
//     }
//     cout << endl;
//   }

// Construction of Polynomial Basis
// ================================

  jA = 0;
  for (int q=0;q<=p;q++) {
    jA+=tulong(BinomCoeff(M+q-1,M-1));
  }
//   cout << jA << endl;
  qS = new int*[jA+1];
    for (size_t i=0;i<=jA;++i) qS[i] = NULL;
  if (M==1) {
    for (int i=0;i<int(jA);i++) {
      qS[i] = new int [1];
      qS[i][0]=i;
    }
  } else {
    tulong pos=0;
    for (int q=0;q<=p;q++) {
      int n = M+q-1;
      int m = M-1;
      int *balls = new int[m];
      for (int i = 0; i<m; i++)
      {
	balls[i]=q+i;
      }
      do {
	qS[pos] = ExtrSeq(n,balls); 
	pos++;
      } while (NxtSmp(n,m,balls));
      delete [] balls;
    }
  }
  
//   // DEBUG-OUTPUT
//   cout << "Polynomial Basis" << endl;
//   for (int i = 0; i<jA;i++) {
//     for (int j=0; j<M;j++) {
//       cout << qS[i][j] << " " ;
//     }
//     cout << endl;
//   }

  ExpSq = new tdouble[jA];
  for (tulong i = 0; i<jA;i++) {
    ExpSq[i] = calc_ExpPB2(i);
  }
}

void PolyChaos::delMem()
{
  if (qS!=NULL) {
    for (tulong i=0;i<=jA;i++) {
      if (qS[i]) delete [] qS[i];
    }
    delete [] qS;
    qS=NULL;
  }
  if (H!=NULL) {
    for (int i = 0; i <= p; i++) {
      delete [] H[i];
    }
    delete [] H;
    H=NULL;
  }
  if (ExpSq!=NULL) {
    delete[] ExpSq;
    ExpSq=NULL;
  }
  M=0;
  p=0;
  jA=0;
}

const tdouble PolyChaos::get_ExpPB2(const tulong j) const
{
  #if FLX_DEBUG
    if ( j < 0 || j >= jA) {
      ostringstream ssV_2;
      ssV_2 << "j (j=" << j << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_ExpPB2_1", ssV_2.str() );
    }
  #endif
  return ExpSq[j];
}

FlxPeval* PolyChaos::get_H1D(const int q, const int m) const
{
  #if FLX_DEBUG
    if ( q < 0 || q > p) {
      ostringstream ssV_2;
      ssV_2 << "q (q=" << q << "; q.th Hermite Polynomail) has to be in the Range [0;" << p << "].";
      throw FlxException("PolyChaos::get_H1D_1", ssV_2.str() );
    }
    if ( m < 1 || m > M) {
      ostringstream ssV_2;
      ssV_2 << "ERROR";
      throw FlxException("PolyChaos::get_H1D_2", ssV_2.str() );
    }
  #endif
  FlxPeval* hm = new FlxPeval(M);
  int t = (q+2)/2;	// number of coefficients unequal to 0
  int u = (q) % 2;	// even number=0; uneven number=1
  for (int i=0; i<t;i++) {
    FlxPeval_term tp(M,m,i*2+u);
    hm->add_term(tp,H[q][i]);
  }
  return hm;
}

FlxPeval* PolyChaos::get_Poly(const tulong j) const
{
  #if FLX_DEBUG
    if ( j < 0 || j >= jA ) {
      ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("PolyChaos::get_Poly_1", ssV_2.str() );
    }
  #endif
  FlxPeval* hm = new FlxPeval(M,true);
  FlxPeval *hmT1; FlxPeval *hmT2;
  int *qSj = qS[j];
  for (int i = 0; i < M; ++i) {
    if (qSj[i]!=0) {
      hmT1 = hm;
      hmT2 = get_H1D(qSj[i],i+1);
      hm = new FlxPeval(*hmT1,*hmT2);
      delete hmT1;
      delete hmT2;
    }
  }
  return hm;  
}

const tdouble sfem::PolyChaos::calc_Realization(const flxVec& U, const flxVec& x) const
{
  #if FLX_DEBUG
    if ( U.get_N() != jA ) {
      ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("sfem::PolyChaos::calc_Realization_1", ssV_2.str() );
    }
    if (x.get_N() != static_cast<tuint>(M)) {
      ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("sfem::PolyChaos::calc_Realization_2", ssV_2.str() );
    }
  #endif
  tdouble u = U[0];
  for (tulong i = 1; i < jA; ++i) {
    u+= U[i]*calc_Poly(i,x);
  }
  return u;  
}

FlxPeval* PolyChaos::get_Realization(const flxVec& U) const
{
  FlxPeval* sp = new FlxPeval(M);
  sp->add_term(FlxPeval_term(M),U[0]);
  
  FlxPeval* spT;
  for (tulong i = 1; i < jA; ++i) {
    spT = get_Poly(i);
    spT->operator*=(U[i]);
    sp->operator+=(*spT);
    delete spT;
  }
  return sp; 
}

const tdouble PolyChaos::calc_ExpPB2(const tulong j) 
{
  #if FLX_DEBUG
    if ( j < 0 || j >= jA) {
      ostringstream ssV_2;
      ssV_2 << "j (j=" << j << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_ExpPB2_1", ssV_2.str() );
    }
  #endif
  tdouble E = 1;
  for (int i=0;i<M;i++) {
    E *= Factorial(qS[j][i]);
  }
  return E;
}

const tdouble PolyChaos::calc_c_ijk(const int i, tulong j, tulong k)  // TODO: inline?
{
  #if FLX_DEBUG
  // Checking for Input-Errors
    if ( j < 0 || j >= jA) {
      ostringstream ssV_2;
      ssV_2 << "j (j=" << j << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_c_ijk_2", ssV_2.str() );
    }
    if ( k < 0 || k >= jA) {
      ostringstream ssV_2;
      ssV_2 << "k (k=" << k << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_c_ijk_2", ssV_2.str() );
    }
    if ( i < 0 || i > M) {
      ostringstream ssV_2;
      ssV_2 << "i (i=" << i << "; has to be in the Range [0;" << M << "].";
      throw FlxException("sfem::PolyChaos::calc_c_ijk_3", ssV_2.str() );
    }
  #endif
  
  // i = 0 has to be treated separately
  if (i == 0) {
    if (j != k) return ZERO;
    else return ExpSq[j];
  }
  
  if (j==k) return ZERO;
  // Consider orthogonality
    for (int u=0;u<M;u++) {
      if (u!=i-1) {
	if (qS[j][u]!=qS[k][u]) return ZERO;
      }
    }
    int vp = qS[j][i-1];
    int vq = qS[k][i-1];
  // make sure vp > vq
    if (vp<vq) {
      std::swap(vp,vq);
      std::swap(j,k);
    }
    if (vp-1!=vq) return ZERO;
  // Compute c_ijk
    return ExpSq[j];
}

void PolyChaos::calc_Kjk(tulong j, tulong k, int& i, tdouble& f)
{
  #if FLX_DEBUG
  // Checking for Input-Errors
    if ( j < 0 || j >= jA) {
      ostringstream ssV_2;
      ssV_2 << "j (j=" << j << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_Kjk_1", ssV_2.str() );
    }
    if ( k < 0 || k >= jA) {
      ostringstream ssV_2;
      ssV_2 << "k (k=" << k << "; has to be in the Range [0;" << jA-1 << "].";
      throw FlxException("sfem::PolyChaos::calc_Kjk_2", ssV_2.str() );
    }
  #endif
  
  if ( j == k) {
    i = 0;
    f = ExpSq[j];
    return;
  }
  
  int *qS1 = qS[j];
  int *qS2 = qS[k];
  
  bool hit = false;
  int ihit = 0;
  for (int m = 0; m < M; m++) {
    if (qS1[m] != qS2[m]) {
      if (hit) {
	i = 0; 
	f = 0;
	return;
      } else {
	hit = true;
	ihit = m;
      }
    }
  }
  #if FLX_DEBUG
  // Checking for Input-Errors
    if ( !hit ) {
      ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("sfem::PolyChaos::calc_Kjk_3", ssV_2.str() );
    }
  #endif
  
  int vp = qS1[ihit];
  int vq = qS2[ihit];
  
  // make sure vp > vq
    if (vp<vq) {
      std::swap(vp,vq);
      std::swap(j,k);
    }
    if (vp-1!=vq) {
      i = 0; 
      f = 0;
      return;
    } else {
      i = ihit+1;
      f = ExpSq[j];
      return;
    }
}

void PolyChaos::dependOn_RV(tulong j, bVec& bv) const
{
  #if FLX_DEBUG
    if ( j < 0 || j >= jA ) {
      std::ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("sfem::PolyChaos::dependOn_RV_1", ssV_2.str() );
    }
    if (bv.size() != static_cast<tuint>(M) ) {
      std::ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("sfem::PolyChaos::dependOn_RV_2", ssV_2.str() );
    }
  #endif
  int *qSj = qS[j];
  for (register int i = 0; i < M; ++i) {
    bv[i] = (qSj[i]!=0);
  }
}


// -------------------------- flxStiffnessMtxSFEM -------------------------------------------------------

flxStiffnessMtxSFEM::flxStiffnessMtxSFEM(const int dim, const int order, flxFEM* FEM)
: flxStiffnessMtxBase(FEM),pc(dim,order),StfMtxV(NULL)
{

}

flxStiffnessMtxSFEM::~flxStiffnessMtxSFEM()
{
  delete_StfMtxV();
}

void flxStiffnessMtxSFEM::delete_StfMtxV()
{
  if (StfMtxV == NULL) return;
  int M = pc.get_dim();
  for (int i = 0; i <= M; ++i) {
    if (StfMtxV[i]) delete StfMtxV[i];
  }
  delete [] StfMtxV;
}

void flxStiffnessMtxSFEM::output_solVec(ostream& sout, bool verbose)
{
  if (SolutionVec==NULL) return;
  sout << "Solution of the system:" << std::endl;
  const tulong P = pc.get_SeqNumb();  
  flxVec sD(DofA);
  flxVec vtmp(DofA);
  for (tulong i = 1; i < P; ++i) {
    flxVec h1(&((*SolutionVec)[i*DofA]),DofA);
    vtmp.mult_coeff(h1,h1);
    sD.add(vtmp,pc.get_ExpPB2(i));
  }
  for (tnlong i = 0; i < DofA; ++i) {
    sout << i << " " << (*SolutionVec)[i] << " (" << sqrt(sD[i]) << ")" << std::endl;
  }
  if (verbose) {
    sout << (*SolutionVec) << std::endl;
  }
}

const tdouble flxStiffnessMtxSFEM::get_U(const tnlong dofindex) const
{
  if (SolutionVec==NULL) {
    std::ostringstream ssV;
    ssV << "Solution vector is empty.";
    throw FlxException("flxStiffnessMtxSFEM::get_U_1", ssV.str() );
  }
  if (dofindex >= DofA) return 0.0; 	// TODO predisplacement ??? FunNodeU::data
  if (FlxRnd_base::get_Current_rv_index() < 0) {
    const tulong P = pc.get_SeqNumb();  
    flxVec Uvec(P);
    Uvec.slice(&((*SolutionVec)[dofindex]),DofA);
    flxSFEM* SFEMptr = dynamic_cast<flxSFEM*>(FEM);
    #if FLX_DEBUG
      if ( SFEMptr == NULL ) {
	std::ostringstream ssV_2;
	ssV_2 << "Error.";
	throw FlxException("flxStiffnessMtxSFEM::get_U_2", ssV_2.str() );
      }
    #endif
    return pc.calc_Realization(Uvec,SFEMptr->RndBox->get_x_Vec());
  } else {
    return (*SolutionVec)[FlxRnd_base::get_Current_rv_index()*DofA+dofindex];
  }
}

void flxStiffnessMtxSFEM::get_U(const tnlong dofindex, tdouble* v, const tuint N) const
{
  for (tuint i=0;i<N;++i) {
    v[i] = get_U(dofindex+i);
  }
}

const tdouble flxStiffnessMtxSFEM::get_F(const flxDOFn* Dof)
{
  throw FlxException_NotImplemented("flxStiffnessMtxSFEM::get_F");
}

FlxPeval* flxStiffnessMtxSFEM::get_Pol(const tnlong dofindex)
{
  if (SolutionVec==NULL) {
    std::ostringstream ssV;
    ssV << "Solution vector is empty.";
    throw FlxException("flxStiffnessMtxSFEM::get_Pol_1", ssV.str() );
  }  
  if (dofindex >= DofA) return new FlxPeval(pc.get_dim()); 	// TODO predisplacement ??? FunNodeU::data
  
  const tulong P = pc.get_SeqNumb();  
  flxVec Uvec(P);
  Uvec.slice(&((*SolutionVec)[dofindex]),DofA);
  return pc.get_Realization(Uvec);
}

void flxStiffnessMtxSFEM::assemble_forc(bool errSerious)
{
  // TODO assemble force vector in case of random forces
  const tulong P = pc.get_SeqNumb();  
  if (ForceVec == NULL) ForceVec = new flxVec(P*DofA);
  
  flxVec* SSFEM_ForceVec = ForceVec;
  ForceVec = NULL;
  
  flxStiffnessMtxBase::assemble_forc();
  flxVec f_tmp(SSFEM_ForceVec->get_tmp_vptr(),tuint(DofA));
  f_tmp = *ForceVec;
  delete ForceVec;
  
  ForceVec = SSFEM_ForceVec;
}

void flxStiffnessMtxSFEM::allocate_SolutionVec()
{
  if (SolutionVec==NULL) {
    SolutionVec=new flxVec(pc.get_SeqNumb()*DofA);
  }
}

void flxStiffnessMtxSFEM::assemble_stf()
{
  if (preDispA>0) {
    throw FlxException_NotImplemented("flxStiffnessMtxSFEM::assemble_stf");
  }
  // First level of assembly
    int& Current_rv_index = FlxRnd_base::reference_Current_rv_index();
    const int cri = Current_rv_index;
    delete_StfMtxV();
    const int M = pc.get_dim();
    StfMtxV = new FlxMtx_baseS*[M+1];
      for (int i=0;i<=M;++i) StfMtxV[i] = NULL;
    for (int i = 0; i <= M; ++i) {
      Current_rv_index = i;
      flxStiffnessMtxBase::assemble_stf();
      StfMtxV[i] = StfMtx;
      StfMtx = NULL;
    }
    Current_rv_index = cri;
  // Second level of assembly
    tulong P = pc.get_SeqNumb();
    valarray<tdouble> fM((P*P+P)/2);
    valarray<int> KM((P*P+P)/2);
    int i;
    tdouble f;
    for (tulong j = 0, l=0; j < P; ++j) {
      for (tulong k = 0; k <= j; ++k,++l) {
	pc.calc_Kjk(j,k,i,f);
	if (f>0.0) {
	  fM[l] = f;
	  KM[l] = i;
	}
      }
    }
    StfMtx = new FlxMtxSparsSFEMSym(fM,KM,StfMtxV,P);
}




// -------------------------- SFEM ----------------------------------------------------------------------


flxSFEM::flxSFEM(int order, FlxConstantBox* constBox, FlxRndBox* RndBox)
 :flxFEM(constBox),RndBox(RndBox),p(order)
{
  
}

flxSFEM::~flxSFEM()
{
  if (StiffnessMtx) {
    RndBox->set_lock(false);
  }
}

flxStiffnessMtxSFEM* flxSFEM::get_Sstfmtx()
{
  if (StiffnessMtx == NULL) {
    std::ostringstream ssV;
    ssV << "The stiffness matrix has not yet been assembled.";
    throw FlxException("flxSFEM::get_Sstfmtx_1", ssV.str() );
  }
  flxStiffnessMtxSFEM* sStfMtx = dynamic_cast<flxStiffnessMtxSFEM*> (StiffnessMtx);
  #if FLX_DEBUG
    if (sStfMtx == NULL) {
      std::ostringstream ssV;
      ssV << "This is not a stochastic finite element analysis.";
      throw FlxException("flxSFEM::get_Sstfmtx_2", ssV.str() );
    }
  #endif
  return sStfMtx;
}

FlxPeval* flxSFEM::get_Pol(const tnlong dofindex)
{
  flxStiffnessMtxSFEM* sStfMtx = get_Sstfmtx();
  return sStfMtx->get_Pol(dofindex);
}

void flxSFEM::assembleDOF()
{
  if (RndBox->has_userDefined()) {
    std::ostringstream ssV;
    ssV << "There exist user defined random variables. The stochastic finite element module can not deal with them.";
    throw FlxException("flxSFEM::assembleDOF", ssV.str() );
  }
  if (StiffnessMtx) {
    delete StiffnessMtx;
  } else {
    RndBox->set_lock(true);
  }
  StiffnessMtx = new flxStiffnessMtxSFEM(RndBox->get_NumbRnd(), p, this);
}

const tnlong calc_PC_projection_numerical_MCI::calc_expectation(flxVec& IntegT, FunBase* fun, const tnlong& N_Interv, const tnlong& N_mciI_1, const tnlong& N_mciI_2, const tdouble& pRedistribute)
{
  GlobalVar.slog(4) << "rsm_set pc_project: estimate the moments using Monte Carlo integration ..." << std::endl;
  #if FLX_DEBUG
    if (IntegT.get_N() != PC.get_SeqNumb() ) {
      throw FlxException_Crude("calc_PC_projection_numerical_MCI::calc_expectation");
    }
  #endif
  P = IntegT.get_N();
  funR = fun;
  // remember x and y
    const flxVec orig_X = RndBox->get_x_Vec();
    const flxVec orig_Y = RndBox->get_y_Vec();
     
  flxpVec Ivec(P*N_Interv);
  tnlong i,j;
  uD = new FlxRnd_Uniform(-1.0,1.0);

  // create correlation information
    RndBox->corr_create(true);
  // create reference vectors
    y_smp = RndBox->y_smp;
    x_smp = RndBox->x_smp;
    
  const tdouble Bound=GlobalVar.Exp_Num_Bound_MCI;
  const tdouble step=Bound/N_Interv;
  const tdouble &h_inv = Bound;
  
  std::valarray<tnlong> N_Vec(N_mciI_1,N_Interv);
  if (N_Interv==1) {
    N_Vec[0] += N_mciI_2;
    do_IntegrationLoop(Ivec,step,N_Interv,N_Vec);
  } else {
    // estimate the integral
    GlobalVar.slog(4) << "  estimation step ..." << std::endl;
    do_IntegrationLoop(Ivec,step,N_Interv,N_Vec);
    
    // perform a better integration based on the estimate (for each term in the PC)   
    std::valarray<tnlong> N_Vec_sub(N_Interv);
    tdouble t;
    for (i=0; i<P; ++i) {
      GlobalVar.slog(4) << "  term " << i << " (out of " << P << ") ..." << std::endl;
      t = 0.0;
      for (j=0; j<N_Interv; ++j) {
	t+=sqrt(fabs( (Ivec[j*P+i]).cast2double() )/N_Vec[j]);	// evaluate the estimated ith integral value (square)
      }
      // redistribution
      tnlong max = 0;
      tnlong maxI = 0;
      tnlong count = 0;
      for (j=0;j<N_Interv;++j) {
	N_Vec_sub[j] = tnlong(std::floor((sqrt(fabs((Ivec[j*P+i]).cast2double())/N_Vec[j])/t)*N_mciI_2));
	if (N_Vec_sub[j]>max) {
	  maxI = j;
	  max = N_Vec_sub[j];
	}
	count+=N_Vec_sub[j];
      }
      N_Vec_sub[maxI] += N_mciI_2 - count;
      // redistribute (diffusion)
	count = 0;
	tnlong rd = tnlong(N_Vec_sub[0]*pRedistribute/2.0);
	N_Vec_sub[0]-=rd;
	count +=N_Vec_sub[0];
	for (j=1;j<N_Interv-1;++j) {
	  N_Vec_sub[j]+=rd;
	  rd = tnlong((N_Vec_sub[j]-rd)*pRedistribute/2.0);
	  N_Vec_sub[j]-=2*rd;
	  N_Vec_sub[j-1]+=rd;
	  count += N_Vec_sub[j]+rd;
	}
	N_Vec_sub[N_Interv-1]+=rd;
	rd = tnlong((N_Vec_sub[N_Interv-1]-rd)*pRedistribute/2.0);
	N_Vec_sub[N_Interv-1]-=rd;
	N_Vec_sub[N_Interv-2]+=rd;
	count += N_Vec_sub[N_Interv-1]+rd;
	N_Vec_sub[maxI] += N_mciI_2 - count;
      // take old results into account
	for (j=0;j<N_Interv;++j) {
	  if (N_Vec[j]>=N_Vec_sub[j]) N_Vec_sub[j] = 0;
	  else N_Vec_sub[j]-=N_Vec[j];
	}
      // evaluate the integral (estimate based on previous estimate)
      do_IntegrationLoop(Ivec,step,N_Interv,N_Vec_sub);
      N_Vec+=N_Vec_sub;
    }
  }

  // sum of everything
    j=0;
    for (i=0;i<N_Interv;++i) {
      j+=N_Vec[i];
    }
    const tdouble jm = tdouble(j)/tdouble(N_Interv);
    flxpVec Integ(P);
    for (i=0;i<N_Interv;++i) {
      flxpVec IvecP(&(Ivec[i*P]),P);
      Integ.add(IvecP,jm/N_Vec[i]);
    }
    IntegT = Integ;
    IntegT *= h_inv / j;
  
  // restore x and y
    *(RndBox->x_smp) = orig_X;
    *(RndBox->y_smp) = orig_Y;
  // free memory
    funR = NULL;
    delete uD;
    GlobalVar.slog(4) << "  samples required for the integration: " << GlobalVar.Double2String(j) << std::endl;
    return j;
}

void calc_PC_projection_numerical_MCI::do_IntegrationLoop( flxpVec& Ivec, const tdouble& step, const tnlong& N_Interv, const valarray< tnlong >& N_Vec)
{
  tdouble x_u=0.0, x_o=step;
  tnlong N_mciI_T, N_mciI_T_I;
  flxpVec IntegI(P), IntegI2(P);
  register tnlong i,j,m;
  for (i=0;i<N_Interv;++i) {
    // change properties of uniform distribution
      uD->reset_bound(x_u,x_o);
    IntegI.set_zero();;
    N_mciI_T = N_Vec[i];
    if (N_mciI_T>=1e4) {
      N_mciI_T_I = tnlong(floor(sqrt(tdouble(N_mciI_T))));
      for (int l=1;l<=2;++l) {
	for (j=(l==2)?N_mciI_T_I-1:1;j<N_mciI_T_I;++j) {
	  IntegI2.set_zero();;
	  for (m=0;m<N_mciI_T_I;++m) {
	    Integrationstep(IntegI2);
	  }
	  IntegI+=IntegI2;
	}
	N_mciI_T_I = N_mciI_T - pow2(N_mciI_T_I) + N_mciI_T_I;
      }
    } else {
      for (j=0;j<N_mciI_T;++j) {
	Integrationstep(IntegI);
      }
    }
    flxpVec Ivec_tmp(&(Ivec[i*P]),P);
    Ivec_tmp+=IntegI;
    x_u = x_o;
    x_o += step;
  }
}

void calc_PC_projection_numerical_MCI::Integrationstep(flxpVec& IntegI)
{
  // generate y-vector
    RndBox->RndCreator.gen_smp(*y_smp);

  // transform to z-space
    const tdouble ly2 = y_smp->get_Norm2_NOroot();
    const tdouble lz2 = uD->Inv_cdf_x(rv_cdf_ChiSquare(DIM,ly2));
    (*y_smp)*= sqrt(lz2/ly2);
  
  // calculate f
    const tdouble f = rv_pdf_ChiSquare(DIM,lz2);
    
  // update standard normal and original space
    RndBox->set_y_Vec(*y_smp);
    
  // return function * foverh;
    const tdouble cv = funR->calc() * f;
    for (register tnlong i=0;i<P;++i) {
      IntegI[i]+=cv*PC.calc_Poly(i,*y_smp);      
    }
}


